/*
 * agent.c - Multi-AP easymesh agent
 *
 * Copyright (C) 2019-2023 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for licensing information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/stat.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>

#if (EASYMESH_VERSION >2)
#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/ec.h>
#endif

#include <linux/if_bridge.h>
#include <netlink/socket.h>

#include <netinet/in.h>
#include <net/if.h>

#include <easy/easy.h>
#include <cmdu.h>
#include <1905_tlvs.h>
#include "wifi.h"

#include <i1905_wsc.h>

#include <uci.h>
#include <map_module.h>
#include <easymesh.h>
#include <dirent.h>

#include "timer.h"
#include "utils/1905_ubus.h"
#include "utils/utils.h"
#include "utils/debug.h"
#include "utils/liblist.h"

#include "steer_rules.h"
#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
#include <dpp_api.h>
#endif
#include "utils/dpp_sock.h"
#include "utils/keygen.h"
#include "dpp.h"
#endif
#include "config.h"
#include "nl.h"
#include "agent.h"
#include "agent_ubus.h"
#include "agent_ubus_dbg.h"
#include "agent_map.h"
#include "cmdu_extension.h"
#include "agent_cmdu.h"
#include "agent_tlv.h"
#include "backhaul_blacklist.h"
#include "backhaul.h"
#if (EASYMESH_VERSION > 2)
#include "qos.h"
#endif

#define map_plugin	"ieee1905.map"

#define IFACE_TIMEOUT 1
#define BOOT_UP_SCAN_TIME	(0 * 1000)
#define BOOT_UP_SCAN_ITV	(2 * 1000)
#define BOOT_UP_SCAN_MAX_TRY 5

static struct agent *this_agent;

static int agent_subscribe_for_cmdus(struct agent *a);
static void agent_free_nodes(struct agent *a);
static int wifi_topology_notification(struct agent *agent, uint8_t *mac,
		char *ifname, uint8_t assoc_event, uint16_t reason);
static int wifi_del_sta(struct agent *a, const char *vif,
		unsigned char *macaddr);
static int agent_req_btm_steer(const char *ifname, unsigned char *sta_macaddr,
		uint8_t pref_num, struct nbr *pref, uint32_t validity_int);



static void bsta_steer_cb(atimer_t *t)
{
	struct netif_bk *bk = container_of(t, struct netif_bk, steer_timeout);
	struct agent *a = bk->agent;
	char fmt[64] = {0};
	uint8_t *prev_bssid = bk->bsta_steer.prev_bssid;

	dbg("|%s:%d| steer timer expired, restoring old bssid (" MACFMT ") for"\
	    "bsta %s\n", __func__, __LINE__, MAC2STR(prev_bssid), bk->ifname);

	wifi_set_iface_bssid(bk, prev_bssid);

	/* Flawfinder: ignore */
	snprintf(fmt, sizeof(fmt), "set_network_bssid %s " MACFMT,
		 bk->ifname, MAC2STR(prev_bssid));
	agent_exec_platform_scripts(fmt);

	bk->bsta_steer.expired = true;


	if (bk->bsta_steer.trigger == BK_STEER_LOCAL) {
		agent_bsta_steer_clear(bk);
	} else if (a->connected) {
		/* if connected we can send response right away */
		send_backhaul_sta_steer_response(a, bk, 0x01,
						 bk->bsta_steer.ul_ifname);
		agent_bsta_steer_clear(bk);
	}
}

static void agent_sighandler(int sig)
{
	uloop_end();
}

int agent_exec_platform_scripts(char *arg)
{
	char buf[16] = {0};

	warn("/lib/wifi/multiap %s\n", arg);
	chrCmd(buf, sizeof(buf), "/lib/wifi/multiap %s", arg);
	return atoi(buf);
}

/* find node by macaddress */
struct node *agent_find_node(struct agent *a, uint8_t *almac)
{
	struct node *n = NULL;

	list_for_each_entry(n, &a->nodelist, list) {
		if (!memcmp(n->alid, almac, 6))
			return n;
	}

	return NULL;
}

struct node *agent_alloc_node(struct agent *a, uint8_t *almac)
{
	struct node *n;

	n = calloc(1, sizeof(struct node));
	if (!n) {
		warn("OOM: node malloc failed!\n");
		return NULL;
	}

	n->agent = a;
	memcpy(n->alid, almac, 6);
	n->map_profile = MULTIAP_PROFILE_1;

	list_add(&n->list, &a->nodelist);
	a->num_nodes++;

	dbg("|%s:%d| ------- " MACFMT "\n", __func__, __LINE__, MAC2STR(almac));
	return n;
}

struct node *agent_add_node(struct agent *a, uint8_t *almac)
{
	struct node *n;

	n = agent_find_node(a, almac);
	if (!n) {
		n = agent_alloc_node(a, almac);
		if (!n) {
			err("|%s:%d| failed to allocate node "MACFMT"\n",
			    __func__, __LINE__, MAC2STR(almac));
			return NULL;
		}
	} else {
		return n;
	}

	/* Add actions
	 * 1. Write to UCI?
	 */

#if (EASYMESH_VERSION > 2)
	qos_sync_rules_to_nodes(a);
#endif

	return n;
}

const char *wifi_ifname_to_radio(struct agent *a, char *ifname)
{
	int i, j;
	const char *ret;

	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		for (j = 0; j < WIFI_IFACE_MAX_NUM; j++) {
			if (a->ifs[i].iface[j].name[0] != '\0' &&
				!strncmp(a->ifs[i].iface[j].name, ifname, 15)) {

				ret = a->ifs[i].radio;
				return ret;
			}
		}
	}

	trace("|%s %d| couldn't find radio for ifname %s\n",
		   __func__, __LINE__, ifname);

	return NULL;
}

struct wifi_netdev *wifi_radio_to_netdev(struct agent *a,
		const char *radio)
{
	int i;

	if (!radio)
		return NULL;

	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		if (a->ifs[i].radio[0] != '\0' &&
			!strncmp(a->ifs[i].radio, radio, 15)) {

			return &a->ifs[i];
		}
	}

	return NULL;
}

struct wifi_radio_element *wifi_radio_to_radio_element(struct agent *a,
		const char *radio)
{
	int i;

	if (!radio)
		return NULL;

	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		if (a->radios[i].name[0] != '\0' &&
			!strncmp(a->radios[i].name, radio, 15)) {

			return &a->radios[i];
		}
	}

	return NULL;
}

struct wifi_netdev *wifi_radio_id_to_netdev(struct agent *a,
							  uint8_t *radio_id)
{
	int i;

	if (!radio_id)
		return NULL;

	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		if (!memcmp(a->ifs[i].re->macaddr, radio_id, 6))
			return &a->ifs[i];
	}

	return NULL;
}

struct wifi_radio_element *wifi_radio_id_to_radio_element(struct agent *a,
							  uint8_t *radio_id)
{
	int i;

	if (!radio_id)
		return NULL;

	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		if (!memcmp(a->radios[i].macaddr, radio_id, 6))
			return &a->radios[i];
	}

	return NULL;
}

struct wifi_radio_element *wifi_ifname_to_radio_element(struct agent *a,
		char *ifname)
{
	const char *radio;

	radio = wifi_ifname_to_radio(a, ifname);
	if (!radio)
		return NULL;

	return wifi_radio_to_radio_element(a, radio);
}

struct wifi_radio_element *wifi_get_radio_by_mac(struct agent *a,
		uint8_t *hwaddr)
{
	struct wifi_radio_element *radio;
	int i;

	for (i = 0; i < a->num_radios; i++) {
		radio = a->radios + i;

		if (memcmp(radio->macaddr, hwaddr, 6))
			continue;

		return radio;
	}

	return NULL;
}

static void wifiagent_log_steer(struct agent *a,
				unsigned char *stamac,
				char *vifname,
				char *steer_type,
				unsigned char *to_bss)
{
	char ev[512] = {0};
	unsigned char macaddr[6] = {0};
	char ifname[IFNAMSIZ] = {0};
	char type[32] = {0};

	if (!stamac || !vifname || !steer_type)
		return;

	memcpy(macaddr, stamac, 6);
	strncpy(ifname, vifname, IFNAMSIZ - 1);
	strncpy(type, steer_type, 31);

	snprintf(ev, sizeof(ev), "{\"macaddr\":\""MACFMT"\""
		 ",\"vif\":\"%.16s\",\"action\":\"%.32s\"",
		 MAC2STR(macaddr), ifname, type);
	if (to_bss) {
		snprintf(ev + strlen(ev), sizeof(ev), ",\"to\":\""MACFMT"\"}",
							MAC2STR(to_bss));
	} else {
		snprintf(ev + strlen(ev), sizeof(ev), "%s", "}");
	}

	info("steer: %s\n", ev);
	agent_notify_event(a, "map.agent", ev);
}

static void wifiagent_log_stainfo(struct agent *a, struct sta *s)
{
	int i;
	char ev[512] = {0};
	char rssis[64] = {0};

	if (!a || !s)
		return;

	for (i = 0; i < 4; i++) {
		if (s->rssi[i] > -100 && s->rssi[i] <= -10)
			snprintf(rssis + strlen(rssis), sizeof(rssis), "%d%s",
					s->rssi[i], i == 3 ? "" : ",");
	}

	snprintf(ev, sizeof(ev),
			"{\"macaddr\":\""MACFMT"\""
			",\"rssi\":\"%s\""
			",\"tx_rate\":%u"
			",\"rx_rate\":%u"
			",\"tx_Bps\":%d"
			",\"rx_Bps\":%d"
			",\"tx_pkts\":%" PRIu64
			",\"rx_pkts\":%" PRIu64
			",\"tx_fpkts\":%u,\"rx_fpkts\":%u,\"rtx_pkts\":%u}",
			MAC2STR(s->macaddr), rssis, s->tx_rate, s->rx_rate,
			s->tx_thput, s->rx_thput,
			s->tx_pkts,
			s->rx_pkts,
			s->tx_fail_pkts,
			s->rx_fail_pkts,
			s->rtx_pkts);

	//trace("stainfo: %s\n", ev);
	agent_notify_event(a, "map.agent", ev);
}

void wifiagent_log_cntlrinfo(struct agent *a)
{
	char ev[512] = {0};

	dbg("%s: called.\n", __func__);

	if (!a)
		return;

	sprintf(ev,
		"{\"active_cntlr\":%u"
		",\"cntlr_almac\":\""MACFMT"\""
		",\"local_cntlr\":%u"
		",\"multiple_cntlr_found\":%u}",
		a->active_cntlr ? true : false,
		MAC2STR(a->cntlr_almac),
		is_local_cntlr_running() ? 1 : 0,
		a->multiple_cntlr ? 1 : 0);

	trace("cntlrinfo: %s\n", ev);
	agent_notify_event(a, "map.agent", ev);
}

// static
int ubus_call_object(struct agent *a, wifi_object_t wobj,
		const char *method,
		void (*response_cb)(struct ubus_request *, int, struct blob_attr *),
		void *priv)
{
	struct blob_buf bb = {};
	int ret;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(a->ubus_ctx, wobj, method, bb.head,
				response_cb, priv, 2 * 1000);
	if (ret) {
		err("Failed to get '%s' (ret = %d)\n", method, ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
	return 0;
}

// static
uint32_t ubus_get_object(struct ubus_context *ctx, const char *name)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(ctx, name, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", name);
		return WIFI_OBJECT_INVALID;
	}

	return id;
}

struct wifi_assoc_frame *wifi_get_frame(struct agent *a, uint8_t *macaddr)
{
	struct wifi_assoc_frame *f;

	list_for_each_entry(f, &a->framelist, list) {
		trace("frame mac: " MACFMT " input: " MACFMT "\n",
				MAC2STR(f->macaddr),
				MAC2STR(macaddr));
		if (!memcmp(f->macaddr, macaddr, 6))
			return f;
	}

	return NULL;
}

/* lookup netif struct by name */
struct netif_bk *agent_get_netif_bk_by_name(struct agent *a, const char *name)
{
	struct netif_bk *p;

	list_for_each_entry(p, &a->bklist, list) {
		if (!strncmp(name, p->ifname, 16))
			return p;
	}

	return NULL;
}

/* lookup netif_bk struct by device */
struct netif_bk *agent_get_netif_bk_by_device(struct agent *a,
		const char *device)
{
	struct netif_bk *p;

	list_for_each_entry(p, &a->bklist, list) {
		if (!strncmp(device, p->cfg->device, 16))
			return p;
	}

	return NULL;
}


/* lookup netif struct by name */
struct netif_ap *get_netif_by_name(struct agent *a, const char *name)
{
	struct netif_ap *p;

	list_for_each_entry(p, &a->aplist, list) {
		if (!strncmp(name, p->ifname, 16))
			return p;
	}

	return NULL;
}

/* get netif_ap based on bssid */
struct netif_ap *wifi_get_netif_by_bssid(struct agent *a, uint8_t *bssid)
{
	struct netif_ap *ap;

	list_for_each_entry(ap, &a->aplist, list) {
		if (hwaddr_equal(ap->bssid, bssid))
			return ap;
	}

	return NULL;
}

/* find sta by macaddress */
struct sta *find_sta_by_mac(struct agent *a, uint8_t *mac)
{
	struct netif_ap *p;

	list_for_each_entry(p, &a->aplist, list) {
		struct sta *s;

		list_for_each_entry(s, &p->stalist, list) {
			if (!memcmp(s->macaddr, mac, 6))
				return s;
		}
	}

	return NULL;
}

struct netif_bk *find_bkhaul_by_band(struct agent *a, enum wifi_band band)
{
	struct netif_bk *p;

	list_for_each_entry(p, &a->bklist, list) {
		if (p->cfg && p->cfg->band == band)
			return p;
	}

	return NULL;
}

struct netif_bk *find_bkhaul_by_ifname(struct agent *a, char *ifname)
{
	struct netif_bk *p;

	list_for_each_entry(p, &a->bklist, list) {
		if (!strncmp(p->ifname, ifname, sizeof(p->ifname)))
			return p;
	}

	return NULL;
}

/* find bkhaul by hwaddr */
struct netif_bk *find_bkhaul_by_bssid(struct agent *a, uint8_t *bssid)
{
	struct netif_bk *p;

	list_for_each_entry(p, &a->bklist, list) {
		if (!memcmp(p->macaddr, bssid, 6))
			return p;
	}

	return NULL;
}

/* find wds by interface name */
struct netif_wds *find_wds_by_ifname(struct agent *a, const char *ifname)
{
	struct netif_wds *p;

	list_for_each_entry(p, &a->wdslist, list) {
		if (!strncmp(p->ifname, ifname, sizeof(p->ifname)))
			return p;
	}

	return NULL;
}

void agent_link_ap_to_cfg(struct agent *a)
{
	struct netif_apcfg *acfg;

	list_for_each_entry(acfg, &a->cfg.aplist, list) {
		struct netif_ap *f;

		f = get_netif_by_name(a, acfg->name);
		if (!f)
			continue;

		f->cfg = acfg;
	}
}


void agent_config_load_post_action(struct agent *a)
{
	agent_link_ap_to_cfg(a);
	/* TODO: agent_link_dpp_to_bkcfg(a); */

	if (a->cfg.pcfg)
		a->pvid = a->cfg.pcfg->pvid;

#if (EASYMESH_VERSION > 2)
	/* TODO: perhaps these fit better under agent_init_interface? */

	/* TODO: every single one of these iterates the URI list --
	 * inefficient, pass URI as argument instead
	 */
//	dpp_gen_uri(a);
//	dpp_chirp(a);
//	dpp_listen_enrollee(a);
#endif
}

int agent_config_reload(struct agent *a)
{
	int ret;

	ret = agent_config_load(&a->cfg);
	agent_config_load_post_action(a);

	return ret;
}

typedef void (*destructor)(void *);

#define delete_expired_entries(vif, type, h, l, ts_member, tmo, func, _nr) \
do {									   \
	type *e, *etmp;							   \
	typeof((destructor) func) fptr = (func);			   \
									   \
	list_for_each_entry_safe(e, etmp, h, l) {			   \
		if (timestamp_expired(&e->ts_member, tmo)) {		   \
			dbg("%s: Entry aged out.. delete.\n", __func__);   \
			if (fptr)					   \
				fptr(e);				   \
			list_del(&e->l);				   \
			(_nr)--;                                           \
			free(e);					   \
		}							   \
	}								   \
} while (0)

#ifdef VENDOR_EXTENSION
static int agent_load_extensions(struct agent *a)
{
#define EXTENSION_PLUGIN_MAX_NUM	8
	char *argv[EXTENSION_PLUGIN_MAX_NUM] = {0};
	struct agent_extension_config *e;
	int argc = 0;
	int ret;


	list_for_each_entry(e, &a->cfg.extlist, list) {
		argv[argc++] = e->name;

		if (argc >= EXTENSION_PLUGIN_MAX_NUM)
			break;
	}

	info("map-agent: load extensions\n");
	INIT_LIST_HEAD(&a->extlist);
	ret = load_extensions(argc, argv, &a->extlist);

	return ret;
}

static void agent_unload_extensions(struct agent *a)
{
	if (a)
		unload_extensions(&a->extlist);
}
#endif

int (*compare)(const void *a, const void *b);

/* Private 'cmp()' function to sort neighbors by RSSI decreasing */
static int cmp_nbr_rssi(void *priv, struct list_head *a, struct list_head *b)
{
	struct pref_neighbor *ea, *eb;

	UNUSED(priv);
	ea = container_of(a, struct pref_neighbor, list);
	eb = container_of(b, struct pref_neighbor, list);

	return eb->rssi - ea->rssi;
}

#define alloc_entry(type)		\
({					\
	type *__n;			\
	__n = calloc(1, sizeof(type));	\
	__n ? &__n->list : NULL;	\
})

#define free_entry(ptr, type)					\
({								\
	if (ptr)						\
		free(container_of(ptr, type, list));		\
})

#define copy_entry(from, to, type)			\
({							\
	type *__s, *__d;				\
	if (from && to)	{				\
		__s = container_of(from, type, list);	\
		__d = container_of(to, type, list);	\
		memcpy(__d, __s, sizeof(type));		\
	}						\
})

static void *alloc_neighbor_entry(void)
{
	struct neighbor *n;

	n = calloc(1, sizeof(struct neighbor));
	if (n)
		return &n->list;

	return NULL;
}

static void free_neighbor_entry(struct list_head *n)
{
	free_entry(n, struct neighbor);
}

static void copy_neighbor_entry(struct list_head *from, struct list_head *to)
{
	copy_entry(from, to, struct neighbor);
}

static void *alloc_sta_neighbor_entry(void)
{
	struct sta_neighbor *n;

	n = calloc(1, sizeof(struct sta_neighbor));
	if (n)
		return &n->list;

	return NULL;
}

static void free_sta_neighbor_entry(struct list_head *n)
{
	free_entry(n, struct sta_neighbor);
}

static void copy_sta_neighbor_entry(struct list_head *from,
		struct list_head *to)
{
	copy_entry(from, to, struct sta_neighbor);
}

/* Function used to match duplicate entries in neighbor list */
static int match_bssid(void *priv, struct list_head *a, struct list_head *b)
{
	struct neighbor *ea;
	struct sta_neighbor *eb;

	UNUSED(priv);
	ea = container_of(a, struct neighbor, list);
	eb = container_of(b, struct sta_neighbor, list);

	return hwaddr_equal(ea->nbr.bssid, eb->nbr.bssid);
}

static struct list_head *create_joined_node(void *priv, struct list_head *a,
		struct list_head *b)
{
	struct pref_neighbor *new;
	struct neighbor *ea = NULL;
	struct sta_neighbor *eb = NULL;
	struct wifi_bss bss;
	struct sta *sta = (struct sta *)priv;
	struct netif_ap *nif = NULL;

	if (a)
		ea = container_of(a, struct neighbor, list);

	if (b)
		eb = container_of(b, struct sta_neighbor, list);

	if (!ea && !eb)
		return NULL;

	new = malloc(sizeof(*new));
	if (!new) {
		warn("OOM: failed to alloc preferred neighbor!\n");
		return NULL;
	}

	memset(new, 0, sizeof(*new));

	if (ea)
		memcpy(new->bssid, ea->nbr.bssid, 6);
	else
		memcpy(new->bssid, eb->nbr.bssid, 6);

	if (sta)
		nif = sta->vif;

	memset(&bss, 0, sizeof(struct wifi_bss));
	if (nif && wifi_scanresults_get_bss(nif->ifname, new->bssid, &bss)) {
		/* new->rssi = bss.rssi; */ /* it is AP's view; meaningless */
		new->ch_util = bss.load.utilization;
		new->num_stas = bss.load.sta_count;
	}

	if (ea) {
		new->bssid_info = ea->nbr.bssid_info;
		new->reg = ea->nbr.reg;
		new->channel = ea->nbr.channel;
		new->phy = ea->nbr.phy;
		new->flags = ea->flags;
		/* TODO: est_bwrate from snr */
	}

	if (eb) {
		new->rssi = eb->nbr.rssi; /* override rssi by sta's view */
		new->rsni = eb->nbr.rsni;
	}

	return &new->list;
}

static void free_joined_node(void *priv, struct list_head *a)
{
	struct pref_neighbor *ea = NULL;

	UNUSED(priv);
	if (a)
		ea = container_of(a, struct pref_neighbor, list);

	free(ea);
}

static void recalc_desired_neighbors(struct sta *s)
{
	struct netif_ap *vif = s->vif;
	struct list_head nbrlist_copy, sta_nbrlist_copy;

	/**
	 * First, does a splice of the two lists -
	 *	11k neighbor list and
	 *	11k STA beacon list.
	 *
	 * It then rearranges the combined list on the basis of decreasing
	 * 'rssi'.
	 * In rule matching modules, walk the pref_nbrlist to select the
	 * next best neighbor Bss based on rssi, bssload, est_dl_throughput etc.
	 */

	INIT_LIST_HEAD(&nbrlist_copy);
	list_dup(&vif->nbrlist, &nbrlist_copy,
			alloc_neighbor_entry,
			free_neighbor_entry,
			copy_neighbor_entry);

	INIT_LIST_HEAD(&sta_nbrlist_copy);
	list_dup(&s->sta_nbrlist, &sta_nbrlist_copy,
			alloc_sta_neighbor_entry,
			free_sta_neighbor_entry,
			copy_sta_neighbor_entry);

	list_flush(&s->pref_nbrlist, struct pref_neighbor, list);

	dbg_list_print("Neighbor nodes", &nbrlist_copy,
			struct neighbor, list, nbr.bssid);

	list_join_uniq(s, &nbrlist_copy, &sta_nbrlist_copy, &s->pref_nbrlist,
			match_bssid, create_joined_node, free_joined_node,
			free_neighbor_entry, free_sta_neighbor_entry);

	dbg_list_print("STA Neighbors", &s->pref_nbrlist,
			struct pref_neighbor, list, bssid);

	list_flush(&nbrlist_copy, struct neighbor, list);
	list_flush(&sta_nbrlist_copy, struct sta_neighbor, list);

	list_sort(NULL, &s->pref_nbrlist, cmp_nbr_rssi);
}

static int update_sta_entry(struct netif_ap *vif, struct wifi_sta *e)
{
	struct sta *s;

	list_for_each_entry(s, &vif->stalist, list) {
		if (!memcmp(s->macaddr, e->macaddr, 6)) {
			bool no_traffic = true;

			/* dbg("%s: update STA " MACFMT " entry\n", __func__,
			 *	MAC2STR(s->macaddr));
			 */
			timestamp_update(&s->last_update);
			if (s->rssi[0] != e->rssi[0] ||
					s->rssi[1] != e->rssi[1] ||
					s->rssi[2] != e->rssi[2] ||
					s->rssi[3] != e->rssi[3]) {

				no_traffic = false;

				s->rssi[0] = e->rssi[0];
				s->rssi[1] = e->rssi[1];
				s->rssi[2] = e->rssi[2];
				s->rssi[3] = e->rssi[3];
			}

			s->connected_ms = e->conn_time * 1000;
			s->tx_rate = e->tx_rate.rate;
			s->rx_rate = e->rx_rate.rate;

			s->tx_thput = 0;
			s->rx_thput = 0;
			if (e->stats.tx_bytes != s->tx_bytes ||
					e->stats.rx_bytes != s->rx_bytes) {
				uint64_t tx_bytes_lastsec;
				uint64_t rx_bytes_lastsec;

				no_traffic = false;

				tx_bytes_lastsec = s->tx_bytes;
				rx_bytes_lastsec = s->rx_bytes;
				s->tx_bytes = e->stats.tx_bytes;
				s->rx_bytes = e->stats.rx_bytes;
				s->tx_thput = s->tx_bytes - tx_bytes_lastsec;
				s->rx_thput = s->rx_bytes - rx_bytes_lastsec;
			}

			s->est_rx_thput = e->est_rx_thput;
			s->est_tx_thput = e->est_tx_thput;

			if (e->stats.tx_pkts != s->tx_pkts ||
					e->stats.rx_pkts != s->rx_pkts) {

				no_traffic = false;

				s->tx_pkts = e->stats.tx_pkts;
				s->rx_pkts = e->stats.rx_pkts;
			}

			if (no_traffic)
				s->stale_stats_cnt++;
			else
				s->stale_stats_cnt = 0;

			s->tx_fail_pkts = e->stats.tx_fail_pkts;
			s->rx_fail_pkts = e->stats.rx_fail_pkts;
			s->rtx_pkts = e->stats.tx_retry_pkts;

		/*loud("STA: " MACFMT " (rssi = %d)   " \
				"Tx-bytes/sec = %u    Rx-bytes/sec = %u\n",
				MAC2STR(s->macaddr), s->rssi[0],
				s->tx_thput, s->rx_thput);*/
			wifiagent_log_stainfo(vif->agent, s);
			return 0;
		}
	}

	return -1;
}

static int agent_req_beacon_metrics(struct agent *a,
	struct netif_ap *ap, uint8_t *sta_addr, uint8_t opclass,
	uint8_t channel, uint8_t *bssid, uint8_t reporting_detail,
	uint8_t ssid_len, char *ssid, uint8_t num_report,
	uint8_t *report, uint8_t num_element, uint8_t *element)
{
	struct wifi_request_neighbor_param param = {};

	trace("agent: %s: --->\n", __func__);

	if (!sta_addr)
		return -1;

	param.opclass = opclass;
	param.channel = channel;
	param.bssid = bssid;
	param.reporting_detail = reporting_detail;
	param.ssid_len = ssid_len;
	param.ssid = ssid;
	param.num_report = num_report;
	param.report = report;
	param.num_element = num_element;
	param.element = element;

	return wifi_req_neighbor(ap->ifname, sta_addr, &param);
}

static void wifi_sta_bcn_req(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_bcn_req_timer);
	struct netif_ap *vif = s->vif;
	struct agent *a = vif->agent;
	struct sta_bcn_req *breq;

	if (s->sta_bcn_req_nr < 1)
		/* No request enqueued */
		return;

	/* LIFO */
	breq = &s->bcn_req_queue[s->sta_bcn_req_nr - 1];

	agent_req_beacon_metrics(a, breq->ap, s->macaddr,
			breq->opclass, breq->channel, breq->bssid,
			breq->reporting_detail,
			breq->ssid_len, breq->ssid, 0, NULL,
			breq->num_element, breq->element);

	/* Dequeue */
	memset(breq, 0, sizeof(struct sta_bcn_req));
	s->sta_bcn_req_nr--;

	if (s->sta_bcn_req_nr >= 1)
		/* requests remainig, send next in 3 sec */
		timer_set(&s->sta_bcn_req_timer, 3 * 1000);
}

static void wifi_sta_finalize(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_finalize_timer);
	struct sta *p, *tmp;
	struct netif_ap *vif = s->vif;

	list_for_each_entry_safe(p, tmp, &vif->stalist, list) {
		if (!memcmp(p->macaddr, s->macaddr, 6)) {
			if (s->legacy_steered)
				s->legacy_steered = false;

			if (!s->ref) {
				dbg("Finalize STA " MACFMT "\n",
						MAC2STR(s->macaddr));
				timer_del(&s->sta_timer);
				timer_del(&s->sta_bcn_req_timer);
				timer_del(&s->sta_steer_timer);
				list_del(&s->list);
				list_flush(&s->pref_nbrlist, struct pref_neighbor, list);
				list_flush(&s->sta_nbrlist, struct sta_neighbor, list);
				if (s->assoc_frame) {
					free(s->assoc_frame->frame);
					free(s->assoc_frame);
				}
				free(s);
			}
		}
	}
}

static int steer_sta_legacy(struct sta *s)
{
	struct netif_ap *vif = s->vif;
	uint16_t reason = 0;

	if (s->legacy_steered) {
		dbg("Skip Legacy Steering " MACFMT " again\n",
				MAC2STR(s->macaddr));
		return 0;
	}

	s->legacy_steered = true;
	dbg("Try to Legacy Steer " MACFMT " ------>\n", MAC2STR(s->macaddr));
	wifi_disconnect_sta(vif->ifname, s->macaddr, reason);
	wifiagent_log_steer(vif->agent, s->macaddr, vif->ifname,
					"steer_legacy", NULL);

	timer_set(&s->sta_finalize_timer,
			vif->cfg->steer_legacy_reassoc_secs * 1000);

	return 0;
}

/* Translate internal 'pref_neighbor' to wifi's 'nbr' struct */
void agent_pref_neighbor_to_nbr(struct pref_neighbor *pref_nbr_src,
		struct nbr *wifi_nbr_dst)
{
	memcpy(wifi_nbr_dst->bssid, pref_nbr_src->bssid, 6);
	wifi_nbr_dst->bssid_info = pref_nbr_src->bssid_info;
	wifi_nbr_dst->reg = pref_nbr_src->reg;
	wifi_nbr_dst->channel = pref_nbr_src->channel;
	wifi_nbr_dst->phy = pref_nbr_src->phy;
}

static int steer_sta(struct sta *s, struct pref_neighbor *pref_nbr)
{
	struct netif_ap *vif = s->vif;
	struct agent *a = vif->agent;
	struct pref_neighbor *pref_nbr_el;
	struct nbr wifi_nbr_el = {};
	int ret = 0;

	s->steer_secs++;
	trace("Steer STA " MACFMT " (steer_secs = %u)\n",
				MAC2STR(s->macaddr), s->steer_secs);

	if (pref_nbr) {
		pref_nbr_el = pref_nbr;
	} else {
		if (list_empty(&s->pref_nbrlist)) {
			trace("No better BSS for STA " MACFMT " found\n",
							MAC2STR(s->macaddr));
			return 0;
		}

		pref_nbr_el = list_first_entry(&s->pref_nbrlist,
						struct pref_neighbor, list);
	}

	if (!pref_nbr_el) {
		warn("Unexpected! pref_nbr_el is NULL or empty!\n");
		return -1;
	}

	/* Translate internal 'pref_neighbor' to wifi's 'nbr' struct */
	agent_pref_neighbor_to_nbr(pref_nbr_el, &wifi_nbr_el);

	/* btm steer ? */
	if (!s->steer_btm_cnt) {
		s->steer_btm_cnt++;
		info("Try {%d} to BTM Steer " MACFMT " =======>\n",
				s->steer_btm_cnt, MAC2STR(s->macaddr));

		ret = agent_req_btm_steer(vif->ifname, s->macaddr, 1, &wifi_nbr_el, 0);
		if (!ret) {
			wifiagent_log_steer(a, s->macaddr, vif->ifname,
					"steer_btmreq", wifi_nbr_el.bssid);
		} else  {
			warn("Failed to send BTM request to " MACFMT "\n",
					MAC2STR(s->macaddr));
		}

		return ret;
	}

	/* retry btm steer ? */
	if (vif->cfg->steer_btm_retry) {
		if (!((s->steer_secs - 1) % vif->cfg->steer_btm_retry_secs) &&
			(s->steer_btm_cnt < vif->cfg->steer_btm_retry + 1)) {

			s->steer_btm_cnt++;
			info("Try {%d} to BTM Steer " MACFMT " =======>\n",
					s->steer_btm_cnt, MAC2STR(s->macaddr));

			ret = agent_req_btm_steer(vif->ifname, s->macaddr,1, &wifi_nbr_el, 0);
			if (!ret)
				wifiagent_log_steer(a, s->macaddr, vif->ifname,
						"steer_btmreq", wifi_nbr_el.bssid);
			else
				warn("Failed to send BTM request to " MACFMT "\n",
						MAC2STR(s->macaddr));

			return ret;
		}
	}

	/* fallback to legacy steer ? */
	if (vif->cfg->fallback_legacy
			&& !s->legacy_steered
			&& (s->steer_secs >= vif->cfg->fallback_legacy))
		return steer_sta_legacy(s);

	return ret;
}

static void rebuild_cntlr_preflist(struct agent *a, struct sta *s,
					int cnt, struct pref_neighbor *pref)
{
	int sz = sizeof(int) + cnt * sizeof(*pref);

	if (s->cntlr_preflist)
		free(s->cntlr_preflist);

	s->cntlr_preflist = malloc(sz);
	if (s->cntlr_preflist) {
		memset(s->cntlr_preflist, 0, sz);
		s->cntlr_preflist->num = cnt;
		memcpy(s->cntlr_preflist->pref, pref, cnt);
	} else
		warn("OOM: cntlr_preflist\n");
}

int calculate_steer_verdict(struct sta *s, struct pref_neighbor **nbr)
{
	struct steer_rule *r;
	steer_verdict_t v = STEER_SKIP;

	list_for_each_entry(r, &regd_steer_rules, list) {
		if (!r->enabled)
			continue;

		if (r->check) {
			loud("STA " MACFMT " Check rule '%s'\n",
					MAC2STR(s->macaddr), r->name);
			v = r->check(r, s, nbr);
		}
		if (v != STEER_SKIP)
			return v;
	}

	return v;
}

/* returns 0 = don't steer, = 1 for steer */
static int should_steer_sta(struct sta *s, unsigned int *res)
{
	*res = 0;

	/* TODO:
	 * Check sta stats viz. active data sessions, retransmits etc. to
	 * decide if should steer sta.
	 * If verdict is OK, steer to cntlr_preflist.
	 */

	return 0;
}

/* returns 0 = don't steer, >= 1 for steer */
static int maybe_steer_sta(struct sta *s, unsigned int *res,
					struct pref_neighbor **nbr)
{
	struct netif_ap *vif = s->vif;
	struct agent *a = vif->agent;
	struct stax *x;

	/* dbg("%s: vif = %p  cfg = %p ifname = %s\n",
	 *	__func__, vif, cfg, vif->ifname);
	 */

	/* TODO: don't check here ..
	 * flag exclude whenever cfg is updated to exclude a STA.
	 */
	list_for_each_entry(x, &a->cfg.pcfg->steer_excludelist, list) {
		unsigned char xmac[6] = {0};
		int m_len;

		hwaddr_aton(x->macstring, xmac);
		m_len = strlen(x->macstring) == 17 ? 6 : strlen(x->macstring)/3;
		if (!memcmp(s->macaddr, xmac, m_len)) {
			loud("STA " MACFMT " in exclude list. Do nothing\n",
					MAC2STR(s->macaddr));
			return 0;
		}
	}

	/* Following if local steering is NOT disallowed */
	if (calculate_steer_verdict(s, nbr) == STEER_OK)
		return 1;

	return 0;
}

static void wifi_sta_steer_timeout(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_steer_timer);
	unsigned int reason;

	if (timestamp_expired(&s->steer_opportunity, s->steer_opportunity_tmo)) {
		/* close steer opportunity window */
		s->steer_policy &= ~STA_STEER_OPPORTUNITY;
		return;
	}

	if (should_steer_sta(s, &reason)) {
		steer_sta(s, NULL);
		/* log_steer_action("Cntlr", s->macaddr, reason, ret); */
	}

	/* check for steer opportunity again in 1 sec */
	timer_set(&s->sta_steer_timer, 1000);
}

static void wifi_sta_steered_deauth(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_steer_deauth_timer);
	struct netif_ap *vif = s->vif;

	warn("%s: Disconnect station: " MACFMT " from interface: %s with reson:%i\n",
		__func__, MAC2STR(s->macaddr), vif->ifname, WIFI_REASON_BSS_TRANSITION_DISASSOC);
	wifi_disconnect_sta(vif->ifname, s->macaddr, WIFI_REASON_BSS_TRANSITION_DISASSOC);
}

static int cond_refresh_sta_neighbor_list(struct agent *a, struct sta *s)
{
	struct netif_ap *vif = s->vif;
#define STA_NBR_REFRESH_BAD_RSSI	-78
	char ev[256] = {0};

	if (s->rssi[0] > STA_NBR_REFRESH_BAD_RSSI) {
		s->inform_leaving = 0;
		return 0;
	}

#if 0
	s->sta_nbr_invalid %= STA_NBR_REFRESH_CNT * a->cfg.runfreq;
	if (!s->sta_nbr_invalid)
		refresh_sta_neighbor_list(vif, s);

	s->sta_nbr_invalid++;
#endif
#if 0
	if (s->supports_bcnreport && s->sta_nbr_nr == 0)
		refresh_sta_neighbor_list(vif, s);
#endif

	if (!(s->inform_leaving++ % 5)) {
		/* tell cntlr about this bad rssi */
		snprintf(ev, sizeof(ev),
			"{\"macaddr\":\""MACFMT"\""
			",\"vif\":\"%s\""
			",\"action\":\"monitor\""
			",\"lowrssi\":%d}",
			MAC2STR(s->macaddr), vif->ifname, s->rssi[0]);

		agent_notify_event(a, "map.agent", ev);
		/* s->inform_leaving = (s->inform_leaving + 1) % 3; */
		s->wait_for_cntlr_nbr = true;

		/* If no pref_nbr reply from cntlr, and
		 * sta->supports_bcnreport = false, then do not steer.
		 * Either of these is true, then steer based on other
		 * criteria.
		 */
	}

	return 0;
}

#define STA_STALE_STATS_CNT_MAX 60
static void wifi_sta_periodic_run(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_timer);
	struct pref_neighbor *pref_nbr = NULL;
	struct netif_ap *vif = s->vif;
	struct wifi_sta sta = {};
	unsigned int reason;
	struct agent *a;
	int ret;

	if (!vif)
		return;

	a = vif->agent;

	/* if (strcmp(vif->ifname, s->vif_name))
	 *	err(stderr, "%s: vif changed under the hood!\n", __func__);
	 */
	if (!vif->cfg || !vif->cfg->enabled)
		return;

	//trace("%s: STA = " MACFMT " ref = %d\n", __func__,
	//					MAC2STR(s->macaddr), s->ref);

	if (s->legacy_steered && s->ref == 2)
		goto rearm_periodic;

	ret = wifi_get_station(vif->ifname, s->macaddr, &sta);
	if (ret)
		goto rearm_periodic;

	update_sta_entry(vif, &sta);
	cond_refresh_sta_neighbor_list(a, s);

	if (s->stale_stats_cnt > STA_STALE_STATS_CNT_MAX) {
		dbg("Stale STA " MACFMT ", sending disconnect.\n",
		    MAC2STR(s->macaddr));

		/* Inform driver about missing STA */
		wifi_disconnect_sta(vif->ifname, s->macaddr,
				WIFI_REASON_DISASSOC_STA_HAS_LEFT);

		/* Clean up STA in map */
		wifi_topology_notification(a, s->macaddr, vif->ifname,
				    0x00 /* left */,
				    0x01 /* unspecified reason */);
		wifi_del_sta(a, vif->ifname, s->macaddr);
		return;
	}

	if (sta_steer_allowed(s->steer_policy)
			&& !list_empty(&s->pref_nbrlist)
			&& maybe_steer_sta(s, &reason, &pref_nbr)) {

		steer_sta(s, pref_nbr);
	}

	if (!list_empty(&vif->nbrlist) && list_empty(&s->pref_nbrlist))
		recalc_desired_neighbors(s);

rearm_periodic:
	timer_set(&s->sta_timer, STA_PERIODIC_RUN_INTERVAL);
}

static int wifi_send_sta_report(struct agent *a, const char *vif,
		uint8_t *macaddr, uint32_t status, uint8_t *bssid)
{
	uint8_t src_bssid[6] = { 0 };
	uint8_t origin[6] = { 0 };
	uint8_t ret = 0;
	struct netif_ap *ifptr;
	bool ifready = false;

	list_for_each_entry(ifptr, &a->aplist, list) {
		if (!strcmp(ifptr->ifname, vif)) {
			ifready = true;
			memcpy(src_bssid, ifptr->bssid, 6);
			break;
		}
	}

	if (!ifready)
		return -1;

	memcpy(origin, a->cntlr_almac, 6);

	/* Here we get need to send the steering report */
	ret = send_steer_btm_report(a, origin, bssid,
			src_bssid, macaddr, status);

	return ret;
}

static int wifi_add_sta(struct agent *a, const char *vif,
		unsigned char *macaddr)
{
	struct netif_ap *ifptr;
	//struct netif_apcfg *cfg;
	struct sta *sptr, *new;
	struct wifi_assoc_frame *af;
	bool ifready = false;

	list_for_each_entry(ifptr, &a->aplist, list) {
		if (!strcmp(ifptr->ifname, vif)) {
			ifready = true;
			break;
		}
	}
	if (!ifready)
		return -1;

#if 0
{
	/* [NOTE] Ext/Rept are now configured in the Disallow Steer List.
	 * There should be no restriction in steering a WiFi Rept (because it
	 * is a normal STA afterall to the upstream AP), but steering it needs
	 * special care. It has cascading effect on the downstream clients.
	 * If a Rept becomes steering candidate, it means it is not hearing
	 * well and/or performing poorly. Such events should be recorded.
	 */
	struct stax *x;

	cfg = ifptr->cfg;
	if (!cfg) {
		warn("%s: NULL config!\n", vif);
		return -1;
	}

//	TODO: test more thorouhgly and comment back in
//	list_for_each_entry(x, &a->cfg.steer_excludelist, list) {
//		unsigned char xmac[6] = {0};
//		int m_len;
//
//		hwaddr_aton(x->macstring, xmac);
//		m_len = strlen(x->macstring) == 17 ? 6 : strlen(x->macstring)/3;
//		if (!memcmp(macaddr, xmac, m_len)) {
//			dbg("Ignore STA " MACFMT " in steer exclude list\n",
//					MAC2STR(macaddr));
//			return 0;
//		}
//	}
}
#endif
	list_for_each_entry(sptr, &ifptr->stalist, list) {
		if (!memcmp(sptr->macaddr, macaddr, 6)) {
			dbg("STA " MACFMT "already in list.\n",
					MAC2STR(macaddr));
			if (sptr->legacy_steered) {
				timer_set(&sptr->sta_finalize_timer,
						ifptr->cfg->steer_legacy_retry_secs * 1000);
			}
			/* In case station is already on stalist, cancel steered station
			 * deauthentication timer if it is stil runnning.
			 */
			if (timer_pending(&sptr->sta_steer_deauth_timer))
				timer_del(&sptr->sta_steer_deauth_timer);
			return 0;
		}
	}

	/* add new sta */
	new = malloc(sizeof(struct sta));
	if (!new) {
		warn("OOM: new sta malloc failed!\n");
		return -1;
	}
	memset(new, 0, sizeof(struct sta));
	memcpy(new->macaddr, macaddr, 6);
	new->vif = ifptr;
	timer_init(&new->sta_timer, wifi_sta_periodic_run);
	timer_init(&new->sta_bcn_req_timer, wifi_sta_bcn_req);
	timer_init(&new->sta_steer_timer, wifi_sta_steer_timeout);
	timer_init(&new->sta_finalize_timer, wifi_sta_finalize);
	timer_init(&new->sta_steer_deauth_timer, wifi_sta_steered_deauth);
	timestamp_update(&new->last_update);
	INIT_LIST_HEAD(&new->sta_nbrlist);
	INIT_LIST_HEAD(&new->pref_nbrlist);
	recalc_desired_neighbors(new);
#if 0
	refresh_sta_neighbor_list(ifptr, new);
#endif
	new->steer_policy |= STA_STEER_ALLOWED;
	dbg("STA steer policy = 0x%x\n", new->steer_policy);
	af = wifi_get_frame(a, new->macaddr);
	if (af) {
		new->assoc_frame = af;
		list_del(&af->list);
	}

	ifptr->nbr_sta++;

	/* INIT_LIST_HEAD(&new->cntlr_preflist); */
	list_add(&new->list, &ifptr->stalist);

	timer_set(&new->sta_timer, 1 * 1000);
	return 0;
}

static int wifi_del_sta(struct agent *a, const char *vif,
		unsigned char *macaddr)
{
	struct netif_ap *ifptr;
	struct sta *s, *tmp;
	bool ifready = false;

	list_for_each_entry(ifptr, &a->aplist, list) {
		if (!strcmp(ifptr->ifname, vif)) {
			ifready = true;
			break;
		}
	}
	if (!ifready)
		return -1;

	list_for_each_entry_safe(s, tmp, &ifptr->stalist, list) {
		if (!memcmp(s->macaddr, macaddr, 6)) {
			/* TODO: defer delete if in periodic_run..lock? */

			dbg("Delete STA " MACFMT "record\n",
						MAC2STR(macaddr));
			timer_del(&s->sta_finalize_timer);
			timer_del(&s->sta_steer_timer);
			timer_del(&s->sta_bcn_req_timer);
			timer_del(&s->sta_timer);
			timer_del(&s->sta_steer_deauth_timer);
			list_del(&s->list);
			list_flush(&s->pref_nbrlist, struct pref_neighbor, list);
			list_flush(&s->sta_nbrlist, struct sta_neighbor, list);
			if (s->assoc_frame) {
				free(s->assoc_frame->frame);
				free(s->assoc_frame);
			}
			free(s);
			return 0;
		}
	}
	ifptr->nbr_sta--;
	return -2;
}

int wifi_topology_notification(struct agent *agent, uint8_t *mac, char *ifname, uint8_t assoc_event, uint16_t reason)
{
	trace("%s: --->\n", __func__);
	struct agent *a = (struct agent *) agent;
	struct cmdu_buff *cmdu;
	struct netif_ap *ap;

	ap = get_netif_by_name(a, ifname);
	if (!ap)
		return -1;

	if (!assoc_event) {
		cmdu = agent_gen_client_disassoc(a, mac, ap->bssid, reason);
		if (cmdu) {
			agent_send_cmdu(a, cmdu);
			cmdu_free(cmdu);
		}
	}

	cmdu = agent_gen_topology_notification(a, mac, ap->bssid, assoc_event);
	if (!cmdu)
		return -1;

	agent_send_cmdu(a, cmdu);
	cmdu_free(cmdu);

	return 0;
}

#if 0	// unused
static char *arp_get_ip_by_mac(const char *mac, char *ip)
{
	FILE *arpt;
	char line[256] = {0};
	char macaddr[24] = {0};
	char ipaddr[24] = {0};
	char device[32] = {0};
	char mask[256] = {0};
	int flag, hw;
	unsigned char sta[6];

	if (!hwaddr_aton(mac, sta))
		return NULL;

	arpt = fopen("/proc/net/arp", "r");
	if (!arpt)
		return NULL;

	while (fgets(line, sizeof(line), arpt) != NULL) {
		int rc;
		unsigned char hwaddr[6];

		rc = sscanf(line, "%23s 0x%x 0x%x %23s %255s %31s",
				ipaddr, &hw, &flag, macaddr, mask, device);
		if (rc != 6)
			continue;

		if (!hwaddr_aton(macaddr, hwaddr))
			continue;

		if (memcmp(sta, hwaddr, 6)) {
			if (!memcmp(sta + 3, hwaddr + 3, 3)) {
				dbg("Arp flush: found (%s) was NOT a match," \
					"however mac was a masked match (%s)," \
					"nothing to flush?\n",
					mac, macaddr);
			}
			continue;
		}
		strncpy(ip, ipaddr, 16);
		break;
	}

	fclose(arpt);

	return ip;
}

static int arp_flush_entry(const char *macaddr)
{
	struct arpreq ar = {0};
	int sd;
	int ret = 0;
	char ipstr[16] = {0};
	struct sockaddr_in *in = (struct sockaddr_in *)&ar.arp_pa;

	dbg("Arp flush: ---- Setting up for arp entry flush mac = %s ----\n",
								macaddr);

	if (!arp_get_ip_by_mac(macaddr, ipstr)) {
		dbg("Arp flush: Flush attempt FAILED, " \
			"no match was found for mac = %s\n", macaddr);
		goto out;
	}

	inet_pton(AF_INET, ipstr, &in->sin_addr.s_addr);
	in->sin_family = AF_INET;
	sd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
	ret = ioctl(sd, SIOCDARP, &ar);
	if (ret) {
		dbg("Arp flush: Failed to flush STA %s (%s) from arp table, " \
				"return value = %d\n", macaddr, ipstr, ret);
	} else {
		dbg("Arp flush: Flushed STA: %s (%s), from ARP table\n",
				macaddr, ipstr);
	}
	close(sd);
out:
	dbg("Arp flush: ---- Flush attempt done ----\n");
	return ret;
}
#endif

static void wifi_chan_change_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = c;
	struct cmdu_buff *cmdu;
	static const struct blobmsg_policy ev_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[ARRAY_SIZE(ev_attr)];
	struct wifi_radio_element *radio;
	char *ifname, *event;
	char *chan = NULL;
	char *bw = NULL;

	blobmsg_parse(ev_attr, ARRAY_SIZE(ev_attr), tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	ifname = blobmsg_data(tb[0]);
	event = blobmsg_data(tb[1]);

	radio = wifi_ifname_to_radio_element(a, ifname);
	if (!radio)
		return;

	if (!strcmp(event, "ap-chan-change")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "target-channel", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "target-width", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "reason", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !data[2]) {
			WARN_ON(1);
			return;
		}

		chan = blobmsg_data(data[0]);
		bw = blobmsg_data(data[1]);

		if (!strcmp(blobmsg_data(data[2]), "radar") ||
		    !strcmp(blobmsg_data(data[2]), "move-start"))
			timer_set(&radio->preference_report_timer, 1000);

	} else if (!strcmp(event, "csa-finished")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "channel", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "bandwidth", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !data[2]) {
			WARN_ON(1);
			return;
		}

		chan = blobmsg_data(data[0]);
		bw = blobmsg_data(data[1]);
	}

	if (!chan || !bw)
		return;

	if (!atoi(chan) || !atoi(bw))
		return;

	if (radio->reported_channel == atoi(chan) &&
	    radio->reported_bandwidth == atoi(bw))
		return;

	radio->current_channel = atoi(chan);
	radio->current_bandwidth = atoi(bw);
	radio->current_opclass = wifi_opclass_find_id_from_channel(&radio->opclass,
			radio->current_channel, radio->current_bandwidth);

	/* Finally send this to controller */
	dbg("[%s] oper_channel_response chan %d bw %d opclass %d\n",
	    ifname, radio->current_channel, radio->current_bandwidth,
	    radio->current_opclass);

	cmdu = agent_gen_oper_channel_response(a, radio, radio->current_channel,
					       radio->current_bandwidth, 0);
	if (WARN_ON(!cmdu))
		return;

	agent_send_cmdu(a, cmdu);
	cmdu_free(cmdu);

	radio->reported_channel = radio->current_channel;
	radio->reported_bandwidth = radio->current_bandwidth;
}

static void wifi_cac_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = c;
	static const struct blobmsg_policy ev_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[ARRAY_SIZE(ev_attr)];
	struct wifi_radio_element *radio;
	char *ifname, *event;

	if (!a->cfg.ap_follow_sta_dfs)
		return;

	blobmsg_parse(ev_attr, ARRAY_SIZE(ev_attr), tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	ifname = blobmsg_data(tb[0]);
	event = blobmsg_data(tb[1]);

	radio = wifi_ifname_to_radio_element(a, ifname);
	if (!radio)
		return;

	if (!strcmp(event, "cac-end")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "success", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		if (strcmp(blobmsg_data(data[0]), "1"))
			return;

		timer_set(&a->dynbh_scheduler, 1 * 1000);
	}
}

bool is_channel_supported_by_radio(struct wifi_radio_element *re,
				   uint8_t opclass,
				   uint8_t channel)
{
	return wifi_opclass_id_channel_supported(&re->opclass, opclass, channel);
}

struct netif_bk *agent_get_netif_bk_with_channel(struct agent *a,
						 uint8_t opclass,
						 uint8_t channel)
{
	int i;
	uint8_t priority = -1;
	struct netif_bk *best = NULL;

	for (i = 0; i < a->num_radios; i++) {
		struct wifi_radio_element *r = &a->radios[i];
		struct netif_bk *bk;

		if (!is_channel_supported_by_radio(r, opclass, channel))
			continue;

		bk = agent_get_netif_bk_by_device(a, r->name);
		if (!bk)
			continue;

		if (best && bk->cfg->priority < priority)
			continue;

		best = bk;
		priority = bk->cfg->priority;
	}

	return best;
}

static int wifi_parse_frame(struct agent *a, struct json_object *frameobj)
{
	struct wifi_assoc_frame *new, *old;
	struct sta *s;
	const char *framestr, *macaddr;
	int len;
	uint8_t *frame;

	trace("|%s:%d| parsing wifi frame\n", __func__, __LINE__);

	macaddr = json_get_string(frameobj, "macaddr");
	if (!macaddr)
		return -1;

	framestr = json_get_string(frameobj, "raw");
	if (!framestr)
		return -1;

	len = strlen(framestr);
	len = len / 2;
	frame = calloc(len, sizeof(uint8_t));
	if (!frame)
		return -1;

	if (!strtob((char *)framestr, len, frame))
		goto out_frame;

	new = calloc(1, sizeof(*new));
	if (!new)
		goto out_frame;

	new->frame = frame;
	hwaddr_aton(macaddr, new->macaddr);
	new->len = len;

	/* if frame for client exists in list, replace it */
	old = wifi_get_frame(a, new->macaddr);
	if (old) {
		list_del(&old->list);
		free(old->frame);
		free(old);
	}

	/* if sta exists attach to sta directly and exit*/
	s = find_sta_by_mac(a, new->macaddr);
	if (s) {
		if (s->assoc_frame) {
			free(s->assoc_frame->frame);
			free(s->assoc_frame);
		}
		s->assoc_frame = new;
		return 0;
	}

	list_add(&new->list, &a->framelist);

	return 0;
out_frame:
	free(frame);
	return -1;
}

static int get_frame_type(struct agent *a, struct json_object *frameobj)
{
	const char *framestr;
	int frame_type = -1;

	framestr = json_object_to_json_string(frameobj);
	if (!framestr)
		return -1;

	if (strstr(framestr, "reassoc"))
		frame_type = WIFI_FRAME_REASSOC_REQ;
	else if (strstr(framestr, "assoc"))
		frame_type = WIFI_FRAME_ASSOC_REQ;
	else if (strstr(framestr, "action"))
		frame_type = WIFI_FRAME_ACTION;

	return frame_type;
}

static void check_protocol(const char *framestr, uint8_t *protocol)
{
	uint8_t *frame;
	uint8_t category;
	uint8_t action;
	int offset = 0;
	int len;

	len = strlen(framestr);
	len = (len - 1)/2;
	frame = calloc(len, sizeof(uint8_t));
	if (!frame)
		return;

	if (!strtob((char *)framestr, len, frame))
		goto out;

	if (len < 26)
		goto out;

	/* 2: frame control,
	 * 2: duration,
	 * 6: destination addr,
	 * 6: source addr,
	 * 6: bssid,
	 * 2: seq control
	 */
	offset = 2 + 2 + 6 + 6 + 6 + 2;
	memcpy(&category, frame + offset, 1);
	memcpy(&action, frame + offset + 1, 1);

	if (category == 0x0a) {
		if (action == 0x06)
			*protocol = 0x02;
		else if (action == 0x1a)
			*protocol = 0x03;
	}

	/* TODO:check the action frame category &
	 * type for ANQP.
	 */

out:
	free(frame);
}

static int wifi_parse_all_frame(struct agent *a, const char *ifname,
		struct json_object *frameobj)
{
	trace("%s\n", __func__);
	const char *framestr;
	int stype;
	uint8_t protocol = 0xff;

	stype = get_frame_type(a, frameobj);
	if (stype == -1)
		return -1;

	switch (stype) {
	case WIFI_FRAME_REASSOC_REQ:
		protocol = 0x01;
		framestr = json_get_string(frameobj, "reassoc");
		break;
	case WIFI_FRAME_ASSOC_REQ:
		protocol = 0x00;
		framestr = json_get_string(frameobj, "assoc");
		break;
	case WIFI_FRAME_ACTION:
		framestr = json_get_string(frameobj, "action");
		check_protocol(framestr, &protocol);
		break;
	default:
		framestr = NULL;
		break;
	}

	if (!framestr)
		return -1;

	prepare_tunneled_message((void *)a, ifname, protocol, framestr);

	return 0;
}

static int wifi_process_action_frame(struct agent *a, const char *framestr)
{
	trace("%s\n", __func__);

	int len;
	uint8_t *frame;

	if (!framestr)
		return -1;

	len = strlen(framestr);
	len = len / 2; /* octets */
	frame = calloc(len, sizeof(uint8_t));
	if (!frame)
		return -1;

	if (!strtob((char *)framestr, len, frame))
		goto out_frame;

	if (frame) {
		struct action_frame_container *frame_ptr;
		uint16_t head_len = sizeof(struct action_frame_container);
		uint16_t tags_len = len - head_len;
		uint16_t pos = 0;
		int count = 0;
		struct sta *s;

		frame_ptr = (struct action_frame_container *) frame;

		/* Count action frame elements (Measurement Reports) */
		while (pos < tags_len) {
			if (frame_ptr->tags[pos] != 0x27) /* EID 39 - Meas Report */
				warn("|%s:%d| action frame contains unsupported" \
					 "Element ID\n", __func__, __LINE__);
			/* +2 for tag_no & tag_len */
			pos += (frame_ptr->tags[pos+1] + 2);
			count++;
		}

		s = find_sta_by_mac(a, frame_ptr->src);
		if (!s)
			goto out_frame;

		if (!s->supports_bcnreport)
			s->supports_bcnreport = true;

		trace("|%s:%d| sending beacon metrics response to the cntlr\n",
				__func__, __LINE__);

		/* TODO: consider adding some time constrains on the metrics
		 * received, so that we don't use outdated data. Current
		 * implementation seems to  work with accordance to
		 * specification (10.3.3 & 17.2.28) though. This is the
		 * good place to add constraints in case of future issues.
		 * Please check issue 5455 for more details.
		 */

		send_beacon_metrics_response(a, frame_ptr->src,
				count, frame_ptr->tags, tags_len);

		free(frame);
	}

	return 0;
out_frame:
	free(frame);
	return -1;
}

static int wifi_process_rx_frame(struct agent *a, struct json_object *frameobj)
{
	trace("%s\n", __func__);
	const char *framestr = NULL;
	int stype;

	stype = get_frame_type(a, frameobj);
	if (stype == -1)
		return -1;

	if (stype == WIFI_FRAME_ACTION) {
		framestr = json_get_string(frameobj, "action");
		if (framestr) {
			dbg("|%s:%d| processing action frame %s\n",
					__func__, __LINE__, framestr);
			return wifi_process_action_frame(a, framestr);
		}
	}

	return -1;
}

static uint16_t wifi_process_pvid_assoc_frame(struct agent *a, uint8_t *frame,
					    int len)
{
	int i;
	uint16_t pvid = 0;

	for (i = 0; i < len; i++) {
		uint8_t *p = frame + i;
		uint8_t elem_len = 0;
		uint8_t wfa_oui[4] = {0x50, 0x6f, 0x9a, 0x1b};

		if (*p++ != 0xdd) /* IEEE 802.1 vendor specific element */
			continue;

		elem_len = *p;

		if (elem_len < 14) { /* At least 14 bytes long */
			i += elem_len;
			continue;
		}

		if (elem_len > len - (i + 2)) {
			dbg("|%s:%d| WFA Element len exceeds raw frame len!\n", __func__, __LINE__);
			break;
		}


		p++; /* len */
		if (memcmp(p, wfa_oui, 4)) {
			i += elem_len;
			continue;
		}

		p += 4; /* wfa oui */

		if (*p++ != 0x06)  { /* multi-ap ext */
			i += elem_len;
			continue;
		}

		p += *p; /* subelem len */
		p++; /* len */


		if (*p++ != 0x07) { /* multi-ap profile */
			i += elem_len;
			continue;
		}

		p += *p; /* subelem len */
		p++; /* len */

		if (*p++ != 0x08) { /* multi-ap def 8021q */
			i += elem_len;
			continue;
		}

		p++; /* len */
		memcpy(&pvid, p, 2);
		dbg("|%s:%d| Found 8021q PVID %d\n", __func__, __LINE__, pvid);
		break;
	}

	return pvid;
}

#define ASSOC_TIMEOUT	60
static void wifi_sta_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = (struct agent *)c;
	char ifname[16] = {0}, event[24] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add = false, del = false;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
		struct blob_attr *data[1];
		static const struct blobmsg_policy data_attr[1] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0};
		uint8_t mac[6] = {0};

		blobmsg_parse(data_attr, 1, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, mac))
			return;

#ifndef UNAUTHORIZED_STA_IN_ASSOCLIST
		if (add) {
			wifi_topology_notification(a, mac, ifname, 0x01 /* joined */,
						   0x00 /* success */);
			wifi_add_sta(a, ifname, mac);
		} else if (del) {
			wifi_topology_notification(a, mac, ifname, 0x00 /* left */,
						   0x01 /* unspecified reason */);
			wifi_del_sta(a, ifname, mac);
		}
#endif
	} else if (!strcmp(event, "deauth")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "reason", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];
		uint8_t mac[6] = {};
		uint16_t reason = 1;
		struct agent_config *cfg = &a->cfg;
		struct policy_cfg *pcfg;

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !cfg)
			return;

		pcfg = cfg->pcfg;
		if (!pcfg)
			return;

		if (!hwaddr_aton(blobmsg_data(data[0]), mac))
			return;

		reason = blobmsg_get_u32(data[1]);

		/* Send failed connection message if the map-agent
		 * has sent fewer than the maximum number of Send Failed
		 * Connection messages (as specified in the
		 * Maximum Reporting Rate element of the Unsuccessful
		 * Association Policy TLV) in the preceding minute
		 */
		if (pcfg->report_sta_assocfails) {
			if (timestamp_expired(&a->last_unassoc,
						ASSOC_TIMEOUT * 1000)) {

				timestamp_update(&a->last_unassoc);
				a->unassoc_cnt = 1;
				send_failed_connection_msg(a, mac, 0, reason);
			} else {
				if (a->unassoc_cnt < pcfg->report_sta_assocfails_rate) {
					a->unassoc_cnt++;
					send_failed_connection_msg(a, mac, 0, reason);
				}
			}
		}
#ifdef UNAUTHORIZED_STA_IN_ASSOCLIST
		wifi_topology_notification(a, mac, ifname, 0, reason);
		wifi_del_sta(a, ifname, mac);
#endif
	} else if (!strcmp(event, "btm-resp")) {
		struct blob_attr *data[3];
		static const struct blobmsg_policy data_attr[3] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "target_bssid", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0}, bssid_str[18] = {0};
		uint8_t mac[6] = {0}, bssid[6] = {0};
		int status;
		char ev_data[512] = {0};

		blobmsg_parse(data_attr, 3, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[2])
			return;

		/* macaddr */
		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);
		if (!hwaddr_aton(mac_str, mac))
			return;

		/* target bssid - may be empty in case of failure */
		if (data[1]) {
			strncpy(bssid_str, blobmsg_data(data[1]),
					sizeof(bssid_str) - 1);
			if (!hwaddr_aton(bssid_str, bssid))
				return;
		}

		/* status */
		status = atoi(blobmsg_data(data[2]));

		/* notify event received */
		snprintf(ev_data, sizeof(ev_data),
				"{\"macaddr\":\""MACFMT"\""
				",\"dst_bssid\":\""MACFMT"\"}"
				",\"status\":%d}",
				MAC2STR(mac),
				MAC2STR(bssid),
				status);

		agent_notify_iface_event(a, ifname, "btm-resp", ev_data);

		wifi_send_sta_report(a, ifname, mac, status, bssid);

		trace("%s: btm-resp for "MACFMT" status %d\n",
			  __func__, MAC2STR(mac), status);

		/* Generally after successful steering station is disassociated from source BSSID,
		 * but sometimes it happens that station is not disassociated from source BSSID
		 * and is associated to targed BSSID at the same time. To avoid that situation
		 * we start timer here to disconnect wifi station when it left on source BSSID.
		 * When driver removes station after successful steer then timer will never end
		 * because it will be deleted together with station, but when timer hits timeout
		 * then it means station was not disconnected and funtion wifi_sta_steered_deauth
		 * will do this.
		 */
		if (!status) {
			struct netif_ap *p = NULL;
			struct sta *s = NULL;

			list_for_each_entry(p, &a->aplist, list) {
				if (!strcmp(p->ifname, ifname)) {
					break;
				}
			}

			if (!p) {
				warn("%s: Unable to find iface %s\n", __func__, ifname);
				return;
			}

			list_for_each_entry(s, &p->stalist, list) {
				if (!memcmp(s->macaddr, mac, 6)) {
					timer_set(&s->sta_steer_deauth_timer, STA_STEERED_DEAUTH_INTERVAL * 1000);
					break;
				}
			}
		} else {
			/* TODO:
			 * update reject counter and retry steer later
			 */
		}
	} else if (!strcmp(event, "bcn-report")) {
		struct blob_attr *data[4];
		static const struct blobmsg_policy data_attr[4] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "token", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "mode", .type = BLOBMSG_TYPE_STRING },
			[3] = { .name = "nbr", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0};
		uint8_t sta_addr[6] = {0};
		struct sta *s;
		uint8_t mode = 0;

		blobmsg_parse(data_attr, 4, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !data[2] || !data[3])
			return;

		mode = atoi(blobmsg_data(data[2]));

		if (mode) /* Measurement Report Mode should be 0x00 */
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, sta_addr))
			return;

		s = find_sta_by_mac(a, sta_addr);
		if (!s)
			return;

		if (!s->supports_bcnreport)
			s->supports_bcnreport = true;
	} else if (!strcmp(event, "wds-station-added") ||
		   !strcmp(event, "wds-station-removed")) {
		struct blob_attr *data[1];
		static const struct blobmsg_policy data_attr[1] = {
			[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		};
		char wdsname[16] = {0};
		struct ts_context *ts = &a->ts;
		struct netif_ap *ap = NULL;
		bool add = !strcmp(event, "wds-station-added");
		uint16_t vid;

		if (a->cfg.brcm_setup)
			return;

		blobmsg_parse(data_attr, 1, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(wdsname, blobmsg_data(data[0]), sizeof(wdsname) - 1);

		if (add)
			netif_alloc_wds(a, wdsname);
		else
			netif_free_wds_by_ifname(a, wdsname);

		list_for_each_entry(ap, &a->aplist, list) {
			vid = ap->cfg->vid;
			if (!is_vid_valid(vid))
				vid = ts->primary_vid;

			if (a->cfg.guest_isolation && (vid != 1 && vid != ts->primary_vid)) {
				dbg("/lib/wifi/multiap ts isolate_wds %s %s %s\n", (add ? "add" : "del"), ap->ifname, wdsname);
				runCmd("/lib/wifi/multiap ts isolate_wds %s %s %s", (add ? "add" : "del"), ap->ifname, wdsname);
			}
		}
	} else if (!strcmp(event, "action")) { /* TODO: this event belongs under wifi.iface handler */
		struct blob_attr *data[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "raw", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0};
		uint8_t bbss_addr[6] = {0};
		struct netif_bk *bk;
		char *framestr;
		uint8_t *frame;
		int len = 0;

		blobmsg_parse(data_attr, 2, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1])
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, bbss_addr))
			return;

		bk = agent_get_netif_bk_by_name(a, ifname);
		if (!bk)
			return;

		framestr = (char *)blobmsg_data(data[1]);

		len = strlen(framestr);
		len = (len - 1) / 2;
		frame = calloc(len, sizeof(uint8_t));
		if (!frame)
			return;

		if (!strtob(framestr, len, frame)) {
			free(frame);
			return;
		}

		if (*frame == 0x09 || *frame == 0x04) {
			struct wifi_radio_element *r;
			uint8_t channel = 0;
			uint8_t opclass = 0;
			uint8_t *ptr = frame;

			ptr++;

			if (*ptr == 0x04) {
				ptr += 2;
				opclass = *ptr;

				ptr++;
				channel = *ptr;

				r = wifi_ifname_to_radio_element(a, ifname);
				if (!r) {
					free(frame);
					return;
				}

				if (is_channel_supported_by_radio(r, opclass, channel)) {
					dbg("|%s:%d| bsta %s supports new "\
					     "channel, no action required\n",
					     __func__, __LINE__, bk->ifname);
					free(frame);
					return;
				}

				bk = agent_get_netif_bk_with_channel(a, opclass, channel);
				if (bk) {
					dynbh_bsta_swap_to_link(a, bk->ifname);
					agent_config_reload(a);
				} else
					dynbh_bsta_enable_all(a);
			}
		}

		free(frame);
	} else if (!strcmp(event, "assoc") || !strcmp(event, "reassoc")) {
		struct json_object *jmsg, *data;
		char *str;
		struct blob_attr *bdata[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "raw", .type = BLOBMSG_TYPE_STRING },
		};
		char *framestr;
		uint8_t *frame;
		int len;
		uint16_t pvid;
#ifdef UNAUTHORIZED_STA_IN_ASSOCLIST
		char mac_str[18] = {0};
		uint8_t mac[6] = {0};
#endif

		blobmsg_parse(data_attr, 2, bdata, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!bdata[0] || !bdata[1])
			return;

		framestr = (char *)blobmsg_data(bdata[1]);

		len = strlen(framestr);
		len = (len - 1) / 2;
		frame = calloc(len, sizeof(uint8_t));
		if (!frame)
			return;

		if (!strtob(framestr, len, frame)) {
			free(frame);
			return;
		}

		pvid = wifi_process_pvid_assoc_frame(a, frame, len);
		if (pvid) {
			agent_fill_8021q_setting(a, pvid, 0);
			agent_apply_traffic_separation(a);
		}

		free(frame);

#ifdef UNAUTHORIZED_STA_IN_ASSOCLIST
		strncpy(mac_str, blobmsg_data(bdata[0]), sizeof(mac_str) - 1);
		hwaddr_aton(mac_str, mac);
		wifi_add_sta(a, ifname, mac);
		wifi_topology_notification(a, mac, ifname,
					   0x01 /* client joined */,
					   0x00 /* success */);
#endif
		trace("%s: ------------>\n", __func__);

		str = blobmsg_format_json(msg, true);
		if (!str)
			return;

		jmsg = json_tokener_parse(str);
		if (!jmsg)
			goto out_str;

		if (!json_object_is_type(jmsg, json_type_object))
			goto out_json;

		json_object_object_get_ex(jmsg, "data", &data);
		if (!data)
			goto out_json;

		wifi_parse_frame(a, data);

		/* parse (re-)assoc message
		*/
		wifi_parse_all_frame(a, ifname, data);

out_json:
		json_object_put(jmsg);
out_str:
		free(str);
	} else if (!strcmp(event, "assoc-failure")) {
		/* { "wifi.sta": {"ifname":"wl0","event":"assoc-failure","data":{"macaddr":"00:0a:52:06:2e:b9", "status":12}} } */
		struct blob_attr *bdata[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "status", .type = BLOBMSG_TYPE_INT32 },
		};
		char mac_str[18] = {0};
		uint8_t mac[6] = {0};
		struct agent_config *cfg = &a->cfg;
		struct policy_cfg *pcfg;
		int status, status_code = 0;

		blobmsg_parse(data_attr, 2, bdata, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!bdata[0] || !bdata[1] || !cfg)
			return;

		pcfg = cfg->pcfg;
		if (!pcfg)
			return;

		status = blobmsg_get_u32(bdata[1]);
		if (status == 12) /* mac list deny */
			status_code = 33; /* QoS AP lacks sufficient bandwidth */

		strncpy(mac_str, blobmsg_data(bdata[0]), sizeof(mac_str) - 1);
		hwaddr_aton(mac_str, mac);
		if (pcfg->report_sta_assocfails) {
			if (timestamp_expired(&a->last_unassoc,
					      ASSOC_TIMEOUT * 1000)) {
				timestamp_update(&a->last_unassoc);
				a->unassoc_cnt = 1;
				send_failed_connection_msg(a, mac, status_code, 0);
			} else {
				if (a->unassoc_cnt < pcfg->report_sta_assocfails_rate) {
					a->unassoc_cnt++;
					send_failed_connection_msg(a, mac, status_code, 0);
				}
			}
		}
		wifi_topology_notification(a, mac, ifname, 0, 0x01);
		wifi_del_sta(a, ifname, mac);
	}
}

static void agent_iface_disconnect_all_sta(const struct netif_ap *iface,
					   uint16_t reason)
{
	struct sta *sta;

	list_for_each_entry(sta, &iface->stalist, list) {
		wifi_disconnect_sta(iface->ifname, sta->macaddr, reason);
	}

}

static void agent_init_ifaces_assoc_mode(struct agent *a)
{
	const struct netif_ap *iface;

	list_for_each_entry(iface, &a->aplist, list) {
		if (iface->cfg->disallow_assoc) {
			agent_iface_disconnect_all_sta(
				iface,
				WIFI_REASON_SERVICE_NOT_AUTHORIZED_IN_THIS_LOCATION);
		}

		wifi_set_backhaul_bss_association_mode(iface->ifname,
						       iface->cfg->disallow_assoc);
	}
}

static void apply_iface_assocation_mode(struct agent *a, const char *ifname)
{
	const struct netif_ap *iface = get_netif_by_name(a, ifname);

	if (iface) {
		wifi_set_backhaul_bss_association_mode(ifname,
						       iface->cfg->disallow_assoc);
	}
}


static void wifi_iface_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = (struct agent *)c;
	const char *ifname, *event;
	struct json_object *jmsg, *data;
	char *str;

	trace("%s: ------------>\n", __func__);

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	jmsg = json_tokener_parse(str);
	if (!jmsg)
		goto out_str;

	if (!json_object_is_type(jmsg, json_type_object))
		goto out_json;

	json_object_object_get_ex(jmsg, "data", &data);
	if (!data)
		goto out_json;

	ifname = json_get_string(jmsg, "ifname");
	if (!ifname)
		goto out_json;

	event = json_get_string(jmsg, "event");
	if (!event)
		goto out_json;

	if (!strcmp(event, "assoc") || !strcmp(event, "reassoc")) {
		wifi_parse_frame(a, data);

		/* parse (re-)assoc message
		 */
		wifi_parse_all_frame(a, ifname, data);
	} else if (!strcmp(event, "frame-rx")) {
		wifi_process_rx_frame(a, data);
	} else if (!strcmp(event, "action")) {
		/* parse action frame
		 * BTM, WNM, ANQP
		 */
		wifi_parse_all_frame(a, ifname, data);
	} else if (!strcmp(event, "bss-up")) {
		apply_iface_assocation_mode(a, ifname);
	}

out_json:
	json_object_put(jmsg);
out_str:
	free(str);
}

static void wifi_channel_event_handler(void *c, struct blob_attr *msg)
{
	/* struct agent *a = (struct agent *)c; */
	const char *ifdev, *res;
	int ch = 0;
	struct json_object *jmsg;
	char *str;

	trace("%s: ------------>\n", __func__);
	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	jmsg = json_tokener_parse(str);
	if (!jmsg)
		goto out_str;

	if (!json_object_is_type(jmsg, json_type_object))
		goto out_json;

	ifdev = json_get_string(jmsg, "radio");
	if (!ifdev)
		goto out_json;

	ch = json_get_int(jmsg, "channel");
	UNUSED(ch);

	res = json_get_string(jmsg, "reason");
	if (!res)
		goto out_json;

	if (!strcmp(res, "radar")) {
		fprintf(stderr, "TODO handle ch change due to radar!\n");
		/* wifi_handle_channel_event(a, ifdev, ch, res); */
	}

out_json:
	json_object_put(jmsg);
out_str:
	free(str);
}

uint8_t wifi_noise_to_anpi(int n)
{
	if (!n)
		return 255;
	if (n < -110)
		return 0;
	if (n > 0)
		return 220;
	return (n + 110) * 2;
}

static bool may_enable_fhs(struct agent *a)
{
#ifdef AGENT_ISLAND_PREVENTION
	if (a->cfg.island_prevention)
		return agent_has_active_backhaul(a) ? true : false;
#endif
	return agent_has_active_backhaul(a) ? true : false;
}

static void wifi_radio_handle_survey_evt(struct agent *a, char *radio,
		struct blob_attr *tb_data)
{
	dbg("%s --->\n", __func__);

	struct blob_attr *data[3];
	static const struct blobmsg_policy supp_attrs[3] = {
		[0] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "noise", .type = BLOBMSG_TYPE_INT32 },
		[2] = { .name = "utilization", .type = BLOBMSG_TYPE_INT32 },
	};
	struct wifi_radio_element *r;
	uint32_t cur_ch;
	int dBm;
	int i;

	if (!tb_data)
		return;

	blobmsg_parse(supp_attrs, 3, data,
		blobmsg_data(tb_data), blobmsg_data_len(tb_data));

	if (!data[0] || !data[1] || !data[2]) {
		dbg("|%s:%d| Wifi event parse error.\n", __func__, __LINE__);
		return;
	}
	cur_ch = (uint8_t)blobmsg_get_u32(data[0]);

	r = wifi_radio_to_radio_element(a, radio);
	if (!r || !r->scanlist)
		return;

	for (i = 0; i < r->scanlist->num_opclass; i++) {
		struct wifi_scanres_opclass_element *op;
		int j;

		op = r->scanlist->opclass_scanlist + i;

		for (j = 0; j < op->num_channels; j++) {
			struct wifi_scanres_channel_element *ch;

			if (op->channel_scanlist[j].channel == cur_ch) {
				ch = op->channel_scanlist + j;
				dBm = (int8_t)blobmsg_get_u32(data[1]);
				ch->anpi = wifi_noise_to_anpi(dBm);
				ch->utilization = (uint8_t)blobmsg_get_u32(data[2]);
				return;
			}
		}
	}
}

static void wifi_radio_event_handler(void *c, struct blob_attr *msg)
{
	dbg("%s --->\n", __func__);

	struct agent *a = (struct agent *) c;
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[6] = {
		[0] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	char radio[18] = {0}, event[14] = {0};

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1])
		return;

	strncpy(radio, blobmsg_data(tb[1]), sizeof(radio) - 1);
	strncpy(event, blobmsg_data(tb[0]), sizeof(event) - 1);

	dbg("%s: Handling %s for %s\n", __func__, event, radio);

	if (!strcmp(event, "ap-enabled")) {
		/* TODO: do this on event-basis, meaning don't teardown
		* everything and rebuild everything
		*/
		/* completely re-init data */
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);

		if (may_enable_fhs(a))
			/* Enable fronthauls */
			timer_set(&a->enable_fhs_scheduler,
					(IFACE_TIMEOUT + 1) * 1000);

	} else if (!strcmp(event, "ap-disabled")) {
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
	} else if (!strcmp(event, "ap-updated")) {
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
	} else if (!strcmp(event, "radar")) {
		struct cmdu_buff *cmdu;

		cmdu = agent_gen_channel_preference_report(a, NULL);
		if (cmdu) {
			agent_send_cmdu(a, cmdu);
			cmdu_free(cmdu);
		}
	} else if (!strcmp(event, "scan_finished")) {
		struct wifi_netdev *ndev =
			wifi_radio_to_netdev(a, radio);

		if (WARN_ON(!ndev || !ndev->re)) {
			dbg("%s: Current radio_element not found.\n", __func__);
			return;
		}
		ndev->re->scan_state = SCAN_DONE;

		handle_wifi_radio_scan_finished(a, ndev);
	} else if (!strcmp(event, "survey")) {
		if (!tb[2])
			return;
		wifi_radio_handle_survey_evt(a, radio, tb[2]);
	} else
		dbg("%s: Unhandled wifi event %s!\n", __func__, event);

}

int wifi_mod_bridge(struct agent *a, char *ifname, char *action)
{
	int ret = -1;
	bool add;
	add = !strcmp(action, "add");

	/* add wds iface to bridge */

	dbg("|%s:%d| %s interface %s to bridge %s using libeasy API\n", __func__,
	    __LINE__, (add ? "Adding" : "Deleting"),
	    ifname, a->cfg.al_bridge);


	if (add && !if_isbridge_interface(ifname))
		ret = br_addif(a->cfg.al_bridge, ifname);
	else if(!add && if_isbridge_interface(ifname))
		ret = br_delif(a->cfg.al_bridge, ifname);

	/* explicitly disable if removing from bridge */
	if (!ret && !add) {
		int rc;

		rc = wifi_set_4addr(ifname, false);
		dbg("|%s:%d| Disabled 4addr mode for %s (rc:%d)\n", __func__,
				__LINE__, ifname, rc);
	}

	return ret;
}

static void agent_reload_local_cntlr(struct agent *a, bool on)
{
	set_value_by_string("mapcontroller", "controller",
			"enabled", on ? "1" : "0", UCI_TYPE_STRING);
	trace("Reloading mapcontroller\n");

	/* procd is too slow to send signal - manual kill prior to reload
	 * TODO: leverage pidfile
	 */
	runCmd("kill `pidof mapcontroller`");
	uci_reload_services("mapcontroller");

	/* Enable is now set to '1' in config; ensure service is started */
	if (on) {
		runCmd("/etc/init.d/mapcontroller reload");
	}
}

static void agent_enable_local_cntlr(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, cntlr_scheduler);

	dbg("|%s:%d| Attempting to start local controller\n", __func__, __LINE__);

	if (is_local_cntlr_running()) {
		warn("Skip starting local mapcontroller: already running\n");
	} else {
		agent_reload_local_cntlr(a, true);
	}
}

void agent_disable_local_cntlr(struct agent *a)
{
	dbg("|%s:%d| Disable local controller in UCI\n", __func__, __LINE__);
	agent_reload_local_cntlr(a, false);
}

static void agent_schedule_cntlr_start(struct agent *a, int secs)
{
	dbg("|%s:%d| Scheduled controller start in %d sec\n",
			__func__, __LINE__, secs);
	timer_set(&a->cntlr_scheduler, secs * 1000);
}

static void agent_rcpi_thresold_timer_cb(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent,
			rcpi_threshold_timer);
	struct netif_ap *ap;

	list_for_each_entry(ap, &a->aplist, list) {
		struct sta *s;
		struct agent_config_radio *rcfg;


		dbg("%s %d checking clients on %s\n", __func__, __LINE__, ap->ifname);

		rcfg = get_agent_config_radio(&a->cfg, ap->cfg->device);
		if (!rcfg)
			continue;

		list_for_each_entry(s, &ap->stalist, list) {
			uint8_t rcpi;

			dbg("%s %d found client "MACFMT"\n", __func__, __LINE__, MAC2STR(s->macaddr));

			rcpi = rssi_to_rcpi(s->rssi[0]);
			if (rcpi < rcfg->report_rcpi_threshold) {
				struct cmdu_buff *cmdu;
				char ev_data[512] = {0};

				snprintf(ev_data, sizeof(ev_data),
					"{\"macaddr\":\""MACFMT"\""
					",\"low_rcpi_threshold\":%d"
					",\"rcpi\":%d}",
					MAC2STR(s->macaddr),
					rcfg->report_rcpi_threshold, rcpi);

				agent_notify_iface_event(a, ap->ifname, "sta_low_rcpi", ev_data);

				cmdu = agent_gen_assoc_sta_metric_responsex(a,
					a->cntlr_almac, s, ap);
				if (cmdu) {
					agent_send_cmdu(a, cmdu);
					cmdu_free(cmdu);
				}
			}

		}

	}

/*	curr_rcpi = calculate_radio_rcpi(a, p->ifname);
	prev_rcpi = p->prev_rcpi;
	if (((prev_rcpi > p->cfg->rcpi_threshold) &&
				(curr_rcpi > p->cfg->rcpi_threshold)) ||
			((prev_rcpi < p->cfg->rcpi_threshold) &&
			 (curr_rcpi < p->cfg->rcpi_threshold)))
		goto refresh_interval;
*/
	//prepare_assoc_sta_metric_response(a, p->ifname);

//refresh_interval:
//	p->prev_rcpi = curr_rcpi;
	timer_set(&a->rcpi_threshold_timer, RCPI_THRESHOLD_TIMER);
}

static bool agent_ap_get_state(struct agent *a, struct netif_ap *ap)
{

	/* TODO: refresh cashed data */
	if (!ap)
		return -1;

	return ap->enabled ? true : false;
}

static int agent_ap_set_state(struct agent *a, struct netif_ap *ap, bool up)
{
	bool is_enabled;

	if (!ap)
		return -1;

	is_enabled = agent_ap_get_state(a, ap);

	dbg("|%s:%d| setting AP %s %s (current: %s)\n",
	    __func__, __LINE__,
	    ap->ifname, up ? "up" : "down",
	    is_enabled ? "up" : "down");

	if (is_enabled == up)
		return 0;

	if (up && ap->cfg && !ap->cfg->enabled) {
		/* start_disabled set for given ap in cfg */
		dbg("|%s:%d| will not enable ap %s - start_disabled set\n",
		    __func__, __LINE__, ap->ifname);
		return 0;
	}

	if (wifi_ap_set_state(ap->ifname, up)) {
		dbg("|%s:%d| failed to set AP %s %s\n",
		    __func__, __LINE__, ap->ifname, up ? "up" : "down");
		return -1;
	} /* check if the state has changed */ if (agent_ap_get_state(a, ap) != up) {
		dbg("|%s:%d| sheduling iface init\n", __func__, __LINE__);
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
	}

	return 0;
}

static void agent_enable_fronthauls(struct agent *a)
{
	struct netif_ap *ap;

	dbg("|%s:%d| enabling all fronthauls\n", __func__, __LINE__);

	list_for_each_entry(ap, &a->aplist, list)
		agent_ap_set_state(a, ap, true);
}

static int agent_req_btm_steer(const char *ifname, unsigned char *sta_macaddr,
		uint8_t pref_num, struct nbr *pref, uint32_t validity_int)
{
	struct wifi_btmreq req = {};

	dbg("Calling wifi.ap.%s request_btm for sta " MACFMT "\n",
			ifname, MAC2STR(sta_macaddr));

	if (pref_num)
		req.mode |= WIFI_BTMREQ_PREF_INC;

	/* Inform remaining clients of imminent disassociation */
	/* disassoc_tmo kept as 0*/
	req.mode |= WIFI_BTMREQ_DISASSOC_IMM;

	/* BTM Abridged bit set to 1 */
	req.mode |= STEER_REQUEST_BTM_ABRIDGED;

	req.validity_int = validity_int / WIFI_BEACON_PERIOD_MS;

	if (wifi_req_btm(ifname, sta_macaddr, pref_num, pref, &req)) {
		err("|%s:%d|Failed calling wifi.ap.%s request_btm for sta " MACFMT "\n",
			__func__, __LINE__, ifname, MAC2STR(sta_macaddr));
		return -1;
	}

	return 0;
}

#ifdef AGENT_ISLAND_PREVENTION
/* Time in seconds until the AP sends a Disassociation frame to the STA */
#define STA_DISCONNECT_TM	30
static int agent_req_btm_bssterm(const char *ap_name, unsigned char *sta_macaddr,
		uint32_t disassoc_tmo)
{
	struct wifi_btmreq req = {};

	dbg("Calling wifi.ap.%s request_btm for sta " MACFMT "\n",
			ap_name, MAC2STR(sta_macaddr));

	/* Inform remaining clients of imminent disassociation */
	req.mode |= WIFI_BTMREQ_DISASSOC_IMM;
	/* Inform clients about BSS termination */
	req.mode |= WIFI_BTMREQ_BSSTERM_INC;

	/* Duration - number of minutes for which the BSS is not present.
	 * Value of 0 is reserved. Value 65 535 when the BSS is terminated
	 * for a period longer than or equal to 65 535 minutes.
	 */
	req.bssterm_dur = 1; /* away (min) */
	req.disassoc_tmo = disassoc_tmo / WIFI_BEACON_PERIOD_MS;

	if (wifi_req_btm(ap_name, sta_macaddr, 0, NULL, &req)) {
		dbg("|%s:%d| failed to send wifi_req_btm\n", __func__, __LINE__);
		return -1;
	}

	return 0;
}

#define STA_DISCONNECT_TM 30
void agent_schedule_fh_disable(struct agent *a)
{
	trace("%s: --->\n", __func__);

	struct netif_ap *ap;
	struct sta *s;

	if (agent_has_active_backhaul(a))
		/* backhaul regained, don't send BTM */
		return;

	dbg("|%s:%d| will disable aps in %d seconds\n",
	    __func__, __LINE__, STA_DISCONNECT_TM + 2);

	list_for_each_entry(ap, &a->aplist, list) {
		/* Try to steer all remaining (b)STAs away */
		list_for_each_entry(s, &ap->stalist, list) {
			agent_req_btm_bssterm(ap->ifname, s->macaddr, STA_DISCONNECT_TM * 1000);
		}
	}

	/* Disconnect all STAs in in STA_DISCONNECT_TM seconds */
	timer_set(&a->sta_disconnect_timer, STA_DISCONNECT_TM * 1000);
}

static int agent_sta_disconnect(struct agent *a, struct netif_ap *ap, struct sta *s)
{
	dbg("Calling wifi.ap.%s disconnect for sta " MACFMT "\n",
	    ap->ifname, MAC2STR(s->macaddr));

	return wifi_disconnect_sta(ap->ifname, s->macaddr,
			WIFI_REASON_BSS_TRANSITION_DISASSOC);
}

static void agent_fh_disable_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, fh_disable_timer);
	struct netif_ap *ap;

	if (agent_has_active_backhaul(a))
		/* backhaul regained, don't disable FHs */
		return;

	list_for_each_entry(ap, &a->aplist, list) {
		agent_ap_set_state(a, ap, false);
	}
}

static void agent_sta_disconnnect_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, sta_disconnect_timer);
	struct netif_ap *ap;
	struct sta *s;

	if (agent_has_active_backhaul(a))
		/* backhaul regained, don't disconnect STAs */
		return;

	list_for_each_entry(ap, &a->aplist, list) {
		list_for_each_entry(s, &ap->stalist, list) {
			agent_sta_disconnect(a, ap, s);
		}
	}

	/* Disable all fronthauls in 2 seconds */
	timer_set(&a->fh_disable_timer, 2 * 1000);
}
#endif /* AGENT_ISLAND_PREVENTION */

int wifiagent_toggle_fh(struct ubus_object *obj, bool isl_prev,
		char *fh_ifname, int enable)
{
	struct agent *a = container_of(obj, struct agent, obj);

#ifdef AGENT_ISLAND_PREVENTION
	if (isl_prev && a->cfg.island_prevention && !enable) {
		/* schedule bringing down all STAs & FHs */
		dbg("|%s:%d| island prevention: schedule FHs down\n",
		    __func__, __LINE__);
		agent_schedule_fh_disable(a);
		return 0;
	}
#else
	UNUSED(isl_prev);
#endif

	if (fh_ifname && fh_ifname[0]) {
		struct netif_ap *ap = NULL;

		/* enable/disable fronthaul(s)  */
		if (strncmp(fh_ifname, "all", 16)) {
			ap = get_netif_by_name(a, fh_ifname);
			if (!ap) {
				dbg("interface \'%s\' not found\n", fh_ifname);
				return -1;
			}
			agent_ap_set_state(a, ap, enable ? true : false);
		} else { /* all interfaces */
			list_for_each_entry(ap, &a->aplist, list) {
				agent_ap_set_state(a, ap, enable ? true : false);
			}
		}
	}

	return 0;
}

static void preference_report_timer_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct wifi_radio_element *re = container_of(t, struct wifi_radio_element, preference_report_timer);
	struct cmdu_buff *cmdu;

	cmdu = agent_gen_channel_preference_report(re->agent, NULL);
	if (cmdu) {
		agent_send_cmdu(re->agent, cmdu);
		cmdu_free(cmdu);
	}
}

bool agent_has_downstream(struct agent *a)
{
	struct netif_ap *ap;

	list_for_each_entry(ap, &a->aplist, list) {
		struct sta *s;

		/* only bbss and combined type interfaces may have downstream APs */
		if (ap->cfg->multi_ap != 1 && ap->cfg->multi_ap != 3)
			continue;

		list_for_each_entry(s, &ap->stalist, list)
			return true;
	}

	return false;
}

static void ethport_event_handler(void *agent, struct blob_attr *msg)
{
	char ifname[16] = {0}, link[8] = {0};
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[4];
	static const struct blobmsg_policy ev_attr[4] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "link", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "speed", .type = BLOBMSG_TYPE_TABLE },
		[3] = { .name = "duplex", .type = BLOBMSG_TYPE_TABLE },
	};
	bool up, down;

	blobmsg_parse(ev_attr, 4, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1]) /* only need ifname and link status */
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(link, blobmsg_data(tb[1]), sizeof(link) - 1);

	up = !strcmp(link, "up");
	down = !strcmp(link, "down");

	UNUSED(down);

	if (up) {
		dbg("|%s:%d| Scheduling next ACS in 1 second\n", __func__,
				__LINE__);
		timer_set(&a->autocfg_dispatcher, 1 * 1000);
		timestamp_update(&a->eth_connect_t);
	}

	return;
}


static void parse_i1905_info(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct agent *a = (struct agent *)req->priv;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ieee_attrs[2] = {
		[0] = { .name = "ieee1905id", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "interface", .type = BLOBMSG_TYPE_ARRAY }
	};

	blobmsg_parse(ieee_attrs, 2, tb, blob_data(msg), blob_len(msg));

	if (tb[0]) {
		char *mac;

		mac = blobmsg_get_string(tb[0]);
		hwaddr_aton(mac, a->almac);
		dbg("almac = " MACFMT "\n", MAC2STR(a->almac));
	}
}

static void uobj_add_event_handler(void *agent, struct blob_attr *msg)
{
	char path[32] = {0};
	uint32_t id = 0;
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "path", .type = BLOBMSG_TYPE_STRING }
	};

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(path, blobmsg_data(tb[1]), sizeof(path) - 1);
	id = (uint32_t) blobmsg_get_u32(tb[0]);
	dbg("|%s:%d| path = [%s] id = [%d] [%u]\n", __func__, __LINE__, path, id, id);
	if (!strncmp(path, map_plugin, strlen(map_plugin))) {
		/* TODO: how to handle failure? */
		agent_subscribe_for_cmdus(a);
	} else if (!strncmp(path, "wifi.ap.", 8)) {
		char *ifname = path + 8;
		struct netif_ap *ap;

		dbg("|%s:%d| iface = [%s]\n", __func__, __LINE__, ifname);
		ap = get_netif_by_name(a, ifname);
		if (!ap)
			return;
		ap->wifi = id;
		dbg("|%s:%d|wifi.ap.%s object added is [%d] [%u]\n", __func__, __LINE__, ap->ifname, ap->wifi, ap->wifi);
	} else if (!strncmp(path, "wifi.backhaul.", 14)) {
		char *ifname = path + 14;
		struct netif_bk *bk;

		warn("|%s:%d| iface = [%s]\n", __func__, __LINE__, ifname);
		bk = agent_get_netif_bk_by_name(a, ifname);
		if (!bk)
			return;
		bk->wifi = id;
		warn("|%s:%d|wifi.backhaul.%s object added is [%d] [%u]\n", __func__, __LINE__, bk->ifname, bk->wifi, bk->wifi);
	} else if (!strncmp(path, "ieee1905.al.", 12)) {
		uint32_t ieee1905_obj = 0;
		int ret;

		ret = ubus_lookup_id(a->ubus_ctx, "ieee1905", &ieee1905_obj);
		if (ret)
			return;

		ubus_call_object(a, ieee1905_obj, "info", parse_i1905_info, a);
	} else if (!strncmp(path, "wpa_supplicant.", 15)) {
		char *ifname = path + 15;

		/* re-init blacklist upon supplicant coming back up*/
		backhaul_blacklist_update_ifname(agent, ifname);
	}
#if (EASYMESH_VERSION > 2)
	else if (!strncmp(path, "hostapd.", 8)) {
		char *ifname = path + 8;
		struct netif_ap *ap;
		struct dpp_enrollee *e;

		/*
		 * QoS rules are volatile and cannot be stored, so they have to be
		 * reapplied after hostapd restarts
		 */
		qos_apply_dscp_rules(agent);

		if (memcmp(a->cntlr_almac, a->almac, 6))
			return;

		/* need to re-do csign etc. upon hostapd going up/down */
		/* we actually need to re-do all uris etc. */
		warn("|%s:%d| iface = [%s]\n", __func__, __LINE__, ifname);

		ap = get_netif_by_name(a, ifname);
		if (!ap)
			return;

		list_for_each_entry(e, &a->dpp_ctx.enrolleelist, list) {
			char fmt[1024] = {0};
			char encryption[64] = {0};
			char hex_key[129] = {0};
			char hex_ssid[65] = {0};
			uint16_t port = 0;
			enum wifi_band band;
			char *ifname;

			if (!ap->cfg || !(ap->cfg->multi_ap & 0x02))
				continue;

			band = ap->cfg->band;

			if (e->band != band)
				continue;

			ifname = agent_get_backhaul_hex(a, band, encryption, hex_ssid, hex_key);
			if (!ifname)
				continue;

			if (e->bootstrap_id) {
				snprintf(fmt, sizeof(fmt), "dpp del_uri_controller %d", e->bootstrap_id);
				agent_exec_platform_scripts(fmt);
				e->bootstrap_id = 0;
			}

			port = dpp_get_controller_tcp_port(a, band);

			/* TODO: may want to split into 2 different CMDs */
			snprintf(fmt, sizeof(fmt), "dpp set_bootstrap %s \"%s\" %s %s %s %d", ap->ifname, e->uri, encryption, hex_ssid, hex_key, port);
			e->bootstrap_id = agent_exec_platform_scripts(fmt);
			if (!strcmp(encryption, "dpp")) {
				snprintf(fmt, sizeof(fmt), "dpp csign_gen %s %s %s", ifname, encryption, hex_ssid);
				agent_exec_platform_scripts(fmt);
			}
		}
	}
#endif
}

void agent_bsta_steer_clear(struct netif_bk *bk)
{
	memset(bk->bsta_steer.target_bssid, 0, 6);
	memset(bk->bsta_steer.prev_bssid, 0, 6);
	bk->bsta_steer.mid = 0;
	memset(bk->bsta_steer.ul_ifname, 0, sizeof(bk->bsta_steer.ul_ifname));
	bk->bsta_steer.expired = false;
}

bool agent_bsta_steer_is_active(struct netif_bk *bk)
{
	return (bk->bsta_steer.mid != 0 && bk->bsta_steer.expired == false);
}

bool agent_bsta_steer_is_active_all(struct agent *a)
{
	struct netif_bk *bk = NULL;

	list_for_each_entry(bk, &a->bklist, list)
		if (agent_bsta_steer_is_active(bk))
			return true;

	return false;
}


int agent_bsta_steer(struct agent *a, char *ifname, uint8_t *target_bssid,
		     bool force_disassoc)
{
	char fmt[128] = {0};

	UNUSED(a);

	if (force_disassoc) {
		snprintf(fmt, sizeof(fmt), "bsta_steer_with_disassoc %s",
			 ifname);
	} else
		snprintf(fmt, sizeof(fmt), "bsta_steer %s",
			 ifname);

	if (hwaddr_is_zero(target_bssid)|| hwaddr_is_bcast(target_bssid)) {
		snprintf(fmt + strlen(fmt), sizeof(fmt) - strlen(fmt), " any");
	} else {
		snprintf(fmt + strlen(fmt), sizeof(fmt) - strlen(fmt), /* Flawfinder: ignore */
			 " " MACFMT, MAC2STR(target_bssid));
	}

	return agent_exec_platform_scripts(fmt);
}

int agent_bsta_steer_handler(struct agent *a, struct netif_bk *bk, uint8_t *bssid)
{
	uint8_t *target_bssid = bk->bsta_steer.target_bssid;
	uint8_t result = 0x01;

	if (!memcmp(bssid, target_bssid, 6) ||
	    /* if target was any BSSID, treat connection over correct
	     * bsta as success */
	    hwaddr_is_zero(target_bssid) || hwaddr_is_bcast(target_bssid)) {
		/* success */
		char fmt[64] = {0};

		result = 0x00;

		/* write to config */
		wifi_set_iface_bssid(bk, bssid);
		snprintf(fmt, sizeof(fmt), "write_bsta_config %s", bk->ifname);
		agent_exec_platform_scripts(fmt);
	}

	if (bk->bsta_steer.trigger == BK_STEER_MANDATE)
		send_backhaul_sta_steer_response(a, bk, result, bk->ifname);

	err("\n\n%s %d bssid = " MACFMT ", target = " MACFMT "\n", __func__,
	    __LINE__, MAC2STR(bssid), MAC2STR(target_bssid));

	timer_del(&bk->steer_timeout);
	agent_bsta_steer_clear(bk);
	return 0;
}

static bool wifi_is_band_onboarded(struct agent *a, enum wifi_band band)
{
	struct agent_config_radio *rcfg = NULL;

	list_for_each_entry(rcfg, &a->cfg.radiolist, list) {
		if (rcfg->onboarded && rcfg->band == band)
			return true;
#if (EASYMESH_VERSION > 2)
		if (rcfg->dpp_onboarded && rcfg->band == band)
			return true;
#endif
	}

	return false;
}

static bool wifi_radio_cac_required(struct agent *a, struct wifi_radio_element *re,
				    int channel, int bandwidth, uint32_t *cac_time)
{
	struct wifi_radio_opclass opclass = {};

	if (!wifi_radio_to_ap(a, re->name))
		return false;

	/* Get fresh opclass preferences */
	if (WARN_ON(wifi_opclass_preferences(re->name, &opclass))) {
		return false;
	}

	return wifi_opclass_cac_required(&opclass, channel, bandwidth, cac_time);
}

static int wifi_bsta_handle_cac_required(struct agent *a,
					 struct wifi_radio_element *re,
					 struct netif_bk *bk)
{
	struct netif_bk *new_bk = NULL;
	struct netif_bk *sta;
	uint32_t cac_time = 0;

	if (!a->cfg.ap_follow_sta_dfs)
		return 0;

	re->cac_required = wifi_radio_cac_required(a, re, re->current_channel,
						   re->current_bandwidth, &cac_time);

	dbg("bsta %s connected on %d/%d cac required %d time %us\n", bk->ifname,
	    re->current_channel, re->current_bandwidth, re->cac_required,
	    cac_time);

	if (!re->cac_required)
		return 0;

	/* Find new bsta */
	list_for_each_entry(sta, &a->bklist, list) {
		if (!strcmp(sta->ifname, bk->ifname))
			continue;

		new_bk = sta;
		break;
	}

	/* Disable current connection */
	config_disable_bsta(bk->cfg);
	bk->connected = false;


	if (new_bk) {
		dbg("[%s] connect cac required use new link %s\n", bk->ifname, new_bk->ifname);
		config_enable_bsta(new_bk->cfg);
		dynbh_bsta_use_link(a, new_bk->ifname);
	} else {
		dbg("connect cac required disconnect current link %s\n", bk->ifname);
		dynbh_bsta_use_link(a, "none");
	}

	wifi_bsta_disconnect(a, bk->ifname);
	cac_time += 20;
	if (!a->dynbh_upgrade_ongoing)
		timer_set(&a->dynbh_scheduler, cac_time * 1000);

	timestamp_update(&bk->cac_start);
	bk->cac_time = cac_time - 5;

	/* Use 80MHz */
	if (re->current_bandwidth < 80)
		re->current_opclass =
			wifi_opclass_find_id_from_channel(&re->opclass,
					re->current_channel, 80);

	/* Start CAC */
	if (WARN_ON(agent_channel_switch(a, re->macaddr, re->current_channel, re->current_opclass))) {
		bk->cac_time = 0;
		if (!a->dynbh_upgrade_ongoing)
			timer_set(&a->dynbh_scheduler, 5 * 1000);
	}

	/* Don't wait - switch bsta */
	dynbh_bsta_scan_on_enabled(a);

	return 1;
}

void wifi_bsta_check_cac_done(struct agent *a)
{
	struct wifi_radio_element *re = NULL;
	struct netif_bk *bk;
	uint8_t enable_prio = 255;
	bool cac_required;
	uint32_t cac_time;

	if (!a->cfg.ap_follow_sta_dfs)
		return;

	/* lower prio is better */
	list_for_each_entry(bk, &a->bklist, list) {
		if (!bk->cfg->enabled)
			continue;
		if (bk->cfg->priority < enable_prio)
			enable_prio = bk->cfg->priority;
	}

	list_for_each_entry(bk, &a->bklist, list) {
		if (bk->cac_time) {
			re = wifi_ifname_to_radio_element(a, bk->ifname);
			if (re) {
				cac_required = wifi_radio_cac_required(a, re,
							re->current_channel,
							re->current_bandwidth,
							&cac_time);
				dbg("[%s] %d/%d cac_required %d\n", bk->ifname, re->current_channel,
				    re->current_bandwidth, cac_required);
			} else {
				cac_required = true;
			}

			if (timestamp_expired(&bk->cac_start, bk->cac_time * 1000) || !cac_required) {
				dbg("[%s] connect enable bsta again, cac end (required %d)\n", bk->ifname, cac_required);
				config_enable_bsta(bk->cfg);
				bk->cac_time = 0;
				if (a->dynbh_upgrade_ongoing) {
					a->dynbh_upgrade_ongoing = false;
					timer_set(&a->dynbh_scheduler, 5 * 1000);
				}
			}
		} else {
			/* In case we loose bk->cac_time during agent stop/start */
			if (!bk->cfg->enabled && bk->cfg->priority < enable_prio) {
				dbg("[%s] connect !cac_time enable higher prio (%d) bsta\n", bk->ifname, enable_prio);
				config_enable_bsta(bk->cfg);
			}
		}
	}
}

static void wifi_bsta_connect(struct agent *a, struct netif_bk *bk,
			      uint8_t *bssid, uint32_t freq)
{
	int i;
	char fmt[64] = {0};
	int bssid_expired = 0;
	struct wifi_bsta_status status = {};
	struct wifi_radio_element *re = NULL;
	uint32_t channel = 0;
	uint32_t bandwidth = 0;

	/* no operation if no bsta on same band is onboarded! */
	if (!wifi_is_band_onboarded(a, bk->cfg->band)) {
		dbg("|%s:%d| band %d is not onboard\n", __func__, __LINE__,
		    bk->cfg->band);
		return;
	}

	if (!a->cntlr_select.local)
		agent_disable_local_cntlr(a);

	dbg("|%s:%d| connect event received\n", __func__, __LINE__);

	bssid_expired = timestamp_expired(&a->disconnect_t,
			(a->cfg.dbhcfg.bh_miss_tmo + 15) * 1000);
	dbg("|%s:%d| new bssid " MACFMT " old bssid " MACFMT ","\
			" difftime exceeds 60s = %d\n", __func__, __LINE__,
			MAC2STR(bssid),	MAC2STR(bk->bssid), bssid_expired);


	if (((!hwaddr_is_zero(bk->bssid) && memcmp(bssid, bk->bssid, 6)) ||
	      bssid_expired) && !a->connected) {
		dbg("|%s:%d| new bssid or difftime exceeded 60s,"\
		    "setting radio states to ACTIVE\n",
		    __func__, __LINE__);
		/** upon connecting to new network - allow radios to be
		 * reconfigured
		 */
		for (i = 0; i < a->num_radios; i++)
			a->radios[i].state = AUTOCFG_ACTIVE;
	}

	agent_enable_fronthauls(a);

	memcpy(bk->bssid, bssid, 6);

	/* Assign bssid to configs
	 * If a bsta is solely onboarded using DPP, do not write anything
	 * as rest of DPP creds cannot persist
	 */
	if (bk->cfg->onboarded) {
		wifi_set_iface_bssid(bk, bssid);
		snprintf(fmt, sizeof(fmt), "write_bsta_config %s", bk->ifname);
		agent_exec_platform_scripts(fmt);
	}

	//backhaul_blacklist_clear(a);

	timestamp_update(&a->connect_t);
	/* TODO: eth always takes priority */
	if (!agent_is_backhaul_type_eth())
		agent_update_active_uplink(a, bk->ifname);
	a->autocfg_interval = 5;
	dbg("|%s:%d| Scheduling next ACS in %u seconds\n", __func__, __LINE__,
			a->autocfg_interval);
	timer_set(&a->autocfg_dispatcher, a->autocfg_interval * 1000);

	/* Get channel/bandwidth from bsta */
	if (!wifi_bsta_status(bk->ifname, &status)) {
		channel = status.channel;
		bandwidth = wifi_bw_to_bw(status.bandwidth);
	}

	/* Update radio channel/bw/opclass */
	re = wifi_ifname_to_radio_element(a, bk->ifname);
	if (re) {
		struct wifi_radio_status radio_status = {};

		re->current_channel = channel;
		re->current_bandwidth = bandwidth;
		re->current_opclass = wifi_opclass_find_id_from_channel(&re->opclass,
					re->current_channel, re->current_bandwidth);

		/* Fall back to radio status */
		if (!re->current_channel) {
			if (!wifi_radio_status(re->name, &radio_status)) {
				channel = radio_status.channel;
				bandwidth = radio_status.bandwidth;

				re->current_channel = radio_status.channel;
				re->current_bandwidth = radio_status.bandwidth;
				re->current_opclass = radio_status.opclass;
			}
		}

		/* Check CAC AP/BSTA action required */
		if (wifi_bsta_handle_cac_required(a, re, bk))
			return;
	}

	if (!channel && freq)
		channel = f2c(freq);

	if (channel) {
		char fmt[512] = {};

		snprintf(fmt, sizeof(fmt), "bsta_set_channel %s %u %u", bk->ifname, channel, bandwidth);
		agent_exec_platform_scripts(fmt);
	}

	bk->connected = true;
	a->connected = true;
}

static void wifi_bsta_event_scan_failed(struct agent *a, struct netif_bk *bk)
{
	struct wifi_radio_element *re = NULL;
	struct wifi_radio_opclass opclass = {};

	trace("[%s] %s called\n", bk->ifname, __func__);

	if (bk->connected)
		return;

	if (!bk->cfg->enabled)
		return;

	if (bk->cfg->band != BAND_5)
		return;

	re = wifi_ifname_to_radio_element(a, bk->ifname);
	if (!re)
		return;

	if (!wifi_radio_to_ap(a, re->name)) {
		trace("[%s] %s !ap_found\n", bk->ifname, __func__);
		return;
	}

	/* Get fresh opclass preferences */
	if (WARN_ON(wifi_opclass_preferences(re->name, &opclass))) {
		return;
	}

	if (!wifi_opclass_cac_ongoing(&opclass)) {
		trace("[%s] no cac_ongoing\n", bk->ifname);
		return;
	}

	dbg("[%s] cac ongoing skip CAC %s\n", bk->ifname, re->name);
	WARN_ON(wifi_stop_cac(re->name, 36, BW80));
	return;
}

static void wifi_bsta_event_handler(void *agent, struct blob_attr *msg)
{
	char ifname[16] = {0}, event[16] = {0};
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct netif_bk *bk;
	bool add, del;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	bk = find_bkhaul_by_ifname(a, ifname);
	if (!bk)
		return;

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	/* if scan event and eth backhaul is used - disconnect to prevent loop */
	if ((!strcmp(event, "scan-started") || !strcmp(event, "scan-failed"))
			&& agent_is_backhaul_type_eth())
		wifi_bsta_disconnect(a, ifname);

	/** if receive scan event from non-active link, while we are not trying
	 *  to progress or backhaul steer */
	if ((!strcmp(event, "scan-started") || !strcmp(event, "scan-failed"))
			&& ((!bk->connected && !a->dynbh_upgrade_ongoing && a->connected))
			&& !agent_bsta_steer_is_active(bk))
		wifi_bsta_disconnect(a, ifname);

	if (!strcmp(event, "scan-failed")) {
		wifi_bsta_event_scan_failed(a, bk);
		return;
	}

	if (add) {
		uint8_t bssid[6] = {0};
		char bssid_str[18] = {0};
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];
		uint32_t freq = 0;
		int remaining;
		struct netif_bk *b;

		if (!tb[2])
			return;

		/* don't handle event if bk is not enabled */
		if (!bk->cfg->enabled) {
			dbg("|%s:%d| bsta is not enabled - higher prio is "\
					"active, disconnect\n",
					__func__, __LINE__);
			wifi_bsta_disconnect(a, ifname);
			return;
		}


		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));
		if (!data[0])
			return;

		strncpy(bssid_str, blobmsg_data(data[0]),
				sizeof(bssid_str) - 1);

		hwaddr_aton(bssid_str, bssid);

		if (data[1])
			freq = blobmsg_get_u32(data[1]);

		wifi_bsta_connect(a, bk, bssid, freq);
		agent_manage_bsta(a, bk);

		timer_del(&a->bh_lost_timer);
		timer_del(&a->bh_reconf_timer);
#ifdef AGENT_ISLAND_PREVENTION
		if (a->cfg.island_prevention) {
			timer_del(&a->sta_disconnect_timer);
			timer_del(&a->fh_disable_timer);
		}
#endif /* AGENT_ISLAND_PREVENTION */

		/* send any pending failed backhaul steer responses */
		list_for_each_entry(b, &a->bklist, list) {
			if (b->bsta_steer.expired == false ||
			    b->bsta_steer.mid == 0)
				continue;

			/* should always be true from here */
			if (bk->bsta_steer.trigger == BK_STEER_MANDATE)
				send_backhaul_sta_steer_response(a, b, 0x01, bk->ifname);
			agent_bsta_steer_clear(b);
		}

		if (agent_bsta_steer_is_active(bk)) {
			/* steer was successful */

			set_value_by_string("mapagent", "agent",
					    "backhaul_override", "1",
					    UCI_TYPE_STRING);
			a->cfg.backhaul_override = 1;

			dynbh_bsta_use_link(a, bk->ifname);
			list_for_each_entry(b, &a->bklist, list) {
				if (strncmp(b->ifname, bk->ifname, IFNAMSIZ) /* && !agent_bsta_steer_is_active(b) */)
					config_disable_bsta(b->cfg);
			}
			agent_bsta_steer_handler(a, bk, bssid);
			return;
		}

		remaining = timer_remaining_ms(&a->onboarding_scheduler);
		if (remaining == -1 && bk->connected) {
			dynbh_bsta_disable_lower_priority(a, ifname);
			agent_config_reload(a);
			timer_set(&a->disable_unconnected_bstas_scheduler,
					  15 * 1000);
		}
	} else if (del) {
		bk->connected = false;
		/* Handle link loss if this was supposed to be an active bsta */
		if (bk->cfg->enabled) {
			char ul_ifname[16] = {0};

			dynbh_handle_bh_lost(a, bk);

			if (agent_get_backhaul_ifname(a, ul_ifname)) {
				if (!strncmp(ul_ifname, bk->ifname, IFNAMSIZ)) {
					char ev[512] = {0};

					dbg("%s: called.\n", __func__);

					if (!a)
						return;

					sprintf(ev,
						"{\"action\":\"unset_uplink\""
						",\"type\":\"wifi\"}");

					trace("cntlrinfo: %s\n", ev);
					agent_notify_event(a, "map.agent", ev);
					agent_exec_platform_scripts("unset_uplink wifi");
				}
			}
		}
		wifi_mod_bridge(a, ifname, "remove");
		a->connected = agent_is_bsta_connected(a);
	} else if (!strcmp(event, "wps-pbc-active")) {
		bk->wps_active = true;
	} else if (!strcmp(event, "wps-timeout")) {
		bk->wps_active = false;
	} else if (!strcmp(event, "csa")) {
		struct wifi_radio_element *re = NULL;
		struct wifi_bsta_status status = {};
		uint32_t cac_time;

		dbg("[%s] connect csa\n", ifname);

		if (a->cfg.ap_follow_sta_dfs) {
			re = wifi_ifname_to_radio_element(a, ifname);

			if (re) {
				/* Get channel/bandwidth from bsta */
				if (!wifi_bsta_status(ifname, &status)) {
					re->current_channel = status.channel;
					re->current_bandwidth = wifi_bw_to_bw(status.bandwidth);
					re->current_opclass =
						wifi_opclass_find_id_from_channel(
								&re->opclass,
								re->current_channel,
								re->current_bandwidth);
				}

				if (wifi_radio_cac_required(a, re, re->current_channel,
							    re->current_bandwidth, &cac_time)) {
					/* Simple disconnect, new connect will do required actions */
					wifi_bsta_disconnect(a, bk->ifname);
					dynbh_bsta_scan_on_enabled(a);
				}
			}
		}
	}
}


#define ONBOARDING_TIMER 15
static void wifi_wps_creds_event_handler(void *c, struct blob_attr *msg)
{
	char encryption[32] = {0}, ifname[16] = {0}, ssid[33] = {0},
			key[65] = {0};
	//uint16_t vlan = 0;
	struct blob_attr *tb[5];
	struct wifi_radio_element *radio;
	struct agent_config_radio *rcfg;
	struct agent *a = (struct agent *) c;
	static const struct blobmsg_policy ap_attr[5] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "encryption", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "key", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "vlan_id", .type = BLOBMSG_TYPE_INT32 },
	};
	struct netif_bk *bk, *p;
	int timeout = 0;

	blobmsg_parse(ap_attr, 5, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2] || !tb[3])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(encryption, blobmsg_data(tb[1]), sizeof(encryption) - 1);
	strncpy(ssid, blobmsg_data(tb[2]), sizeof(ssid) - 1);
	strncpy(key, blobmsg_data(tb[3]), sizeof(key) - 1);
	bk = find_bkhaul_by_ifname(a, ifname);
	if (!bk)
		return;

	/* don't wait for other bands if creds are synced via AP-Autoconfig */
	if (!a->cfg.eth_onboards_wifi_bhs) {
		list_for_each_entry(p, &a->bklist, list) {
			if (p->wps_active) {
				int remaining = timer_remaining_ms(&a->onboarding_scheduler);

				/* allow up to extra time for the other
				* band(s) to complete
				*/
				timeout = ONBOARDING_TIMER - (remaining == -1 ? 0 : remaining);
			}
		}
	}

	wifi_apply_iface_cfg(bk->ifname, encryption, ssid, key);
	bk->wps_active = false;
	bk->cfg->onboarded = true;

	//uci_reload_services("wireless"); /* move to platform script? */

	radio = wifi_ifname_to_radio_element(a, ifname);
	if (!radio)
		return;

	radio->onboarded = 1;
	uci_set_wireless_interface_option("mapagent", "radio", "device",
			radio->name, "onboarded", "1");

	timer_set(&a->onboarding_scheduler, timeout * 1000);

	rcfg = get_agent_config_radio(&a->cfg, ifname);
	if (!rcfg)
		return;

	rcfg->onboarded = true;

#if 0 /* Not used at present */
	vlan = 0;
	if (tb[4])
		vlan = (uint16_t) blobmsg_get_u16(tb[4]);

	if (vlan && bk->agent) {
		if (!(a->cfg.pcfg)) {
			a->cfg.pcfg = (struct policy_cfg *)calloc(1, sizeof(struct policy_cfg));
			if (!(a->cfg.pcfg)) {
				err("%s:%d - memory allocation failed\n",
				    __func__, __LINE__);
				return;
			}
		}
		bk->agent->cfg.pcfg->pvid = vlan;

		agent_apply_traffic_separation(bk->agent);
	}
#endif
}

static void agent_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev,
		const char *type, struct blob_attr *msg)
{
	int i;
	char *str;
	struct agent *a = container_of(ev, struct agent, evh);
	struct wifi_ev_handler {
		const char *ev_type;
		void (*handler)(void *ctx, struct blob_attr *ev_data);
	} evs[] = {
		{ "wifi.sta", wifi_sta_event_handler },
		{ "wifi.channel", wifi_channel_event_handler },
		{ "wifi.radio", wifi_radio_event_handler },
		{ "wifi.iface", wifi_iface_event_handler },
		{ "wps_credentials", wifi_wps_creds_event_handler },
		{ "wifi.bsta", wifi_bsta_event_handler },
		{ "ethport", ethport_event_handler },
		{ "ubus.object.add", uobj_add_event_handler },
		{ "wifi.radio", wifi_chan_change_event_handler },
		{ "wifi.radio", wifi_cac_event_handler},
#if (EASYMESH_VERSION > 2)
		{ "wifi.sta", dpp_sta_event_handler },
		{ "wifi.ap", dpp_configurator_evh},
		{ "wifi.bsta", dpp_enrollee_evh}
#endif
	};

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	info("[ &agent = %p ] Received [event = %s]  [val = %s]\n",
			a, type, str);

	for (i = 0; i < ARRAY_SIZE(evs); i++) {
		if (!strcmp(type, evs[i].ev_type)) {
			evs[i].handler(a, msg);
		}
	}

	free(str);
}

static void ieee1905_cmdu_event_handler(void *c, struct blob_attr *msg)
{
	char src[18] = { 0 }, src_origin[18] = { 0 };
	struct agent *a = (struct agent *)c;
	uint8_t srcmac[6], origin[6];
	char in_ifname[16] = {0};
	struct blob_attr *tb[6];
	char *cmdustr = NULL;
	uint8_t *cmdu = NULL;
	int cmdu_strlen = 0;
	uint16_t mid = 0;
	int cmdulen = 0;
	uint16_t type;
	int ret;
	static const struct blobmsg_policy cmdu_attrs[6] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT16 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT16 },
		[2] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "source", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "origin", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "cmdu", .type = BLOBMSG_TYPE_STRING },
	};

	blobmsg_parse(cmdu_attrs, 6, tb, blob_data(msg), blob_len(msg));

	trace("%s: --->\n", __func__);

	if (!tb[0] || !tb[1]) {
		trace("%s: no type or mid.\n", __func__);
		return;
	}

	if (tb[0]) {
		type = (uint16_t)blobmsg_get_u16(tb[0]);
		if (!is_cmdu_for_us(a, type)) {
			trace("%s: type: 0x%x is_cmdu_for_us false.\n", __func__, type);
			return;
		}
	}

	if (tb[1])
		mid = (uint16_t)blobmsg_get_u16(tb[1]);

	if (tb[2])
		strncpy(in_ifname, blobmsg_data(tb[2]), 15);

	if (tb[3]) {
		strncpy(src, blobmsg_data(tb[3]), 17);
		hwaddr_aton(src, srcmac);
	}

	if (tb[4]) {
		strncpy(src_origin, blobmsg_data(tb[4]), 17);
		hwaddr_aton(src_origin, origin);
	}

	if (tb[5]) {
		cmdu_strlen = blobmsg_data_len(tb[5]);
		cmdulen = (cmdu_strlen - 1) / 2;

		cmdustr = calloc(cmdu_strlen + 1, sizeof(char));
		if (!cmdustr)
			return;

		strncpy(cmdustr, blobmsg_data(tb[5]), cmdu_strlen);
		cmdu = calloc(cmdulen, sizeof(uint8_t));
		if (!cmdu) {
			free(cmdustr);
			return;
		}

		strtob(cmdustr, cmdulen, cmdu);
		free(cmdustr);
	}

	ret = CMDU_HOOK(a, type, mid,
			srcmac,
			cmdu, cmdulen,
			CMDU_RX);
	if (ret == CMDU_DROP) {
		warn("[%s]: cmdu 0x%04X dropped by extension at Rx\n", __func__, type);
		goto out;
	}

	agent_handle_map_event(a, type, mid, in_ifname, srcmac, origin, cmdu + 8, cmdulen - 8);
out:
	free(cmdu);
}

static void sync_neighbor_lists(struct netif_ap *ap)
{
	struct nbr nbr[64] = {};
	int nbr_num = 64;
	struct neighbor *n;
	bool found;
	int ret;
	int i;

	/* Update new & changed nbr entries in driver */
	list_for_each_entry(n, &ap->nbrlist, list) {

		if (n->flags & NBR_FLAG_DRV_UPDATED)
			continue;

		/* update nbr in wifi driver */
		dbg("[%s] Update neighbor "
		    MACFMT " entry in driver\n",
		    ap->ifname,
		    MAC2STR(n->nbr.bssid));
		wifi_del_neighbor(ap->ifname, n->nbr.bssid);
		wifi_add_neighbor(ap->ifname, &n->nbr);
		n->flags |= NBR_FLAG_DRV_UPDATED;
	}

	/* Trigger removal of dead entries from list in driver */
	ret = wifi_get_neighbor_list(ap->ifname, nbr, &nbr_num);
	if (ret)
		return;

	for (i = 0; i < nbr_num; i++) {
		found = false;
		list_for_each_entry(n, &ap->nbrlist, list) {
			if (!memcmp(nbr[i].bssid, n->nbr.bssid, 6)) {
				found = true;
				break;
			}
		}

		if (!found) {
			/* Remove dead entry from driver list */
			dbg("[%s] Delete neighbor "
			    MACFMT " entry in driver\n",
			    ap->ifname, MAC2STR(nbr[i].bssid));
			wifi_del_neighbor(ap->ifname, nbr[i].bssid);
		}
	}
}

static void refresh_neighbor_list(atimer_t *t)
{
	struct netif_ap *ap = container_of(t, struct netif_ap, nbr_timer);

	trace("%s: ap = %s\n", __func__, ap->ifname);

	/* First remove aged out entries from FH's nbrlist */
	delete_expired_entries(ap, struct neighbor, &ap->nbrlist, list,
				tsp, NBR_AGEOUT_INTERVAL, NULL,
				ap->nbr_nr);

	/* Now update list in driver so that it matches FH's nbrlist */
	sync_neighbor_lists(ap);

	/* Check again in couple of minutes */
	timer_set(&ap->nbr_timer, NBR_REFRESH_INTERVAL);
}

static void refresh_bssinfo(atimer_t *t)
{
	struct netif_ap *vif = container_of(t, struct netif_ap, bss_timer);
	struct agent *a = vif->agent;
	struct wifi_ap_status ap_status = {};
	uint8_t sta[128 * 6] = {};
	int num_sta = 128;
	int i;

	dbg("%s: vif = %s\n", __func__, vif->ifname);
	if (!wifi_ap_status(vif->ifname, &ap_status)) {
		vif->channel = ap_status.ap.bss.channel;
		memcpy(vif->bssid, ap_status.ap.bss.bssid, 6);
		memcpy(vif->ssid, ap_status.ap.bss.ssid, 32);
		/* others.. */
		if (ap_status.ap.bss.load.utilization != 0xff)
			vif->bssload = ap_status.ap.bss.load.utilization;

		dbg("ap: %s   bssid: " MACFMT
				"  channel: %d   bssload: %d\n",
				vif->ifname, MAC2STR(vif->bssid),
				vif->channel, vif->bssload);

		if (!wifi_get_assoclist(vif->ifname, sta, &num_sta)) {
			for (i = 0; i < num_sta; i++)
				wifi_add_sta(a, vif->ifname, &sta[i * 6]);
		}
	}

	timer_set(&vif->bss_timer, BSS_REFRESH_INTERVAL);
}

#if 0
static int run_agent(struct agent *a)
{
	struct netif_ap *ap;
	int i = 0;

	if (a->cfg.runfreq == AGENT_RUN_OFF)
		return 0;

	/* for each enabled wifi(life) ap interface, */
	list_for_each_entry(ap, &a->aplist, list) {
		if (!ap->cfg->enabled)
			continue;

		trace("Agent: Set up refresh timers for '%s'...\n", ap->ifname);
		/* do ... */
		timer_set(&ap->bss_timer, (2 + i++) * 1000);
		timer_set(&ap->nbr_timer, (5 + i++) * 1000);
	}
	return 0;
}
#endif

static void parse_radio_stats(struct ubus_request *req, int type,
			      struct blob_attr *msg)
{
	struct wifi_radio_element *re = (struct wifi_radio_element *)req->priv;
	static const struct blobmsg_policy stats_attr[] = {
		[0] = { .name = "tx_bytes", .type = BLOBMSG_TYPE_INT64 },
		[1] = { .name = "tx_packets", .type = BLOBMSG_TYPE_INT64 },
		[2] = { .name = "tx_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[3] = { .name = "tx_dropped_packets", .type = BLOBMSG_TYPE_INT64 },
		[4] = { .name = "rx_bytes", .type = BLOBMSG_TYPE_INT64 },
		[5] = { .name = "rx_packets", .type = BLOBMSG_TYPE_INT64 },
		[6] = { .name = "rx_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[7] = { .name = "rx_dropped_packets", .type = BLOBMSG_TYPE_INT64 },
		[8] = { .name = "rx_plcp_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[9] = { .name = "rx_fcs_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[10] = { .name = "rx_mac_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[11] = { .name = "rx_unknown_packets", .type = BLOBMSG_TYPE_INT64 },
	};
	struct blob_attr *tb[ARRAY_SIZE(stats_attr)];

	blobmsg_parse(stats_attr, ARRAY_SIZE(stats_attr), tb, blobmsg_data(msg), blob_len(msg));

	if (tb[0])
		re->tx_bytes = blobmsg_get_u64(tb[0]);

	if (tb[1])
		re->tx_packets = blobmsg_get_u64(tb[1]);

	if (tb[2])
		re->tx_error_packets = blobmsg_get_u64(tb[2]);

	if (tb[3])
		re->tx_dropped_packets = blobmsg_get_u64(tb[3]);

	if (tb[4])
		re->rx_bytes = blobmsg_get_u64(tb[4]);

	if (tb[5])
		re->rx_packets = blobmsg_get_u64(tb[5]);

	if (tb[6])
		re->rx_error_packets = blobmsg_get_u64(tb[6]);

	if (tb[7])
		re->rx_dropped_packets = blobmsg_get_u64(tb[7]);

	if (tb[8])
		re->rx_plcp_error_packets = blobmsg_get_u64(tb[8]);

	if (tb[9])
		re->rx_fcs_error_packets = blobmsg_get_u64(tb[9]);

	if (tb[10])
		re->rx_mac_error_packets = blobmsg_get_u64(tb[10]);

	if (tb[11])
		re->rx_unknown_packets = blobmsg_get_u64(tb[11]);
}

static void parse_radio_cac_methods(struct ubus_request *req, int type,
				    struct blob_attr *msg)
{
	struct wifi_radio_element *re = (struct wifi_radio_element *)req->priv;
	struct blob_attr *method;
	int rem_methods;

	blobmsg_for_each_attr(method, msg, rem_methods) {
		if (blobmsg_type(method) != BLOBMSG_TYPE_STRING)
			continue;

		if (!strcmp(blobmsg_data(method), "continous"))
			re->cac_methods |= 1U << WIFI_CAC_CONTINUOUS;
		else if (!strcmp(blobmsg_data(method), "mimo-reduced"))
			re->cac_methods |= 1U << WIFI_CAC_MIMO_REDUCED;
		else if (!strcmp(blobmsg_data(method), "continous-dedicated"))
			re->cac_methods |= 1U << WIFI_CAC_DEDICATED;
		else if (!strcmp(blobmsg_data(method), "time-sliced"))
			re->cac_methods |= 1U << WIFI_CAC_TIME_SLICED;
	}
}

static void parse_radio(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct wifi_radio_element *re = (struct wifi_radio_element *)req->priv;
	static const struct blobmsg_policy radio_attr[] = {
		[0] = { .name = "isup", .type = BLOBMSG_TYPE_BOOL },
		[1] = { .name = "band", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "noise", .type = BLOBMSG_TYPE_INT32 },
		[3] = { .name = "rx_streams", .type = BLOBMSG_TYPE_INT8 },
		[4] = { .name = "tx_streams", .type = BLOBMSG_TYPE_INT8 },
		[5] = { .name = "supp_channels", .type = BLOBMSG_TYPE_ARRAY },
		[6] = { .name = "opclass", .type = BLOBMSG_TYPE_INT32 },
		[7] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[8] = { .name = "regdomain", .type = BLOBMSG_TYPE_STRING },
		[9] = { .name = "txpower", .type = BLOBMSG_TYPE_INT32 },
		[10] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[11] = { .name = "diagnostics", .type = BLOBMSG_TYPE_TABLE },
		[12] = { .name = "bandwidth", .type = BLOBMSG_TYPE_INT32 },
		[13] = { .name = "vendor_id", .type = BLOBMSG_TYPE_STRING },
		[14] = { .name = "stats", .type = BLOBMSG_TYPE_TABLE },
		[15] = { .name = "cac_methods", .type = BLOBMSG_TYPE_ARRAY },
	};
	struct blob_attr *tb[ARRAY_SIZE(radio_attr)];

	dbg("%s ---> %s\n", __func__, re->name);

	blobmsg_parse(radio_attr, ARRAY_SIZE(radio_attr), tb, blob_data(msg), blob_len(msg));

	if (tb[0])
		re->enabled = blobmsg_get_bool(tb[0]);

	if (tb[1]) {
		char *band;

		band = blobmsg_get_string(tb[1]);
		if (!strncmp(band, "2.4GHz", strlen("2.4GHz")))
			re->band = BAND_2;
		else if (!strncmp(band, "5GHz", strlen("5GHz")))
			re->band = BAND_5;
		else if (!strncmp(band, "6GHz", strlen("6GHz")))
			re->band = BAND_6;
		else
			re->band = BAND_UNKNOWN;
	}

	if (tb[2])
		re->anpi = wifi_noise_to_anpi((int)blobmsg_get_u32(tb[2]));

	if (tb[3])
		re->rx_streams = blobmsg_get_u8(tb[3]);

	if (tb[4])
		re->tx_streams = blobmsg_get_u8(tb[4]);
	if (tb[6])
		re->current_opclass = (uint8_t) blobmsg_get_u32(tb[6]);

	if (tb[7])
		re->current_channel = (uint8_t) blobmsg_get_u32(tb[7]);

	if (tb[8])
		memcpy(re->country_code, blobmsg_data(tb[8]), 2);

	if (tb[9])
		re->current_txpower_percent = (uint8_t) blobmsg_get_u32(tb[9]);

	if (tb[10]) {
		char macaddr[18] = {0};

		strncpy(macaddr, blobmsg_data(tb[10]), 17);
		if (!hwaddr_aton(macaddr, re->macaddr))
			return;
	}

	if (tb[11]) {
		struct blob_attr *tb1[4];
		static const struct blobmsg_policy radio_diag_attr[4] = {
			[0] = { .name = "tx_airtime", .type = BLOBMSG_TYPE_INT64 },
			[1] = { .name = "rx_airtime", .type = BLOBMSG_TYPE_INT64 },
			[2] = { .name = "channel_busy", .type = BLOBMSG_TYPE_INT64 },
			[3] = { .name = "obss_airtime", .type = BLOBMSG_TYPE_INT64 },
		};

		blobmsg_parse(radio_diag_attr, 4, tb1, blobmsg_data(tb[11]), blob_len(tb[11]));

		/* get percentage of utilization and scale linearly with 255 */
		if (tb1[0])
			re->tx_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[0]) / (float)1000000) * 255) & 0xff;

		if (tb1[1])
			re->rx_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[1]) / (float)1000000) * 255) & 0xff;

		if (tb1[2])
			re->total_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[2]) / (float)1000000) * 255) & 0xff;

		if (tb1[3])
			re->other_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[3]) / (float)1000000) * 255) & 0xff;

	}

	if (tb[12])
		re->current_bandwidth = blobmsg_get_u32(tb[12]);

	if (tb[13])
		strncpy(re->vendor, blobmsg_data(tb[13]),
				sizeof(re->vendor) - 1);

	if (tb[14])
		parse_radio_stats(req, type, tb[14]);

	if (tb[15])
		parse_radio_cac_methods(req, type, tb[15]);

	//Put the preference as per the config file
	agent_config_opclass(re);
}

struct wifi_scanres_channel_element *wifi_get_scanres_ch_element(
		struct wifi_radio_element *re, uint8_t classid, uint8_t ch)
{
	int i, j;

	if (!re || !re->scanlist)
		return NULL;

	for (i = 0; i < re->scanlist->num_opclass; i++) {
		struct wifi_scanres_opclass_element *op_el;

		op_el = &re->scanlist->opclass_scanlist[i];
		if (op_el->opclass != classid)
			continue;

		for (j = 0; j < op_el->num_channels; j++) {
			if (op_el->channel_scanlist[j].channel == ch)
				return &op_el->channel_scanlist[j];
		}
	}

	dbg("No operating class with channel %d\n", ch);
	return NULL;
}

static void free_scanlist(struct wifi_radio_element *re)
{
	trace("agent: %s: --->\n", __func__);

	struct wifi_scanres_element *scanres_el;
	int j;

	/* return if no scanresults to free */
	if (!re || !re->scanlist)
		return;

	scanres_el = re->scanlist;
	for (j = 0; j < scanres_el->num_opclass; j++) {
		int k;
		struct wifi_scanres_opclass_element *opclass;

		opclass = &scanres_el->opclass_scanlist[j];
		if (!opclass)
			continue;

		for (k = 0; k < opclass->num_channels; k++) {
			struct wifi_scanres_channel_element *ch_el;

			ch_el = &opclass->channel_scanlist[k];
			if (!ch_el)
				continue;

			free(ch_el->nbrlist);
			ch_el->nbrlist = NULL;
		}

		free(opclass->channel_scanlist);
		opclass->channel_scanlist = NULL;
	}

	free(scanres_el->opclass_scanlist);
	scanres_el->opclass_scanlist = NULL;
	free(scanres_el);
	scanres_el = NULL;
}

#define ADD_NBR_TIMER 10
void reschedule_nbrlist_update(struct netif_ap *ap)
{
	int elapsed = 0;
	int remaining = timer_remaining_ms(&ap->nbr_timer);

	if (remaining < ADD_NBR_TIMER)
		elapsed = ADD_NBR_TIMER - (remaining == -1 ? 0 : remaining);

	timer_set(&ap->nbr_timer, (ADD_NBR_TIMER - elapsed) * 1000);
}

void free_scanlist_neighbors(struct wifi_radio_element *re)
{
	struct wifi_scanres_element *scanres_el;
	int j;

	scanres_el = re->scanlist;
	if (!scanres_el)
		return;

	for (j = 0; j < scanres_el->num_opclass; j++) {
		int k;
		struct wifi_scanres_opclass_element *opclass = &scanres_el->opclass_scanlist[j];

		for (k = 0; k < opclass->num_channels; k++) {
			struct wifi_scanres_channel_element *ch_el = &opclass->channel_scanlist[k];

			if (ch_el->nbrlist)
				free(ch_el->nbrlist);
			ch_el->nbrlist = NULL;
			ch_el->num_neighbors = 0;
		}
	}
}

static int init_scanlist(struct agent *a, const char *radio)
{
	struct wifi_radio_opclass *opclass;
	struct wifi_radio_opclass_entry *entry;
	struct wifi_radio_opclass_channel *channel;
	struct wifi_radio_element *re;
	int num_channels;
	int opclass_supported;
	int i, j, k;

	trace("agent: %s: --->\n", __func__);

	re = wifi_radio_to_radio_element(a, radio);
	if (WARN_ON(!re))
		return -1;

	opclass = &re->opclass;
	k = 0;

	free_scanlist(re);

	re->scanlist = calloc(1, sizeof(*re->scanlist));
	if (!re->scanlist) {
		dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
		return -1;
	}

	opclass_supported = wifi_opclass_num_supported_bw20(opclass);
	re->scanlist->opclass_scanlist =
		calloc(opclass_supported,
			sizeof(struct wifi_scanres_opclass_element));

	if (!re->scanlist->opclass_scanlist) {
		dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
		goto error;
	}

	re->scanlist->num_opclass = 0;
	for (i = 0; i < opclass->entry_num; i++) {
		entry = &opclass->entry[i];

		if (!wifi_opclass_id_supported(opclass, entry->id)
				|| entry->bandwidth != 20) /* caps */
			continue;

		re->scanlist->opclass_scanlist[k].opclass = entry->id;
		re->scanlist->opclass_scanlist[k].bandwidth = entry->bandwidth;
		/* Store for all supported channels in that opclass */
		num_channels = entry->channel_num;
		re->scanlist->opclass_scanlist[k].num_channels =
				num_channels;

		re->scanlist->opclass_scanlist[k].channel_scanlist =
			calloc(num_channels,
				sizeof(struct wifi_scanres_channel_element));

		if (!re->scanlist->opclass_scanlist[k].channel_scanlist) {
			dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
			goto error;
		}
		/* Store for all supported opclasses */
		re->scanlist->num_opclass++;

		for (j = 0; j < entry->channel_num; j++) {
			channel = &entry->channel[j];
			re->scanlist->opclass_scanlist[k].channel_scanlist[j].channel =
							channel->channel;
		}

		k++;
	}

	return 0;

error:
	free_scanlist(re);
	return -1;
}

static void agent_available_scan_timeout(atimer_t *t)
{
	dbg("[Status code] SCAN NOT COMPLETED\n\n");

	struct wifi_netdev *ndev;
	int ret;

	ndev = container_of(t, struct wifi_netdev, available_scan_timer);

	if (WARN_ON(!ndev))
		return;

	if (WARN_ON(!ndev->re))
		return;

	if (ndev->re->scan_state != SCAN_REQUESTED || !ndev->scan_req.mid)
		return;

	// TODO: revisit - maybe redundant
	ndev->scan_req.status = CH_SCAN_STATUS_SCAN_NOT_COMPLETED;

	ret = agent_send_ch_scan_response(ndev->agent, ndev, &ndev->scan_req);
	if (ret)
		return;

	ndev->scan_req = (const struct wifi_scan_request_radio){ 0 };
}

static void _enumerate_wifi_objects(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	trace("%s: --->\n", __func__);
	struct radio_apcfg_state {
		char name[16];
		enum autocfg_state state;
		uint16_t mid;
		uint16_t wsc_mid;
	} apcfg_state[WIFI_DEVICE_MAX_NUM] = {0};
	struct agent *a = (struct agent *)req->priv;
	struct json_object *json_msg;
	struct json_object *radio_array;
	char *json_str;
	int i, j, k, len, prev_len = 0;

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	json_object_object_get_ex(json_msg, "radios", &radio_array);
	len = json_object_array_length(radio_array);
	if (len > WIFI_DEVICE_MAX_NUM)
		len = WIFI_DEVICE_MAX_NUM;

	trace("%s: num_radios(len) = %d\n", __func__, len);
	if (len) {
		/* store current radio apcfg states */
		for (i = 0; i < a->num_radios; i++) {
			strncpy(apcfg_state[i].name, a->radios[i].name,
				IFNAMSIZ - 1);
			apcfg_state[i].state = a->radios[i].state;
			apcfg_state[i].mid = a->radios[i].mid;
			apcfg_state[i].wsc_mid = a->radios[i].wsc_mid;
		}
		prev_len = a->num_radios;
		/* clears radio apcfg states */
		agent_free_radios(a);
		clear_aplist(a);
		a->num_radios = len;
	}

	for (i = 0; i < len; i++) {
		struct json_object *radio_obj, *radio_obj_name;
		struct json_object *ap_obj, *ap_obj_name;
		struct json_object *bk_obj, *bk_obj_name;
		struct json_object *ap_array;
		struct json_object *bk_array;
		int ap_len, bk_len;
		const char *radio_name;
		char r_objname[32] = {0};
		//const char *r_fmt = "wifi.radio.%s";
		wifi_object_t r_wobj = WIFI_OBJECT_INVALID;
		struct agent_config_radio *r_cfg;
		radio_obj = json_object_array_get_idx(radio_array, i);
		json_object_object_get_ex(radio_obj, "name", &radio_obj_name);
		radio_name = json_object_get_string(radio_obj_name);
		strncpy(a->ifs[i].radio, radio_name, 15);
		strncpy(a->radios[i].name, radio_name, 15);
		timer_init(&a->ifs[i].available_scan_timer,
				agent_available_scan_timeout);
		a->ifs[i].agent = a;
		a->ifs[i].re = &a->radios[i];

		a->radios[i].agent = a;
		timer_init(&a->radios[i].preference_report_timer, preference_report_timer_cb);

		for (j = 0; j < prev_len; j++) {
			/* write back apcfg state if any */
			if (!strncmp(apcfg_state[j].name, a->radios[i].name,
				     IFNAMSIZ - 1)) {
				a->radios[i].state = apcfg_state[j].state;
				a->radios[i].mid = apcfg_state[j].mid;
				a->radios[i].wsc_mid = apcfg_state[j].wsc_mid;
			}
		}

		list_for_each_entry(r_cfg, &a->cfg.radiolist, list) {
			if (strncmp(r_cfg->name, a->radios[i].name, 15))
				continue;

			a->radios[i].onboarded = r_cfg->onboarded;
			a->radios[i].dedicated_backhaul = r_cfg->dedicated_backhaul;
			break;
		}

		// parse channel range
		// if channel range is stricly beneath <= 64, assign low
		// else if channel rangei stricly above 100, >= 100, assign high
		// else if both ranges, no assignation

		json_object_object_get_ex(radio_obj, "accesspoints", &ap_array);
		ap_len = json_object_array_length(ap_array);
		for (j = 0; j < ap_len; j++) {
			const char *ap_name;

			ap_obj = json_object_array_get_idx(ap_array, j);
			json_object_object_get_ex(ap_obj, "ifname",
					&ap_obj_name);
			ap_name = json_object_get_string(ap_obj_name);
			strncpy(a->ifs[i].iface[j].name, ap_name, 15);
			a->ifs[i].iface[j].mode = WIFI_IFACE_AP;
		}

		json_object_object_get_ex(radio_obj, "backhauls", &bk_array);
		bk_len = json_object_array_length(bk_array);
		for (k = 0; k < bk_len; k++) {
			const char *bk_name;

			bk_obj = json_object_array_get_idx(bk_array, k);
			json_object_object_get_ex(bk_obj, "ifname",
					&bk_obj_name);
			bk_name = json_object_get_string(bk_obj_name);
			strncpy(a->ifs[i].iface[k + j].name, bk_name, 15);
			a->ifs[i].iface[k + j].mode = WIFI_IFACE_BK;
		}

		snprintf(r_objname, 31, "wifi.radio.%s", radio_name);
		r_wobj = ubus_get_object(a->ubus_ctx, r_objname);
		if (r_wobj == WIFI_OBJECT_INVALID) {
			//dbg("%s not present! skipping '%s' from config\n",
			//			r_objname, f->name);
			continue;
		}

		dbg("%s: getting radio status = radios[%d]\n", __func__, i);
		ubus_call_object(a, r_wobj, "status", parse_radio, &a->radios[i]);

		/* Get fresh opclass preferences after scan */
		wifi_radio_update_opclass_preferences(a, radio_name, 1);
		agent_set_post_scan_action_pref(a, radio_name, true);

		/* Allocate radio_element's scanlist */
		init_scanlist(a, radio_name);
	}

	agent_init_wsc_attributes(a);
out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

static int enumerate_wifi_objects(struct agent *a)
{
	trace("%s: --->\n", __func__);

	struct blob_buf bb = {};
	int ret, retry = 0;

	blob_buf_init(&bb, 0);
	while ((ret = ubus_invoke(a->ubus_ctx, a->wifi, "status", bb.head,
			_enumerate_wifi_objects, a, 2 * 1000) || !a->num_radios)
			&& retry < 5) {

		err("|%s:%d| Failed to get wifi status(ret = %d), OR "\
		    "num_radios is 0 (%d) retry in 0.5s, wifi object:%u, "\
		    "retry:%d\n", __func__, __LINE__, ret, a->num_radios,
		    a->wifi, retry);
		usleep(500 * 1000); /* 0.5s sleep */
		retry++;
		a->wifi = ubus_get_object(a->ubus_ctx, "wifi");
	}
	trace("%s: num of retries = %d\n", __func__, retry);

	if (ret)
		timer_set(&a->init_ifaces_scheduler, 30 * 1000);

	blob_buf_free(&bb);
	return 0;
}

struct netif_ap *netif_alloc_ap(const char *ifname)
{
	struct netif_ap *n = NULL;

	n = calloc(1, sizeof(struct netif_ap));
	if (!n)
		return NULL;

	INIT_LIST_HEAD(&n->stalist);
	INIT_LIST_HEAD(&n->nbrlist);
	INIT_LIST_HEAD(&n->restrict_stalist);
	snprintf(n->ifname, 15, "%s", ifname);

	timer_init(&n->nbr_timer, refresh_neighbor_list);
	timer_init(&n->bss_timer, refresh_bssinfo);

	return n;
}

static struct netif_bk *netif_alloc_bk(const char *ifname)
{
	struct netif_bk *n = NULL;

	n = calloc(1, sizeof(struct netif_bk));
	if (!n)
		return NULL;

	snprintf(n->ifname, 15, "%s", ifname);
	timer_init(&n->steer_timeout, bsta_steer_cb);
	n->wifi = WIFI_OBJECT_INVALID;
	n->radio = WIFI_OBJECT_INVALID;

#if (EASYMESH_VERSION > 2)
	timer_init(&n->dpp_do_chirp, dpp_period_chirp);
#endif

	return n;
}

struct netif_wds *netif_alloc_wds(struct agent *a, const char *ifname)
{
	struct netif_wds *n = NULL;

	n = calloc(1, sizeof(struct netif_wds));
	if (!n)
		return NULL;

	snprintf(n->ifname, 15, "%s", ifname);

	list_add(&n->list, &a->wdslist);

	return n;
}

void netif_free_wds(struct netif_wds *n)
{
	list_del(&n->list);
	free(n);
}

void netif_free_wds_by_ifname(struct agent *a, const char *ifname)
{
	struct netif_wds *n = NULL;

	n = find_wds_by_ifname(a, ifname);
	if (!n)
		return;

	netif_free_wds(n);
}

/* get first ap on the radio */
struct netif_ap *wifi_radio_to_ap(struct agent *a, const char *radio)
{
	struct netif_ap *p;

	list_for_each_entry(p, &a->aplist, list) {
		const char *ap_radio;

		ap_radio = wifi_ifname_to_radio(a, p->ifname);
		if (!ap_radio)
			continue;

		if (strncmp(radio, ap_radio, 15))
			continue;

		return p;
	}

	return NULL;
}

static void parse_dot11n(struct netif_ap *ap, struct blob_attr *arg)
{
	int supp_mcs = 0;
	struct blob_attr *tb[4];
	static const struct blobmsg_policy ap_attr[4] = {
		[0] = { .name = "dot11n_sgi20", .type = BLOBMSG_TYPE_BOOL },
		[1] = { .name = "dot11n_sgi40", .type = BLOBMSG_TYPE_BOOL },
		[2] = { .name = "dot11n_40", .type = BLOBMSG_TYPE_BOOL },
		[3] = { .name = "dot11n_supp_max_mcs", .type = BLOBMSG_TYPE_INT32 }
	};

	blobmsg_parse(ap_attr, 4, tb, blobmsg_data(arg), blobmsg_data_len(arg));

	supp_mcs = blobmsg_get_u32(tb[3]);
	supp_mcs = supp_mcs/8;
	ap->caps.ht |= (supp_mcs & 0x3) << 6;
	ap->caps.ht |= (supp_mcs & 0x3) << 4;
	ap->caps.ht |= ((blobmsg_get_bool(tb[0]) ? 1 : 0) << 3);
	ap->caps.ht |= ((blobmsg_get_bool(tb[1]) ? 1 : 0) << 2);
	ap->caps.ht |= ((blobmsg_get_bool(tb[2]) ? 1 : 0) << 1);
}

#define VHT_CAP_MAX_MCS	8
void parse_dot11ac_mcs(uint8_t *supp_mcs, int mcs, int nss)
{
	int i;
	int octel;
	int shift;
	uint8_t mask[4] = {0xfc, 0xf3, 0xcf, 0x3f};
	int nss_cnt = min(nss, VHT_CAP_MAX_MCS);

	for (i = 0; i < nss_cnt; i++) {
		octel = (2 * i) / 8;
		shift = (2 * i) % 8;

		if (mcs == 9)
			supp_mcs[octel] &= (mask[i%4] | (0x02 << shift));
		else if (mcs == 8)
			supp_mcs[octel] &= (mask[i%4] | (0x01 << shift));
		else if (mcs == 7)
			supp_mcs[octel] &= (mask[i%4] | (0x00 << shift));
	}
}

static void parse_dot11ac(struct netif_ap *ap, struct blob_attr *arg)
{
	struct blob_attr *tb[10];
	int tx_mcs = 0, tx_nss = 0;
	int rx_mcs = 0, rx_nss = 0;
	uint8_t tx_supp_mcs[2] = {0xff, 0xff};
	uint8_t rx_supp_mcs[2] = {0xff, 0xff};
	static const struct blobmsg_policy ap_attr[10] = {
		[0] = { .name = "dot11ac_sgi80", .type = BLOBMSG_TYPE_BOOL },
		[1] = { .name = "dot11ac_sgi160", .type = BLOBMSG_TYPE_BOOL },
		[2] = { .name = "dot11ac_8080", .type = BLOBMSG_TYPE_BOOL },
		[3] = { .name = "dot11ac_160", .type = BLOBMSG_TYPE_BOOL },
		[4] = { .name = "dot11ac_su_beamformer", .type = BLOBMSG_TYPE_BOOL },
		[5] = { .name = "dot11ac_mu_beamformer", .type = BLOBMSG_TYPE_BOOL },
		[6] = { .name = "dot11ac_supp_max_rx_mcs", .type = BLOBMSG_TYPE_INT32 },
		[7] = { .name = "dot11ac_supp_max_rx_nss", .type = BLOBMSG_TYPE_INT32 },
		[8] = { .name = "dot11ac_supp_max_tx_mcs", .type = BLOBMSG_TYPE_INT32 },
		[9] = { .name = "dot11ac_supp_max_tx_nss", .type = BLOBMSG_TYPE_INT32 },
	};

	blobmsg_parse(ap_attr, 10, tb, blobmsg_data(arg), blobmsg_data_len(arg));

	rx_mcs = blobmsg_get_u32(tb[6]);
	rx_nss = blobmsg_get_u32(tb[7]);
	tx_mcs = blobmsg_get_u32(tb[8]);
	tx_nss = blobmsg_get_u32(tb[9]);
	ap->caps.vht[4] |= ((blobmsg_get_bool(tb[0]) ? 1 : 0) << 1);
	ap->caps.vht[4] |= ((blobmsg_get_bool(tb[1]) ? 1 : 0) << 0);
	ap->caps.vht[5] |= ((blobmsg_get_bool(tb[2]) ? 1 : 0) << 7);
	ap->caps.vht[5] |= ((blobmsg_get_bool(tb[3]) ? 1 : 0) << 6);
	ap->caps.vht[5] |= ((blobmsg_get_bool(tb[4]) ? 1 : 0) << 5);
	ap->caps.vht[5] |= ((blobmsg_get_bool(tb[5]) ? 1 : 0) << 4);

	ap->caps.vht[4] |= ((tx_nss - 1) & 0x07) << 5;
	ap->caps.vht[4] |= ((rx_nss - 1) & 0x07) << 2;

	parse_dot11ac_mcs(tx_supp_mcs, tx_mcs, tx_nss);
	memcpy(&ap->caps.vht[0], tx_supp_mcs, 2);
	parse_dot11ac_mcs(rx_supp_mcs, rx_mcs, rx_nss);
	memcpy(&ap->caps.vht[2], rx_supp_mcs, 2);
}

#define HE_CAP_MAX_MCS	8
static void parse_dot11ax_mcs(uint8_t *he, int *he_mcs_len, int mcs,
			int nss, int *max_nss)
{
	int i;
	int octel;
	int shift;
	int offset = 1 + *he_mcs_len;
	uint8_t supp_mcs[2] = {0xff, 0xff};
	uint8_t mask[4] = {0xfc, 0xf3, 0xcf, 0x3f};
	int nss_cnt = min(nss, HE_CAP_MAX_MCS);

	for (i = 0; i < nss_cnt; i++) {
		octel = (2 * i) / 8;
		shift = (2 * i) % 8;

		if (mcs == 11)
			supp_mcs[octel] &= (mask[i%4] | (0x02 << shift));
		else if (mcs == 9)
			supp_mcs[octel] &= (mask[i%4] | (0x01 << shift));
		else if (mcs == 7)
			supp_mcs[octel] &= (mask[i%4] | (0x00 << shift));
	}

	memcpy(he + offset, supp_mcs, 2);
	*he_mcs_len += 2;
	*max_nss = max(*max_nss, nss);
}

static void parse_dot11ax(struct netif_ap *ap, struct blob_attr *arg)
{
	int max_rx_nss = 0;
	int max_tx_nss = 0;
	int mcs_len = 0;
	int offset = 0;
	struct blob_attr *tb[35];
	struct wifi_wifi6_capabilities *wifi6_caps = &ap->caps.wifi6;
	static const struct blobmsg_policy ap_attr[ARRAY_SIZE(tb)] = {
		[0] = { .name = "dot11ax_5g_160_and_8080", .type = BLOBMSG_TYPE_BOOL },
		[1] = { .name = "dot11ax_5g_160", .type = BLOBMSG_TYPE_BOOL },
		[2] = { .name = "dot11ax_su_beamformer", .type = BLOBMSG_TYPE_BOOL },
		[3] = { .name = "dot11ax_mu_beamformer", .type = BLOBMSG_TYPE_BOOL },
		[4] = { .name = "dot11ax_ul_mumimo_full", .type = BLOBMSG_TYPE_BOOL },
		[5] = { .name = "dot11ax_ofdma_ra", .type = BLOBMSG_TYPE_BOOL },
		[6] = { .name = "dot11ax_supp_max_rx_mcs_20", .type = BLOBMSG_TYPE_INT32 },
		[7] = { .name = "dot11ax_supp_max_rx_nss_20", .type = BLOBMSG_TYPE_INT32 },
		[8] = { .name = "dot11ax_supp_max_tx_mcs_20", .type = BLOBMSG_TYPE_INT32 },
		[9] = { .name = "dot11ax_supp_max_tx_nss_20", .type = BLOBMSG_TYPE_INT32 },
		[10] = { .name = "dot11ax_supp_max_rx_mcs_40", .type = BLOBMSG_TYPE_INT32 },
		[11] = { .name = "dot11ax_supp_max_rx_nss_40", .type = BLOBMSG_TYPE_INT32 },
		[12] = { .name = "dot11ax_supp_max_tx_mcs_40", .type = BLOBMSG_TYPE_INT32 },
		[13] = { .name = "dot11ax_supp_max_tx_nss_40", .type = BLOBMSG_TYPE_INT32 },
		[14] = { .name = "dot11ax_supp_max_rx_mcs_80", .type = BLOBMSG_TYPE_INT32 },
		[15] = { .name = "dot11ax_supp_max_rx_nss_80", .type = BLOBMSG_TYPE_INT32 },
		[16] = { .name = "dot11ax_supp_max_tx_mcs_80", .type = BLOBMSG_TYPE_INT32 },
		[17] = { .name = "dot11ax_supp_max_tx_nss_80", .type = BLOBMSG_TYPE_INT32 },
		[18] = { .name = "dot11ax_supp_max_rx_mcs_160", .type = BLOBMSG_TYPE_INT32 },
		[19] = { .name = "dot11ax_supp_max_rx_nss_160", .type = BLOBMSG_TYPE_INT32 },
		[20] = { .name = "dot11ax_supp_max_tx_mcs_160", .type = BLOBMSG_TYPE_INT32 },
		[21] = { .name = "dot11ax_supp_max_tx_nss_160", .type = BLOBMSG_TYPE_INT32 },
		[22] = { .name = "dot11ax_supp_max_rx_mcs_8080", .type = BLOBMSG_TYPE_INT32 },
		[23] = { .name = "dot11ax_supp_max_rx_nss_8080", .type = BLOBMSG_TYPE_INT32 },
		[24] = { .name = "dot11ax_supp_max_tx_mcs_8080", .type = BLOBMSG_TYPE_INT32 },
		[25] = { .name = "dot11ax_supp_max_tx_nss_8080", .type = BLOBMSG_TYPE_INT32 },
		[26] = { .name = "dot11ax_su_beamformee", .type = BLOBMSG_TYPE_BOOL },
		[27] = { .name = "dot11ax_twt_requester", .type = BLOBMSG_TYPE_BOOL },
		[28] = { .name = "dot11ax_twt_responder", .type = BLOBMSG_TYPE_BOOL },
		[29] = { .name = "dot11ax_bfe_sts_le_80", .type = BLOBMSG_TYPE_BOOL },
		[30] = { .name = "dot11ax_bfe_sts_gt_80", .type = BLOBMSG_TYPE_BOOL },
		[31] = { .name = "dot11ax_rts", .type = BLOBMSG_TYPE_BOOL },
		[32] = { .name = "dot11ax_mu_rts", .type = BLOBMSG_TYPE_BOOL },
		[33] = { .name = "dot11ax_multi_bssid", .type = BLOBMSG_TYPE_BOOL },
		[34] = { .name = "dot11ax_mu_edca", .type = BLOBMSG_TYPE_BOOL },

	};

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blobmsg_data(arg), blobmsg_data_len(arg));

	if (tb[6])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[6]),
				blobmsg_get_u32(tb[7]), &max_rx_nss);
	else if (tb[10])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[10]),
				blobmsg_get_u32(tb[11]), &max_rx_nss);
	else if (tb[14])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[14]),
				blobmsg_get_u32(tb[15]), &max_rx_nss);

	if (tb[8])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[8]),
				blobmsg_get_u32(tb[9]), &max_tx_nss);
	else if (tb[12])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[12]),
				blobmsg_get_u32(tb[13]), &max_tx_nss);
	else if (tb[16])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[16]),
				blobmsg_get_u32(tb[17]), &max_tx_nss);

	if (tb[18])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[18]),
				blobmsg_get_u32(tb[19]), &max_rx_nss);

	if (tb[20])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[20]),
				blobmsg_get_u32(tb[21]), &max_tx_nss);

	if (tb[22])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[22]),
				blobmsg_get_u32(tb[23]), &max_rx_nss);

	if (tb[24])
		parse_dot11ax_mcs(ap->caps.he, &mcs_len,
				blobmsg_get_u32(tb[24]),
				blobmsg_get_u32(tb[25]), &max_tx_nss);

	ap->caps.he[0] = mcs_len;
	offset = 1 + mcs_len;
	ap->caps.he[offset] |= (((max_tx_nss - 1) & 0x07) << 5);
	ap->caps.he[offset] |= (((max_rx_nss - 1) & 0x07) << 2);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[0]) ? 1 : 0) << 1);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[1]) ? 1 : 0) << 0);

	offset++;
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[2]) ? 1 : 0) << 7);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[3]) ? 1 : 0) << 6);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[4]) ? 1 : 0) << 5);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[5]) ? 1 : 0) << 4);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[5]) ? 1 : 0) << 3);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[5]) ? 1 : 0) << 2);
	ap->caps.he[offset] |= ((blobmsg_get_bool(tb[5]) ? 1 : 0) << 1);

	/* Populate Wi-Fi 6 caps structure */
	/* HE 160 supported when HE160 TX & RX supported */
	wifi6_caps->he160 = tb[19] && blobmsg_get_u32(tb[19]) &&
			    tb[21] && blobmsg_get_u32(tb[21]);

	wifi6_caps->he8080 = tb[23] && blobmsg_get_u32(tb[23]) &&
			     tb[25] && blobmsg_get_u32(tb[25]);

	/* 4, 8 or 12 bytes of supported MCS & NSS */
	wifi6_caps->mcs_nss_len = mcs_len;
	memcpy(wifi6_caps->mcs_nss_12, &ap->caps.he[1], mcs_len);

	wifi6_caps->su_beamformer = tb[2] && blobmsg_get_bool(tb[2]);
	wifi6_caps->su_beamformee = tb[26] && blobmsg_get_bool(tb[26]);
	wifi6_caps->mu_beamformer = tb[3] && blobmsg_get_bool(tb[3]);
	wifi6_caps->beamformee_le80 = tb[29] && blobmsg_get_bool(tb[29]);
	wifi6_caps->beamformee_gt80 = tb[30] && blobmsg_get_bool(tb[30]);
	wifi6_caps->ul_mumimo = tb[4] && blobmsg_get_bool(tb[4]);
	wifi6_caps->ul_ofdma = tb[5] && blobmsg_get_bool(tb[5]);
	wifi6_caps->dl_ofdma = tb[5] && blobmsg_get_bool(tb[5]);
	wifi6_caps->max_dl_mumimo = 0;       // todo: ?
	wifi6_caps->max_ul_mumimo = 0;       // todo: ?
	wifi6_caps->max_dl_ofdma = 0;        // todo: ?
	wifi6_caps->max_ul_ofdma = 0;        // todo: ?
	wifi6_caps->rts = tb[31] && blobmsg_get_bool(tb[31]);
	wifi6_caps->mu_rts = tb[32] && blobmsg_get_bool(tb[32]);
	wifi6_caps->multi_bssid = tb[33] && blobmsg_get_bool(tb[33]);
	wifi6_caps->mu_edca = tb[34] && blobmsg_get_bool(tb[34]);
	wifi6_caps->twt_requester = tb[27] && blobmsg_get_bool(tb[27]);
	wifi6_caps->twt_responder = tb[28] && blobmsg_get_bool(tb[28]);
	wifi6_caps->spatial_reuse = false;   // todo: ?
	wifi6_caps->anticipated_ch_usage = false; // todo: ?
}

void parse_bk(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct netif_bk *bk = (struct netif_bk *)req->priv;
	struct agent *a = bk->agent;
	char bssid_str[18] = {0}, macaddr[18] = {0};
	uint8_t bssid[6] = {0};
	uint32_t freq = 0;
	static const struct blobmsg_policy ap_attr[] = {
		[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 }
	};
	struct blob_attr *tb[ARRAY_SIZE(ap_attr)];

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blob_data(msg), blob_len(msg));

	if (!tb[0])
		return;

	strncpy(macaddr, blobmsg_data(tb[0]), 17);
	hwaddr_aton(macaddr, bk->macaddr);

	if (tb[1]) {
		strncpy(bssid_str, blobmsg_data(tb[1]), 17);
		hwaddr_aton(bssid_str, bssid);
	}

	if (tb[2])
		freq = blobmsg_get_u32(tb[2]);

	if (!hwaddr_is_zero(bssid) && memcmp(bk->macaddr, bssid, 6)) {
		dynbh_bsta_disable_lower_priority(a, bk->ifname);

		/* if connected and previously unconnected */
		if (!bk->connected) {
			wifi_bsta_connect(a, bk, bssid, freq);
			timer_set(&a->disable_unconnected_bstas_scheduler, 0);
			timer_del(&a->bh_lost_timer);
			timer_del(&a->bh_reconf_timer);
		}
	} else {
		bk->connected = false;
		/* if disconnected - check if it is a part of the bridge */
		if (if_isbridge_interface(bk->ifname)) {
			wifi_mod_bridge(a, bk->ifname, "remove");

			if (bk->cfg->enabled) {
				char ul_ifname[16] = {0};

				dynbh_handle_bh_lost(a, bk);

				if (agent_get_backhaul_ifname(a, ul_ifname)) {
					if (!strncmp(ul_ifname, bk->ifname, IFNAMSIZ))
						agent_exec_platform_scripts("unset_uplink wifi");
				}
			}
		}
	}

}

static void parse_esp_params(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct wifi_bss_element *bss = (struct wifi_bss_element *)req->priv;
	int rem;
	struct blob_attr *cur;
	struct blob_attr *data[1];
	static const struct blobmsg_policy beacon_attr[1] = {
		[0] = { .name = "beacon-ies", .type = BLOBMSG_TYPE_ARRAY },
	};

	blobmsg_parse(beacon_attr, 1, data, blob_data(msg), blob_len(msg));

	if (!data[0])
		return;

	blobmsg_for_each_attr(cur, data[0], rem) {
		char *iestr = strdup(blobmsg_get_string(cur));

		if (!iestr)
			return;

		/*
		 * esp string having this format: "ff xx 0b"
		 * where xx: 4 or 7 or 10 or 13
		 */
		if ((!strncmp(iestr, "ff", 2)) &&
			(!strncmp(iestr+4, "0b", 2))) {
			uint8_t *ie, *ie1;
			int i, slen, esp_len;
			int ac;		/* Access Category */

			slen = strlen(iestr);
			ie = calloc(slen/2, sizeof(uint8_t));
			if (!ie) {
				free(iestr);
				return;
			}

			strtob(iestr, slen, ie);
			esp_len = ie[1];
			/*
			 * esp data payload start after 3 bytes;
			 * bypass 3 bytes, (1btye: ff, 1 byte: len, 1byte:0b)
			 */
			ie1 = ie + 3;
			for (i = 1; i < esp_len; i += 3) {
				ac = (*ie1) & 0x03;
				switch (ac) {
				case BK:
					bss->is_ac_bk = 1;
					memcpy(bss->est_wmm_bk, ie1, 3);
					break;
				case BE:
					bss->is_ac_be = 1;
					memcpy(bss->est_wmm_be, ie1, 3);
					break;
				case VI:
					bss->is_ac_vi = 1;
					memcpy(bss->est_wmm_vi, ie1, 3);
					break;
				case VO:
					bss->is_ac_vo = 1;
					memcpy(bss->est_wmm_vo, ie1, 3);
					break;
				}
				ie1 += 3;
			}

			free(ie);
			free(iestr);
			return;
		}

		free(iestr);
	}
}

static void parse_ap_stats(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	int bss_index;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;
	struct wifi_bss_element *bsslist;
	struct netif_ap *ap = (struct netif_ap *)req->priv;
	struct blob_attr *tb[6];
	static const struct blobmsg_policy ap_stats_attr[6] = {
		[0] = { .name = "tx_unicast_packets", .type = BLOBMSG_TYPE_INT64},
		[1] = { .name = "rx_unicast_packets", .type = BLOBMSG_TYPE_INT64},
		[2] = { .name = "tx_multicast_packets", .type = BLOBMSG_TYPE_INT64},
		[3] = { .name = "rx_multicast_packets", .type = BLOBMSG_TYPE_INT64},
		[4] = { .name = "tx_broadcast_packets", .type = BLOBMSG_TYPE_INT64},
		[5] = { .name = "rx_broadcast_packets", .type = BLOBMSG_TYPE_INT64},
	};


	radio = wifi_ifname_to_radio_element(ap->agent, ap->ifname);
	if (!radio)
		return;

	bsslist = (struct wifi_bss_element *)realloc(radio->bsslist,
			(radio->num_bss + 1) * sizeof(struct wifi_bss_element));

	if (bsslist) {
		radio->bsslist = bsslist;
		radio->num_bss++;
	} else
		return;

	bss_index = radio->num_bss - 1;
	bss = radio->bsslist + bss_index;
	memset(bss, 0, sizeof(struct wifi_bss_element));
	memcpy(bss->bssid, ap->bssid, 6);
	strncpy(bss->ssid, ap->ssid, sizeof(bss->ssid) - 1);
	bss->enabled = ap->enabled;

	blobmsg_parse(ap_stats_attr, 6, tb, blob_data(msg), blob_len(msg));

	if (tb[0])
		bss->tx_ucast_bytes = blobmsg_get_u64(tb[0]);

	if (tb[1])
		bss->rx_ucast_bytes = blobmsg_get_u64(tb[1]);

	if (tb[2])
		bss->tx_mcast_bytes = blobmsg_get_u64(tb[2]);

	if (tb[3])
		bss->rx_mcast_bytes = blobmsg_get_u64(tb[3]);

	if (tb[4])
		bss->tx_bcast_bytes = blobmsg_get_u64(tb[4]);

	if (tb[5])
		bss->rx_bcast_bytes = blobmsg_get_u64(tb[5]);

	/* Fill ESP params for bss */
	ubus_call_object(ap->agent, ap->wifi, "dump_beacon",
			parse_esp_params, bss);
}

static void parse_ap(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct netif_ap *ap = (struct netif_ap *)req->priv;
	char bssid[18] = { 0 }, ifname[16] = { 0 };
	struct blob_attr *tb[10];
	static const struct blobmsg_policy ap_attr[10] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "capabilities", .type = BLOBMSG_TYPE_TABLE },
		[3] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[4] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "standard", .type = BLOBMSG_TYPE_STRING },
		[6] = { .name = "bandwidth", .type = BLOBMSG_TYPE_INT32 },
		[7] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		[8] = { .name = "num_stations", .type = BLOBMSG_TYPE_INT32 },
		[9] = { .name = "enabled", .type = BLOBMSG_TYPE_BOOL },
	};

	blobmsg_parse(ap_attr, 10, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(ifname, blobmsg_data(tb[0]), 15);
	strncpy(ap->ifname, blobmsg_data(tb[0]), 15);

	strncpy(bssid, blobmsg_data(tb[1]), 17);
	if (!hwaddr_aton(bssid, ap->bssid))
		return;

	if (tb[2]) { /* capabilities */
		struct blob_attr *data[4];
		static const struct blobmsg_policy cap_attr[4] = {
			[0] = { .name = "dot11n", .type = BLOBMSG_TYPE_TABLE },
			[1] = { .name = "dot11ac", .type = BLOBMSG_TYPE_TABLE },
			[2] = { .name = "wmm", .type = BLOBMSG_TYPE_BOOL },
			[3] = { .name = "dot11ax", .type = BLOBMSG_TYPE_TABLE }
		};

		blobmsg_parse(cap_attr, 4, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (data[0])
			parse_dot11n(ap, data[0]);
		if (data[1])
			parse_dot11ac(ap, data[1]);
		if (data[2])
			ap->caps.wmm = blobmsg_get_bool(data[2]);
		if (data[3])
			parse_dot11ax(ap, data[3]);
	}

	if (tb[3])
		ap->channel = blobmsg_get_u32(tb[3]);

	if (tb[4])
		strncpy(ap->ssid, blobmsg_data(tb[4]), sizeof(ap->ssid) - 1);

	if (tb[5])
		strncpy(ap->standard, blobmsg_data(tb[5]), sizeof(ap->standard) - 1);

	if (tb[6])
		ap->bandwidth = blobmsg_get_u32(tb[6]);

	if (tb[9])
		ap->enabled = blobmsg_get_bool(tb[9]);

	if (tb[8])
		ap->nbr_sta = blobmsg_get_u32(tb[8]);


	if (!ap->enabled && may_enable_fhs(ap->agent))
		/* Enable fronthauls */
		timer_set(&ap->agent->enable_fhs_scheduler,
				  (IFACE_TIMEOUT + 1) * 1000);

}

static int agent_radio_update_scanlist(struct agent *a,
		struct wifi_radio_element *re, struct wifi_bss *bsss, int bss_num)
{
	int i, idx;
	struct wifi_scanres_channel_element *scanres_el;

	for (i = 0; i < bss_num; i++) {
		uint8_t classid = wifi_opclass_find_id_from_channel(
					&re->opclass,
					bsss[i].channel,
					20); /* caps */

		scanres_el = wifi_get_scanres_ch_element(re,
							classid, bsss[i].channel);
		if (!scanres_el)
			continue;

		// no neighbors, allocate first
		if (!scanres_el->num_neighbors) {
			scanres_el->nbrlist = calloc(1, sizeof(*(scanres_el->nbrlist)));
			if (!scanres_el->nbrlist)
				continue;

			scanres_el->num_neighbors = 1;
		}
		// at least one neighbor, allocate one more
		else {
			struct wifi_scanres_neighbor_element *nbr;
			int size;

			size = (scanres_el->num_neighbors + 1) * sizeof(*nbr);
			nbr = (struct wifi_scanres_neighbor_element *)realloc(
						scanres_el->nbrlist, size);

			if (!nbr)
				continue;

			scanres_el->num_neighbors++;
			scanres_el->nbrlist = nbr;
		}

		idx = scanres_el->num_neighbors - 1; // fill index

		/* Fill in the neighbour data */
		memcpy(scanres_el->nbrlist[idx].bssid, bsss[i].bssid, 6);

		strncpy(scanres_el->nbrlist[idx].ssid, (char *) bsss[i].ssid,
				sizeof(scanres_el->nbrlist->ssid) - 1);

		scanres_el->nbrlist[idx].rssi = bsss[i].rssi;
		scanres_el->nbrlist[idx].bw = wifi_bw_to_bw(bsss[i].curr_bw);
		scanres_el->nbrlist[idx].utilization = bsss[i].load.utilization;
		scanres_el->nbrlist[idx].num_stations = bsss[i].load.sta_count;
	}

	return 0;
}

int agent_radio_scanresults(struct agent *a, struct wifi_radio_element *re)
{
	struct wifi_bss bss[128];
	int bss_num = ARRAY_SIZE(bss);
	int ret;

	if (!re || !strlen(re->name))
		return -1;

	trace("[%s] radio scanresults\n", re->name);

	/* Get scan results from the driver */
	ret = wifi_get_scan_results(re->name, bss, &bss_num);

	if (ret) {
		dbg("[%s] failed to get scanresults\n", re->name);
		return -1;
	}

	/* Update scan cache */
	ret = wifi_scanresults_add(&re->scanresults, &re->opclass, bss, bss_num);
	if (ret) {
		dbg("[%s] failed to update scan cache\n", re->name);
		return -1;
	}

	/* TODO: update scanlist using fresh results from the cache */

	/* Empty scanlist & add most recent results for reporting */
	free_scanlist_neighbors(re);
	ret = agent_radio_update_scanlist(a, re, bss, bss_num);

	return ret;
}

void agent_init_interfaces_post_actions(struct agent *a)
{
	trace("%s: --->\n", __func__);

	struct netif_ap *p = NULL;
	uint8_t sta[128 * 6] = {};
	int num_sta = 128;
	int i;

	list_for_each_entry(p, &a->aplist, list) {
		num_sta = 128;
		if (wifi_get_assoclist(p->ifname, sta, &num_sta))
			continue;

		for (i = 0; i < num_sta; i++)
			wifi_add_sta(a, p->ifname, &sta[i * 6]);
	}

	/* request scan results from the driver, etc */
	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		struct wifi_radio_element *re = &a->radios[i];

		if (re && re->name[0] != '\0' && strlen(re->name)) {
			/* Make sure scanlist is initialized */
			if (!re->scanlist)
				init_scanlist(a, re->name);
			/* Get & store scan results */
			agent_radio_scanresults(a, re);
			/* Finally update neighbor data */
			update_neighbors_from_scancache(a, &re->scanresults);
		}
	}
}

int agent_get_system_info(struct agent *a)
{

	char buf[68] = {0};

	chrCmd(buf, sizeof(buf), "/lib/wifi/multiap db_get device.deviceinfo.SerialNumber");

	strncpy(a->device_inventory.serial_number, buf,
		sizeof(a->device_inventory.serial_number) - 1);

	chrCmd(buf, sizeof(buf), "/lib/wifi/multiap db_get device.deviceinfo.SoftwareVersion");

	strncpy(a->device_inventory.sw_version, buf,
		sizeof(a->device_inventory.sw_version) - 1);

	dbg("|%s:%d| serial number:%s software version:%s\n", __func__, __LINE__, a->device_inventory.serial_number,
		a->device_inventory.sw_version);
	return 0;
}

/* Initialize netif_ap/bk structs from ubus wifi objects */
int agent_init_interfaces(struct agent *a)
{
	struct agent_config *cfg = &a->cfg;
	struct netif_apcfg *f;
	struct netif_bkcfg *b;
#if (EASYMESH_VERSION > 2)
	struct dpp_uri_cfg *u;
#endif
	wifi_object_t wifi_obj;
	uint32_t ieee1905_obj = 0;
	int ret, retry = 0;

	ret = ubus_lookup_id(a->ubus_ctx, "ieee1905", &ieee1905_obj);
	if (ret)
		return -1;

	ubus_call_object(a, ieee1905_obj, "info", parse_i1905_info, a);

	wifi_obj = ubus_get_object(a->ubus_ctx, "wifi");
	while (wifi_obj == WIFI_OBJECT_INVALID && retry < 5) {
		err("|%s:%d| Failed to get wifi object (%u) retry in 0.5s, try num:%d\n", __func__, __LINE__, wifi_obj, retry);
		usleep(500 * 1000); /* 0.5s sleep */
		wifi_obj = ubus_get_object(a->ubus_ctx, "wifi");
		retry++;
	}

	if (wifi_obj == WIFI_OBJECT_INVALID) {
		warn("|%s:%d| Object 'wifi' not present!\n", __func__, __LINE__);
		timer_set(&a->init_ifaces_scheduler, 30 * 1000);
		return -1;
	}

	a->wifi = wifi_obj;
	enumerate_wifi_objects(a);

	a->connected = false;

	list_for_each_entry(f, &cfg->aplist, list) {
		wifi_object_t r_wobj = WIFI_OBJECT_INVALID;
		wifi_object_t wobj = WIFI_OBJECT_INVALID;
		//const char *r_fmt = "wifi.radio.%s";
		//const char *fmt = "wifi.ap.%s";
		const char *radio_name = NULL;
		struct netif_ap *fn = NULL;
		char r_objname[32] = {0};
		char objname[32] = {0};
		struct subscribe_fr {
			uint8_t type;
			uint8_t stype;
		} subfr[] = {{0, 0}, {0, 2}, {0, 13}};
		int num_subfr = ARRAY_SIZE(subfr);
		int k;

		radio_name = wifi_ifname_to_radio(a, f->name);
		if (!radio_name)
			continue;

		fn = netif_alloc_ap(f->name);
		if (fn) {
			/* subscribe following frames
			 * WIFI_FRAME_ASSOC_REQ,
			 * WIFI_FRAME_REASSOC_REQ,
			 * WIFI_FRAME_ACTION
			 */
			// TODO: call unsubscribe_frame method on cleanup.

			fn->wifi = wobj;
			fn->radio = r_wobj;
			fn->cfg = f;
			fn->agent = a;
			strncpy(fn->radio_name, radio_name, IFNAMSIZ-1);

			list_add(&fn->list, &a->aplist);
		}

		snprintf(r_objname, 31, "wifi.radio.%s", radio_name);
		fn->radio = ubus_get_object(a->ubus_ctx, r_objname);
		if (fn->radio == WIFI_OBJECT_INVALID) {
			dbg("%s not present! skipping '%s' from config\n",
						r_objname, f->name);
			continue;
		}

		snprintf(objname, 31, "wifi.ap.%s", f->name);
		fn->wifi = ubus_get_object(a->ubus_ctx, objname);
		if (fn->wifi == WIFI_OBJECT_INVALID) {
			dbg("%s not present! skipping '%s' from config\n",
							objname, f->name);
			continue;
		}

		ubus_call_object(a, fn->wifi, "status", parse_ap, fn);

		for (k = 0; k < num_subfr; k++)
			wifi_subscribe_frame(f->name, subfr[k].type, subfr[k].stype);

		ubus_call_object(a, fn->wifi, "stats", parse_ap_stats, fn);
	}

	list_for_each_entry(b, &cfg->bklist, list) {
		wifi_object_t r_wobj = WIFI_OBJECT_INVALID;
		wifi_object_t wobj = WIFI_OBJECT_INVALID;
		//const char *r_fmt = "wifi.radio.%s";
		//const char *bk_fmt = "wifi.backhaul.%s";
		const char *radio_name = NULL;
		struct netif_bk *bn = NULL;
		char r_objname[32] = {0};
		char objname[32] = {0};


		bn = find_bkhaul_by_ifname(a, b->name);
		if (!bn) {
			bn = netif_alloc_bk(b->name);
			if (bn) {
				list_add(&bn->list, &a->bklist);
				bn->cfg = b;
				bn->agent = a;
			}
		} else {
			bn->cfg = b;
			bn->agent = a;
			memset(bn->macaddr, 0, 6);
		}

		radio_name = wifi_ifname_to_radio(a, b->name);
		if (!radio_name)
			continue;

		snprintf(r_objname, 31, "wifi.radio.%s", radio_name);
		r_wobj = ubus_get_object(a->ubus_ctx, r_objname);
		if (r_wobj == WIFI_OBJECT_INVALID) {
			dbg("%s not present! skipping '%s' from config\n",
						r_objname, f->name);
			continue;
		}

		snprintf(objname, 31, "wifi.backhaul.%s", b->name);
		wobj = ubus_get_object(a->ubus_ctx, objname);
		if (wobj == WIFI_OBJECT_INVALID) {
			dbg("%s not present! skipping '%s' from config\n",
							objname, b->name);
			continue;
		}

		if (bn) {
			bn->wifi = wobj;
			bn->radio = r_wobj;

			ubus_call_object(a, wobj, "status", parse_bk, bn);
			if (bn->connected)
				a->connected = true;
		}

		if (b->onboarded) {
			struct wifi_radio_element *re;

			re = wifi_ifname_to_radio_element(a, b->name);
			if (!re)
				continue;

			re->onboarded = 1;
			trace("radio (%s) has been onboarded\n", re->name);
		}
	}

#if (EASYMESH_VERSION > 2)
	list_for_each_entry(u, &cfg->urilist, list) {
		struct netif_bk *bn;

		bn = find_bkhaul_by_ifname(a, u->ifname);
		if (bn)
			bn->dpp_cfg = u;

		if (timer_remaining_ms(&bn->dpp_do_chirp) == -1) {
			timer_set(&bn->dpp_do_chirp, 0);
		}
	}
#endif

	if (!agent_has_active_backhaul(a)
	    && timer_remaining_ms(&a->bh_lost_timer) == -1
	    && !list_empty(&a->bklist) && !list_empty(&cfg->bklist)) {
		if (a->cfg.ap_follow_sta_dfs) {
			struct netif_bk *sta = NULL;
			struct dyn_bh_cfg *dbcfg = &a->cfg.dbhcfg;

			list_for_each_entry(sta, &a->bklist, list) {
				dbg("[%s] connect %s bsta\n", sta->ifname, sta->cac_time ? "disable" : "enable");
				if (sta->cac_time)
					config_disable_bsta(sta->cfg);
				else
					config_enable_bsta(sta->cfg);
			}

			dynbh_bsta_scan_on_enabled(a);
			if (dbcfg->bh_miss_tmo)
				/* Trigger bh link loss timer */
				timer_set(&a->bh_lost_timer, dbcfg->bh_miss_tmo * 1000);

			timestamp_update(&a->disconnect_t);
		} else {
			struct dyn_bh_cfg *dbcfg = &a->cfg.dbhcfg;

			dynbh_bsta_enable_all(a);
			if (dbcfg) {
				timer_set(&a->bh_reconf_timer,
					  dbcfg->bh_reconf_tmo * 1000);
				timestamp_update(&a->disconnect_t);
			}
		}
	} else if (!list_empty(&a->bklist) && !list_empty(&cfg->bklist)) {
		timer_set(&a->dynbh_scheduler, 30 * 1000);
	}

	agent_config_reload(a);

	agent_init_interfaces_post_actions(a);
	if (cfg && cfg->pcfg && is_vid_valid(cfg->pcfg->pvid)) {
		a->reconfig_reason |= AGENT_RECONFIG_REASON_VLAN_SETUP;
		timer_set(&a->reload_scheduler, 5 * 1000);
	}
	return 0;
}

static void agent_enable_fhs_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, enable_fhs_scheduler);

	agent_enable_fronthauls(a);
}

static void agent_boot_scan_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, boot_scan_scheduler);
	bool radio_missing = false;
	int i;

	if (!a)
		return; /* err */

	dbg("|%s:%d| boot scan %s", __func__, __LINE__,
	    a->cfg.scan_on_boot_only ? "enabled" : "disabled");

	if (!a->cfg.scan_on_boot_only)
		/* On boot scan disabled */
		return;

	for (i = 0; i < a->num_radios; i++) {
		struct wifi_radio_element *re = &a->radios[i];

		if (!re || !re->enabled) {
			radio_missing = true;
			continue;
		}

		if (re->scan_state != SCAN_DONE)
			wifi_radio_scan_req_all(a, re->name);
	}

	if (a->boot_scan_tries++ < BOOT_UP_SCAN_MAX_TRY
			&& radio_missing)
		/* one or more radio not available yet, re-try again later */
		timer_set(&a->boot_scan_scheduler, BOOT_UP_SCAN_ITV * 1000);
}


#define OPER_CHAN_REP_INTERVAL 60
static void agent_periodic_run(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, agent_periodic_run);

	a->uptime += 1;

	trace("|%s:%d| periodic time (elapsed:%"PRIu64")\n", __func__, __LINE__,
	      a->uptime);

	if ((a->uptime % OPER_CHAN_REP_INTERVAL) == 0)
		send_oper_channel_report(a, NULL);

	timer_set(&a->agent_periodic_run, 1 * 1000);
}


static void agent_radio_stats_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	int i;
	struct agent *a = container_of(t, struct agent, radio_stats_scheduler);

	for (i = 0; i < a->num_radios; i++) {
		char r_objname[32] = {0};
		wifi_object_t r_wobj = WIFI_OBJECT_INVALID;
		struct wifi_radio_element *r = &a->radios[i];

		/* mark outdated scanresults */
		wifi_scanresults_mark_expired(&r->scanresults);

		/* get radio status and update stats */
		snprintf(r_objname, 31, "wifi.radio.%s", r->name);
		r_wobj = ubus_get_object(a->ubus_ctx, r_objname);

		if (r_wobj == WIFI_OBJECT_INVALID)
			continue;

		ubus_call_object(a, r_wobj, "status", parse_radio, r);
	}

	timer_set(&a->radio_stats_scheduler, RADIO_STATS_TIMER);
}

static void agent_init_ifaces_cb(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, init_ifaces_scheduler);

	if (!a)
		return;

	agent_init_interfaces(a);
}


#define RELOAD_TIMEOUT 10

static void agent_reload_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);
	struct agent *a = container_of(t, struct agent, reload_scheduler);

	if (a->cfg.eth_onboards_wifi_bhs) {
		/* re-sync any credentials passed with AP-Autoconfig */
		agent_exec_platform_scripts("bsta_to_wireless");
	}

	agent_exec_platform_scripts("sync_credentials");

	dbg("|%s:%d| reconfig_reason = %d\n", __func__, __LINE__, a->reconfig_reason);

	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_VLAN_TEARDOWN)) {
		dbg("|%s:%d| ts teardown\n", __func__, __LINE__);
		agent_clear_traffic_sep(a);
	}

	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_VLAN_SETUP)) {
		dbg("|%s:%d| ts setup\n", __func__, __LINE__);
		agent_apply_traffic_separation(a);
	}

	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_AP_AUTOCONF)) {
		dbg("|%s:%d| apconf reload wireless\n", __func__, __LINE__);
		uci_reload_services("wireless");

#ifdef WIRELESS_EXPLICIT_RELOAD
		agent_exec_platform_scripts("wireless_reload");
#endif
	}

	a->reconfig_reason = 0;
}

/* send the steering compled message this function also resets the value of
 * steering opportunity
 */
static void agent_steering_opp_timeout(atimer_t *t)
{
	trace("agent: %s: --->\n", __func__);
	struct agent *a = container_of(t, struct agent, sta_steer_req_timer);
	uint8_t origin[6] = { 0 };


	memcpy(origin, a->cntlr_almac, 6);
	send_sta_steer_complete((void *)a, origin);
}

#define max_val(a,b) \
	({ typeof(a) _a = (a); \
	typeof(b) _b = (b); \
	_a > _b ? _a : _b; })

static void agent_start_controller_conditionally(struct agent *a)
{
	uint8_t max_retries = a->cntlr_select.retry_int;
	uint16_t probe_int = a->cntlr_select.probe_int;
	int try = a->cntlr_miss_count + 1;
	int max_misstime = max_val((max_retries * probe_int),
				HEARTBEAT_AUTOCFG_INTERVAL + 5);
	bool start_controller = false;

	trace("|%s:%d| Entry max_misstime %d\n", __func__, __LINE__, max_misstime);

	if (!timestamp_expired(&a->observed_time, (try * probe_int) * 1000)
			|| is_local_cntlr_running()) {
		dbg("|%s:%d| Active controller available\n",
			__func__, __LINE__);
		a->cntlr_miss_count = 0;
		a->active_cntlr = true;
		return;
	}

	if (!timestamp_expired(&a->observed_time, max_misstime * 1000)) {
		a->cntlr_miss_count++;
		return;
	}

	a->active_cntlr = false;

	memset(a->cntlr_almac, 0, sizeof(a->cntlr_almac));

	dbg("|%s:%d| No controller observed for over %d seconds\n",
	    __func__, __LINE__, max_misstime);

#ifdef AGENT_DYNAMIC_CNTRL
	if (a->cntlr_select.autostart)
		start_controller = true;
#endif

	if (start_controller) {
		if (is_local_cntlr_available()) {
			time_t now;
			/* Schedule start of local Controller after
			 * random time chosen from a 10s time window
			 * to reduce chance of multiple MAP-Agents
			 * starting own controller at the same time.
			 */
			srand(time(&now));
			agent_schedule_cntlr_start(a, rand() % 10);
		} else {
			warn("Local cntlr unavailable: skip start\n");
		}
	} else if (a->cntlr_miss_count) {
		/* Report controller absence to the user first time
		 * it goes missing
		 */
		wifiagent_log_cntlrinfo(a);
	}

	a->cntlr_miss_count = 0;
}


/* TODO: is it possible to pass private 'reason'? i.e. switch case (reason) */
static void agent_dispatch_autoconfig(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, autocfg_dispatcher);
	int i;

	trace("|%s:%d| Triggering AP-Autoconfig Search\n", __func__, __LINE__);

	agent_start_controller_conditionally(a);
	a->autocfg_interval = a->cntlr_select.probe_int;

	if (!a->active_cntlr) {
		/* Cntlr missing - fall back to active autoconfig probing */
		trace("|%s:%d| No active Controller present!\n", __func__, __LINE__);
		for (i = 0; i < a->num_radios; i++)
			a->radios[i].state = AUTOCFG_ACTIVE;
	}

	agent_check_bsta_connections(a);

	for (i = 0; i < a->num_radios; i++) {
		struct cmdu_buff *cmdu;
		struct wifi_radio_element *radio = &a->radios[i];
		//int mid;

		cmdu = agent_gen_ap_autoconfig_search(a, radio, 0x02);
		if (!cmdu)
			continue;
		trace("|%s:%d| Sending Autoconfig Search for radio %s(%s)\n",
				__func__, __LINE__, radio->name,
				(radio->band == BAND_2 ? "2.4GHz" : "5GHz"));
		radio->mid = agent_send_cmdu(a, cmdu);

		cmdu_free(cmdu);
	}

	dbg("|%s:%d| Scheduling next autoconfig search in %u seconds\n",
				__func__, __LINE__, a->autocfg_interval);
	timer_set(&a->autocfg_dispatcher, a->autocfg_interval * 1000);
}

static void agent_metric_report_timer_cb(atimer_t *t)
{
	struct agent_config *cfg =
		container_of(t, struct agent_config, metric_report_timer);
	struct agent *a = container_of(cfg, struct agent, cfg);
	struct cmdu_buff *cmdu;
	char *ifname = NULL;
	struct node *n;

	n = agent_find_node(a, a->almac);
	if (!n)
		goto refresh_interval;

	cmdu = prepare_ap_metrics_query((void *)a, ifname);
	if (!cmdu)
		goto refresh_interval;

	send_ap_metrics_response(a, cmdu, n);
	cmdu_free(cmdu);
refresh_interval:
	timer_set(&cfg->metric_report_timer,
			cfg->pcfg->report_interval * 1000);
}

void clear_stalist(struct netif_ap *p)
{
	struct sta *s, *tmp;

	list_for_each_entry_safe(s, tmp, &p->stalist, list) {
		dbg("Delete STA " MACFMT "\n", MAC2STR(s->macaddr));
		timer_del(&s->sta_finalize_timer);
		timer_del(&s->sta_steer_timer);
		timer_del(&s->sta_bcn_req_timer);
		timer_del(&s->sta_timer);
		list_del(&s->list);
		list_flush(&s->pref_nbrlist, struct pref_neighbor, list);
		list_flush(&s->sta_nbrlist, struct sta_neighbor, list);
		if (s->assoc_frame) {
			free(s->assoc_frame->frame);
			free(s->assoc_frame);
		}
		free(s);
	}
}

void clear_nbrlist(struct netif_ap *p)
{
	struct neighbor *n, *tmp;

	list_for_each_entry_safe(n, tmp, &p->nbrlist, list) {
		dbg("Delete NBR " MACFMT "\n", MAC2STR(n->nbr.bssid));
		list_del(&n->list);
		free(n);
	}
}

void clear_restrict_stalist(struct netif_ap *p)
{
	struct neighbor *n, *tmp;

	list_for_each_entry_safe(n, tmp, &p->restrict_stalist, list) {
		list_del(&n->list);
		free(n);
	}
}

static void netif_free(struct agent *a, struct netif_ap *n)
{
	/* clear stalist */
	clear_stalist(n);

	/* clear nbrlist */
	clear_nbrlist(n);

	/* clear restricted sta list */
	clear_restrict_stalist(n);

	/* cancel timers */
	timer_del(&n->nbr_timer);
	timer_del(&n->bss_timer);
	timer_del(&n->util_threshold_timer);
	timer_del(&n->una_sta_meas_timer);

	list_del(&n->list);

	free(n);
}

void clear_aplist(struct agent *a)
{
	struct netif_ap *p, *tmp;

	list_for_each_entry_safe(p, tmp, &a->aplist, list) {
		netif_free(a, p);
	}
}

void clear_bklist(struct agent *a)
{
	struct netif_bk *p, *tmp;

	list_for_each_entry_safe(p, tmp,  &a->bklist, list) {
		list_del(&p->list);
		free(p);
	}
}

void clear_wdslist(struct agent *a)
{
	struct netif_wds *p, *tmp;

	list_for_each_entry_safe(p, tmp,  &a->wdslist, list) {
		netif_free_wds(p);
	}
}

int agent_map_sub_cb(void *bus, void *priv, void *data)
{
	struct blob_attr *msg = (struct blob_attr *)data;
	char *str;


	str = blobmsg_format_json(msg, true);
	trace("Received notification '%s'\n", str);
	free(str);

	ieee1905_cmdu_event_handler(priv, msg);

	return 0;
}

int agent_map_del_cb(void *bus, void *priv, void *data)
{
	struct agent *a = (struct agent *)priv;
	uint32_t *obj = (uint32_t *)data;

	a->subscribed = false;
	fprintf(stdout, "Object 0x%x no longer present\n", *obj);

	return 0;
}

static int agent_subscribe_for_cmdus(struct agent *a)
{
	mapmodule_cmdu_mask_t cmdu_mask = {0};
	uint32_t map_id;
	int ret;


	map_prepare_cmdu_mask(cmdu_mask,
			CMDU_TYPE_TOPOLOGY_DISCOVERY,
			CMDU_TYPE_TOPOLOGY_NOTIFICATION,
			CMDU_TYPE_TOPOLOGY_QUERY,
			CMDU_TYPE_TOPOLOGY_RESPONSE,
			CMDU_TYPE_VENDOR_SPECIFIC,
			CMDU_TYPE_LINK_METRIC_QUERY,
			CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
			CMDU_TYPE_AP_AUTOCONFIGURATION_WSC,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW,
			CMDU_TYPE_HIGHER_LAYER_QUERY,
			CMDU_1905_ACK,
			CMDU_BEACON_METRICS_QUERY,
			CMDU_BACKHAUL_STEER_REQUEST,
			CMDU_AP_METRICS_QUERY,
			CMDU_ASSOC_STA_LINK_METRICS_QUERY,
			CMDU_UNASSOC_STA_LINK_METRIC_QUERY,
			CMDU_CHANNEL_SCAN_REQUEST,
			CMDU_CHANNEL_SCAN_REPORT,
			CMDU_POLICY_CONFIG_REQ,
			CMDU_BACKHAUL_STA_CAPABILITY_QUERY,
			CMDU_BACKHAUL_STA_CAPABILITY_REPORT,
			CMDU_CHANNEL_PREFERENCE_QUERY,
			CMDU_CHANNEL_SELECTION_REQ,
#ifdef OPER_CHAN_CHANGE_RELAY_MCAST
			CMDU_OPERATING_CHANNEL_REPORT,
#endif
			CMDU_CLIENT_STEERING_REQUEST,
			CMDU_CLIENT_ASSOC_CONTROL_REQUEST,
			CMDU_AP_CAPABILITY_QUERY,
			CMDU_AP_CAPABILITY_REPORT,
			CMDU_HIGHER_LAYER_DATA,
			CMDU_CLIENT_CAPABILITY_QUERY,
			CMDU_CAC_REQUEST,
			CMDU_CAC_TERMINATION,
#if (EASYMESH_VERSION > 2)
			CMDU_BSS_CONFIG_RESPONSE,
			CMDU_DPP_CCE_INDICATION,
			CMDU_PROXIED_ENCAP_DPP,
			CMDU_DPP_BOOTSTRAPING_URI,
			CMDU_DIRECT_ENCAP_DPP,
			CMDU_AGENT_LIST,
			CMDU_SERVICE_PRIORITIZATION_REQUEST,
#endif
			-1);

	memcpy(a->cmdu_mask, cmdu_mask, sizeof(a->cmdu_mask));

	trace("<----------------------------------- %s\n", __func__);
	for (;;) {
		ret = ubus_lookup_id(a->ubus_ctx, map_plugin, &map_id);
		if (!ret)
			break;

		trace("ieee1905.map not up yet, sleeping for 2s!\n");
		sleep(2);
	}

	a->map_oid = map_id;

	ret = map_subscribe(a->ubus_ctx,
			    &a->map_oid,
			    "mapagent", &cmdu_mask, a,
			    agent_map_sub_cb,
			    agent_map_del_cb,
			    &a->subscriber);
	if (!ret) {
		a->subscribed = true;
	} else {
		trace("mapagent: Failed to 'register' with %s (err = %s)\n",
		      map_plugin, ubus_strerror(ret));
	}

	return ret;
}

int agent_init_defaults(struct agent *a)
{
	INIT_LIST_HEAD(&a->aplist);
	INIT_LIST_HEAD(&a->bklist);
	INIT_LIST_HEAD(&a->ethlist);
	INIT_LIST_HEAD(&a->framelist);
	INIT_LIST_HEAD(&a->nodelist);
	INIT_LIST_HEAD(&a->wdslist);
#if (EASYMESH_VERSION > 2)
	INIT_LIST_HEAD(&a->dpp_ctx.enrolleelist);
	INIT_LIST_HEAD(&a->dpp_ctx.chirplist);
#endif

	a->cntlr_select.local = CONTROLLER_SELECT_LOCAL;
	a->cntlr_select.auto_detect = CONTROLLER_SELECT_AUTODETECT;
	a->cntlr_select.probe_int = CONTROLLER_SELECT_PROBE_INT;
	a->cntlr_select.retry_int = CONTROLLER_SELECT_RETRY_INT;
#ifdef AGENT_DYNAMIC_CNTRL
	a->cntlr_select.autostart = CONTROLLER_SELECT_AUTOSTART;
#endif

	/* don't try to start cntlr right away */
	timestamp_update(&a->observed_time);
	timestamp_update(&a->last_unassoc);

	a->cntlr_miss_count = 0;
	a->active_cntlr = false;
	a->multiple_cntlr = false;
	a->autocfg_interval = CONTROLLER_SELECT_PROBE_INT;


	a->is_sta_steer_start = 0;
	a->sta_steerlist_count = 0;
	timer_init(&a->sta_steer_req_timer, agent_steering_opp_timeout);

#if (EASYMESH_VERSION > 2)
	INIT_LIST_HEAD(&a->qos_rules);
#endif

	return 0;
}

static int agent_ackq_timeout_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *e)
{
	struct agent *a = container_of(q, struct agent, cmdu_ack_q);
	struct cmdu_buff *cmdu = (struct cmdu_buff *) e->cookie;
	int ret;
	uint16_t msgid;

	trace("%s: ---> cmdu = %04x to "MACFMT" \n", __func__,
		cmdu_get_type(cmdu), MAC2STR(cmdu->origin));

	if (e->resend_cnt-- > 0) {
		ret = ieee1905_ubus_send_cmdu(a->ubus_ctx, cmdu, &msgid, a->pvid);
		if (ret)
			err("%s fail to send cmdu\n", __func__);
		dbg("%s CMDU sent, msgid = %d\n", __func__, msgid);
		return CMDU_ACKQ_TMO_REARM;
	}

	return CMDU_ACKQ_TMO_DELETE;
}

static void agent_ackq_delete_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *e)
{
	struct cmdu_buff *cmdu = (struct cmdu_buff *) e->cookie;

	trace("%s: ---> cmdu = %04x to "MACFMT" \n", __func__,
		cmdu_get_type(cmdu), MAC2STR(cmdu->origin));

	cmdu_free(cmdu);
}

void run_agent(void)
{
	struct agent *w;
	struct ubus_context *ctx;

	set_sighandler(SIGHUP, agent_sighandler);
	set_sighandler(SIGPIPE, SIG_IGN);

	w = calloc(1, sizeof(*w));
	if (!w)
		return;

	this_agent = w;
	dbg("Starting wifi_agent... (&agent = %p)\n", w);

	agent_init_defaults(w);

	cmdu_ackq_init(&w->cmdu_ack_q);
	w->cmdu_ack_q.timeout_cb = agent_ackq_timeout_cb;
	w->cmdu_ack_q.delete_cb = agent_ackq_delete_cb;

	uloop_init();
	ctx = ubus_connect(NULL);
	if (!ctx) {
		err("Failed to connect to ubus\n");
		free(w);
		return;
	}

	w->ubus_ctx = ctx;
	w->evh.cb = agent_event_handler;

#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
	timer_init(&w->dpp_legacy_timeout, dpp_legacy_timeout_cb);

	if (dpp_init(&w->dpp, 0, NULL)) {
		fprintf(stderr, "Failed to init dpp context\n");
		goto out_ubus;
	}

	dpp_set_ctx_private_data(w->dpp, w);
	dpp_register_cb(w->dpp, dpp_frame_handler);
	w->own_peer = dpp_create_responder_instance(w->dpp);
	if (!w->own_peer) {
		fprintf(stderr, "Failed to create enrollee instance\n");
		goto out_ubus;
	}

	strcpy(w->own_peer->e_req.tech, "map");
	strcpy(w->own_peer->e_req.dpp_name, "test_map_dpp");
	w->own_peer->e_req.name_len = strlen(w->own_peer->e_req.dpp_name);
	w->own_peer->e_req.netrole = DPP_NETROLE_MAPAGENT;

	dpp_print_bootstrap_info(w->own_peer->own_bi);
#endif
#endif

	agent_config_init(w, &w->cfg);
#ifdef VENDOR_EXTENSION
	agent_load_extensions(w);
#endif

	w->ts.nl_main_sk = nl_init_main_sock(w);
	if (!w->ts.nl_main_sk)
		goto out_ubus;

	w->ts.nl_sk.cb = nl_event_uloop_cb;
	w->ts.nl_sk.fd = nl_socket_get_fd(w->ts.nl_main_sk);
	uloop_fd_add(&w->ts.nl_sk, ULOOP_READ);
#if (EASYMESH_VERSION > 2)
	w->dpp_ctx.dpp_fd = dpp_create_server(8908);
	//w->dpp_ctx.dpp_fd = connect_to_server("0.0.0.0", 8908);
	if (w->dpp_ctx.dpp_fd) {
		int ret;

		w->dpp_ctx.dpp_sk.cb = dpp_event_uloop_cb;
		w->dpp_ctx.dpp_sk.fd = w->dpp_ctx.dpp_fd;
		ret = uloop_fd_add(&w->dpp_ctx.dpp_sk, ULOOP_READ | ULOOP_EDGE_TRIGGER);
		err("setup dpp_sk!! ret:%d fd:%d\n", ret, w->dpp_ctx.dpp_fd);
	}

	if (w->cfg.persist_uri) {
		/* attempt to read ec private key */
		dpp_get_keys(w);

		/* if no found, generate one and write to the file */
		if (!w->dpp_ctx.ec_key)
			dpp_gen_keypair(w);
	}
#endif

	ubus_add_uloop(ctx);

	/* used by some ts script handling */
	setenv("AL_BRIDGE", w->cfg.al_bridge, 1);

	agent_init_wsc_attributes(w);

	//memcpy(w->cntlr_almac, w->cfg.cntlr_almac, 6);

	//agent_config_dump(&w->cfg);

	get_registered_steer_rules();	/* TODO: return rule list and improve */

	//agent_switch_according_to_pref(w); /*switch to the channel according to the prefrence*/

	ubus_register_event_handler(ctx, &w->evh, "ethport");
	ubus_register_event_handler(ctx, &w->evh, "wifi.*");
	ubus_register_event_handler(ctx, &w->evh, "wps_credentials");
	ubus_register_event_handler(ctx, &w->evh, "ubus.object.*");

	timer_init(&w->autocfg_dispatcher, agent_dispatch_autoconfig);
	timer_init(&w->reload_scheduler, agent_reload_cb);
	timer_init(&w->cntlr_scheduler, agent_enable_local_cntlr);
	timer_init(&w->rcpi_threshold_timer, agent_rcpi_thresold_timer_cb);
	timer_init(&w->onboarding_scheduler, agent_trigger_bsta_sync);
	timer_init(&w->bh_lost_timer, dynbh_bh_lost_cb);
	timer_init(&w->bh_reconf_timer, dynbh_bh_reconf_cb);
#ifdef AGENT_ISLAND_PREVENTION
	timer_init(&w->sta_disconnect_timer, agent_sta_disconnnect_cb);
	timer_init(&w->fh_disable_timer, agent_fh_disable_cb);
#endif /* AGENT_ISLAND_PREVENTION */
	timer_init(&w->disable_unconnected_bstas_scheduler, dynbh_disable_unconnected_bsta_cb);
	timer_init(&w->dynbh_scheduler, dynbh_upgrade_backhaul_cb);
	timer_init(&w->init_ifaces_scheduler, agent_init_ifaces_cb);
	timer_init(&w->enable_fhs_scheduler, agent_enable_fhs_cb);
	timer_init(&w->radio_stats_scheduler, agent_radio_stats_cb);
	timer_init(&w->boot_scan_scheduler, agent_boot_scan_cb);
	timer_init(&w->agent_periodic_run, agent_periodic_run);

	timer_init(&w->cfg.metric_report_timer, agent_metric_report_timer_cb);

	agent_init_interfaces(w);
	agent_init_ifaces_assoc_mode(w);
	/* switch to the channel according to the prefrence */
	//agent_switch_according_to_pref(w);

	agent_get_system_info(w);

	/* w->cfg.enabled */
	agent_publish_object(w, MAPAGENT_OBJECT);
	agent_publish_dbg_object(w, MAPAGENT_DBG_OBJECT);

	//run_agent(w);

	timer_set(&w->rcpi_threshold_timer, RCPI_THRESHOLD_TIMER);
	timer_set(&w->autocfg_dispatcher, 0 * 1000);
	timer_set(&w->radio_stats_scheduler, RADIO_STATS_TIMER);
	w->boot_scan_tries = 0;
	timer_set(&w->boot_scan_scheduler, BOOT_UP_SCAN_TIME * 1000);
	timer_set(&w->agent_periodic_run, 1 * 1000);

	agent_subscribe_for_cmdus(w);

	nl_wds_start_all(w);

	uloop_run();

/* out_and_exit: */
	map_unsubscribe(w->ubus_ctx, w->subscriber);
	agent_free_radios(w);
	agent_remove_object(w);
	agent_remove_dbg_object(w);
	agent_config_clean(&w->cfg);
	agent_free_cntlr_sync(w);
	cmdu_ackq_free(&w->cmdu_ack_q);
	clear_aplist(w);
	clear_bklist(w);
	clear_wdslist(w);
#if (EASYMESH_VERSION > 2)
	qos_rule_free_all(&w->qos_rules);
#endif
	ubus_unregister_event_handler(ctx, &w->evh);
#ifdef VENDOR_EXTENSION
	agent_unload_extensions(w);
#endif
	uloop_done();
	nl_free_main_sock(w->ts.nl_main_sk);
	agent_free_nodes(w);
out_ubus:
	ubus_free(ctx);
	free(w);
}

struct wsc_data *agent_free_wsc_data(struct wsc_data *wsc)
{
	if (wsc->m1_frame)
		free(wsc->m1_frame);
	if (wsc->key) {
		free(wsc->key->key);
		free(wsc->key);
	}

	return NULL;
}

static void agent_free_nodes(struct agent *a)
{
	struct node *n, *tmp;

	list_for_each_entry_safe(n, tmp, &a->nodelist, list) {
		list_del(&n->list);
		free(n);
	}
}

void agent_free_radios(struct agent *a)
{
	int i;

	trace("%s: a->num_radios to free = %d\n", __func__, a->num_radios);
	for (i = 0; i < a->num_radios; i++) {
		struct wifi_radio_element *re;

		re = &a->radios[i];

		free_scanlist(re);

		free(re->bsslist);
		re->bsslist = NULL;
		free(re->unassoc_stalist);
		re->unassoc_stalist = NULL;
		re->num_unassoc_sta = 0;
		agent_free_wsc_data(&re->autconfig);
		re->autconfig.key = NULL;
		re->autconfig.m1_frame = NULL;
		re->num_bss = 0;

		timer_del(&re->preference_report_timer);
	}
	a->num_radios = 0;
	memset(&a->radios, 0, sizeof(a->radios));
}

void agent_free_cntlr_sync(struct agent *a)
{
	if (a->sync_config_req)
		free(a->sync_config_req);

	if (a->privkey) {
		struct wsc_key *k = (struct wsc_key *) a->privkey;

		free(k->key);
		free(k);
	}
}

static void wifi_sta_restrict_timeout(atimer_t *t)
{
	struct restrict_sta_entry *s =
		container_of(t, struct restrict_sta_entry, restrict_timer);

	wifi_restrict_sta(s->fh_ifname, s->sta, 0);
	list_del(&s->list);
	free(s);
}

/* following implements 'wifi.agent' ubus methods */
int wifiagent_assoc_control_sta(char *fh_ifname, unsigned char *sta,
							int enable, int tmo)
{
	struct agent *a = this_agent;
	int ret = -1;

	if (!a) {
		dbg("this wifiagent is NULL\n");
		return -1;
	}

	if (fh_ifname && fh_ifname[0]) {
		struct netif_ap *p;

		p = get_netif_by_name(a, fh_ifname);
		if (!p || !p->cfg)
			return -1;

		if (!enable) {
			struct restrict_sta_entry *s, *tmp;
			char stastr[18] = {0};

			ret = wifi_restrict_sta(fh_ifname, sta, 0);

			list_for_each_entry_safe(s, tmp,
					&p->restrict_stalist, list) {
				if (!memcmp(s->sta, sta, 6)) {
					timer_del(&s->restrict_timer);
					list_del(&s->list);
					free(s);
				}
			}

			/* remove from config, if there is one */
			hwaddr_ntoa(sta, stastr);
			ret |= config_update2("wifiagent", &a->cfg,
					"ap", "ifname", fh_ifname,
					"restrict", enable, stastr, 18);

			return ret;
		}

		/* If timeout is positive, create a timed sta assoc-control
		 * entry but don't update config. At the timer expiry, remove
		 * sta from assoc-control list.
		 *
		 * If timeout is 0 or negetive, permanently add sta to the
		 * ap's assoc control list by creating a config entry.
		 */
		ret = wifi_restrict_sta(fh_ifname, sta, 1);

		if (tmo > 0) {
			struct restrict_sta_entry *s;

			s = calloc(1, sizeof(struct restrict_sta_entry));
			if (s) {
				memcpy(s->sta, sta, 6);
				snprintf(s->fh_ifname, 16, "%s", fh_ifname);
				timer_init(&s->restrict_timer, wifi_sta_restrict_timeout);
				timer_set(&s->restrict_timer, tmo * 1000);
				list_add_tail(&s->list, &p->restrict_stalist);
			}
		} else {
			char stastr[18] = {0};

			/* make entry persistent as no timeout is specified */
			hwaddr_ntoa(sta, stastr);
			ret |= config_update2("wifiagent", &a->cfg,
					"ap", "ifname", fh_ifname,
					"restrict", enable, stastr, 18);
		}
	}

	return ret;
}

int wifiagent_steer_sta(struct ubus_context *ctx,
			char *ifname,
			unsigned char *sta,
			int prefcnt, struct pref_neighbor *pref, int optime)
{
	struct agent *a = this_agent;
	struct sta *s;
	struct nbr *wifi_nbr_list = NULL;
	int i;
	int ret = 0;

	UNUSED(ctx);

	if (!a) {
		dbg("this wifiagent is NULL\n");
		return -1;
	}

	s = find_sta_by_mac(a, sta);
	if (!s) {
		dbg("STA not found!\n");
		return -1;
	}

	if (optime) {
		s->steer_policy |= STA_STEER_OPPORTUNITY;
		if (!timer_pending(&s->sta_steer_timer)) {
			rebuild_cntlr_preflist(a, s, prefcnt, pref);
			/* (Re)start steer opportunity timer.
			 * In the opportunity time window constantly check
			 * every second for this sta's steering opportinity.
			 */
			s->steer_opportunity_tmo = optime * 1000;
			clock_gettime(CLOCK_MONOTONIC, &s->steer_opportunity);
			timer_set(&s->sta_steer_timer, 1000);
			return 0;
		}
	}

	/* steer mandate */
	s->steer_policy &= ~STA_STEER_OPPORTUNITY;
	if (/* !(s->caps & STA_CAP_11V_BSS_TRANS) || */   /* TODO */
			!(s->steer_policy & STEER_BTM)) {
		return steer_sta_legacy(s);
	}

	/* Translate internal 'pref_neighbor' to wifi's 'nbr' struct */
	wifi_nbr_list = calloc(prefcnt, sizeof(struct nbr));
	if (WARN_ON(!wifi_nbr_list))
		return -ENOMEM;

	for (i = 0; i < prefcnt; i++) {
		agent_pref_neighbor_to_nbr(&pref[i], &wifi_nbr_list[i]);
	}

	ret = agent_req_btm_steer(ifname, s->macaddr, prefcnt, wifi_nbr_list, 0);

	if (ret)
		dbg("Failed to send BTM request to " MACFMT "\n",
		    MAC2STR(s->macaddr));

	free(wifi_nbr_list);

	return ret;
}

int wifiagent_get_status(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	struct agent *agent = this_agent;
	struct netif_ap *p = NULL;
	struct netif_bk *bk = NULL;
	struct neighbor *n;
	struct blob_buf bb;
	void *a, *b, *c, *t;
	int i;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	a = blobmsg_open_array(&bb, "radios");
	for (i = 0; i < agent->num_radios; i++) {
		struct wifi_radio_element *re = &agent->radios[i];
		char macaddrstr[18] = {0};

		hwaddr_ntoa(re->macaddr, macaddrstr);
		t = blobmsg_open_table(&bb, "");
		blobmsg_add_string(&bb, "name", re->name);
		blobmsg_add_string(&bb, "macaddr", macaddrstr);
		blobmsg_add_u32(&bb, "channel", re->current_channel);
		blobmsg_add_u32(&bb, "bandwidth", re->current_bandwidth);
		blobmsg_add_u32(&bb, "opclass", re->current_opclass);
		b = blobmsg_open_table(&bb, "autoconfig");
		blobmsg_add_string(&bb, "state", (re->state ? "ACTIVE":"HEARTBEAT"));
		c = blobmsg_open_table(&bb, "message_identifier");
		blobmsg_add_u32(&bb, "search", re->mid);
		blobmsg_add_u32(&bb, "wsc", re->wsc_mid);
		blobmsg_add_u32(&bb, "renew", re->renew_mid);
		blobmsg_close_table(&bb, c);
		blobmsg_close_table(&bb, b);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	a = blobmsg_open_array(&bb, "fronthaul");
	list_for_each_entry(p, &agent->aplist, list) {
		struct sta *s;
		char bssidstr[18] = {0};
		void *tt;

		if (!p->cfg)
			continue;

		hwaddr_ntoa(p->bssid, bssidstr);
		t = blobmsg_open_table(&bb, "");
		blobmsg_add_string(&bb, "name", p->ifname);
		blobmsg_add_u8(&bb, "enabled", p->enabled);
		blobmsg_add_string(&bb, "bssid", bssidstr);
		blobmsg_add_string(&bb, "ssid", p->ssid);
		blobmsg_add_u32(&bb, "channel", p->channel);
		blobmsg_add_u32(&bb, "load", p->bssload);

		b = blobmsg_open_array(&bb, "neighbor");
		list_for_each_entry(n, &p->nbrlist, list) {
			char nbr_bssidstr[18] = {0};

			hwaddr_ntoa(n->nbr.bssid, nbr_bssidstr);
			tt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "bssid", nbr_bssidstr);
			blobmsg_add_u32(&bb, "channel", n->nbr.channel);
			blobmsg_close_table(&bb, tt);
		}
		blobmsg_close_array(&bb, b);

		b = blobmsg_open_array(&bb, "stations");
		list_for_each_entry(s, &p->stalist, list) {
			void *aa, *ttt, *tttt;
			char stastr[18] = {0};
			struct pref_neighbor *pn;

			hwaddr_ntoa(s->macaddr, stastr);

			ttt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "addr", stastr);
			/* blobmsg_add_u32(&bb, "conn_time", s->connected_ms / 1000); */
			blobmsg_add_u32(&bb, "rssi", s->rssi[0]);
			blobmsg_add_u32(&bb, "bcnreport_capable",
					s->supports_bcnreport);
			blobmsg_add_u32(&bb, "btmreq_steers",
					s->steer_btm_cnt);
			blobmsg_add_u32(&bb, "legacy_steers",
					s->legacy_steered);
			aa = blobmsg_open_array(&bb, "neighbor");
			list_for_each_entry(pn, &s->pref_nbrlist, list) {
				char pn_bssidstr[18] = {0};

				hwaddr_ntoa(pn->bssid, pn_bssidstr);
				tttt = blobmsg_open_table(&bb, "");
				blobmsg_add_string(&bb, "bssid", pn_bssidstr);
				blobmsg_add_u32(&bb, "rssi", pn->rssi);
				blobmsg_close_table(&bb, tttt);
			}
			blobmsg_close_array(&bb, aa);
			blobmsg_close_table(&bb, ttt);
		}
		blobmsg_close_array(&bb, b);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	a = blobmsg_open_array(&bb, "backhauls");
	list_for_each_entry(bk, &agent->bklist, list) {
		char macaddr[18] = {0};
		char bssidstr[18] = {0};
		char bssidcfgstr[18] = {0};

		if (!bk->cfg)
			continue;

		hwaddr_ntoa(bk->macaddr, macaddr);
		hwaddr_ntoa(bk->bssid, bssidstr);
		hwaddr_ntoa(bk->cfg->bssid, bssidcfgstr);
		t = blobmsg_open_table(&bb, "");
		blobmsg_add_string(&bb, "name", bk->ifname);
		blobmsg_add_u8(&bb, "enabled", bk->enabled);
		blobmsg_add_u8(&bb, "connected", bk->connected);
		blobmsg_add_string(&bb, "macaddr", macaddr);
		blobmsg_add_string(&bb, "bssid", bssidstr);
		blobmsg_add_string(&bb, "ssid", bk->ssid);
		blobmsg_add_u32(&bb, "channel", bk->channel);

		b = blobmsg_open_table(&bb, "cfg");
		blobmsg_add_string(&bb, "name", bk->cfg->name);
		blobmsg_add_u8(&bb, "enabled", bk->cfg->enabled);
		blobmsg_add_u32(&bb, "band", bk->cfg->band);
		blobmsg_add_string(&bb, "device", bk->cfg->device);
		blobmsg_add_string(&bb, "ssid", bk->cfg->ssid);
		blobmsg_add_string(&bb, "key", bk->cfg->key);
		blobmsg_add_string(&bb, "encryption", bk->cfg->encryption);
		blobmsg_add_u8(&bb, "onboarded", bk->cfg->onboarded);
		blobmsg_add_string(&bb, "bssid", bssidcfgstr);
		blobmsg_add_u16(&bb, "priority", (uint16_t) bk->cfg->priority);
		blobmsg_close_table(&bb, b);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);


	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int wifiagent_get_nodes(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	struct agent *agent = this_agent;
	struct node *n;
	struct blob_buf bb;
	void *a;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "num_nodes", agent->num_nodes);

	a = blobmsg_open_array(&bb, "nodes");
	list_for_each_entry(n, &agent->nodelist, list) {
		char aladdrstr[18] = {0};
		void *t;

		hwaddr_ntoa(n->alid, aladdrstr);
		t = blobmsg_open_table(&bb, "");

		blobmsg_add_string(&bb, "almac", aladdrstr);
		blobmsg_add_u16(&bb, "profile", n->map_profile);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int wifiagent_get_info(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	char almac_str[18] = {};
	struct blob_buf bb;
	struct agent *a = this_agent;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "active_cntlr", a->active_cntlr);
	hwaddr_ntoa(a->cntlr_almac, almac_str);
	blobmsg_add_string(&bb, "cntlr_almac", almac_str);
	blobmsg_add_u32(&bb, "local_cntlr", is_local_cntlr_running());
	blobmsg_add_u32(&bb, "multiple_cntlr_found", a->multiple_cntlr);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int wifiagent_get_bk_info(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	struct blob_buf bk = { 0 };
	struct blob_attr *tb[4];
	static const struct blobmsg_policy bk_attr[4] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "backhaul_device_id", .type = BLOBMSG_TYPE_TABLE },
		[3] = { .name = "backhaul_device_macaddr", .type = BLOBMSG_TYPE_TABLE },
	};
	int ret;

	blob_buf_init(&bk, 0);

	if (!blobmsg_add_json_from_file(&bk, MAP_UPLINK_PATH)) {
		dbg("%s: Failed to parse %s\n", __func__, MAP_UPLINK_PATH);
		goto out;
	}

	ret = blobmsg_parse(bk_attr, 4, tb, blob_data(bk.head), blob_len(bk.head));
	if (ret)
		goto out;

	ubus_send_reply(ctx, req, bk.head);
	blob_buf_free(&bk);

	return UBUS_STATUS_OK;
out:
	blob_buf_free(&bk);
	return UBUS_STATUS_UNKNOWN_ERROR;
}


int agent_switch_according_to_pref(struct agent *a)
{
	uint32_t channel = 0;
	uint32_t opclass = 0;
	int ret = 0;
	int l = 0;
	struct wifi_radio_element *radio;

	trace("agent: %s: --->\n", __func__);

	for (l = 0; l < a->num_radios; l++) {
		radio = a->radios + l;
		ret = agent_get_highest_preference(radio, radio->current_opclass, &channel,
			&opclass);

		/* The operating class channel preference has been set
		 * now we want to switch the channel to the max preference */
		trace("agent|%s: %d|: opclass is %d channel is %d--->\n",
				__func__, __LINE__, opclass, channel);

		if ((radio->current_opclass == opclass) && (radio->current_channel == channel))
			continue;

		if (channel != 0 && opclass != 0) {
			ret = agent_channel_switch(a, radio->macaddr, channel, opclass);
			if (ret == -1) {
				/* Here we need to set the default preference for all the
				 * channels in that operating class also set the response
				 * code as rejected*/
				agent_set_channel_preference_to_default(radio);
			}
		}
	}
	return 0;
}

void agent_set_post_scan_action_pref(struct agent *agent, const char *radio, bool opclass_preferences)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(agent->radios); i++) {
		if (radio && strcmp(radio, agent->radios[i].name))
			continue;

		agent->radios[i].post_scan_action.opclass_preferences = opclass_preferences;
	}
}

/* Allocates channel scan report CMDU and puts Timestamp TLV into it */
struct cmdu_buff *agent_prepare_scan_cmdu(struct agent *a, struct tlv *tsp)
{
	trace("%s --->\n", __func__);

	struct cmdu_buff *cmdu_data;

	/* Allocate new fragment CMDU */
	cmdu_data = cmdu_alloc_frame(CH_SCAN_RESP_CMDU_MAX_LEN);
	if (!cmdu_data) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}
	cmdu_set_type(cmdu_data, CMDU_CHANNEL_SCAN_REPORT);
	memcpy(cmdu_data->origin, a->cntlr_almac, 6);

	/* Put timestamp TLV to the fragment CMDU */
	if (cmdu_copy_tlvs(cmdu_data, &tsp, 1)) {
		dbg("%s: error: cmdu_put_tlv()\n", __func__);
		return NULL;
	}
	dbg("|%s:%d| added MAP_TLV_TIMESTAMP\n", __func__, __LINE__);

	return cmdu_data;
}

/* Splits Channel Scan Report TLVs into fragment CMDUs.
 * Sends fragmet CMDUs to controller.
 */
static int agent_send_ch_scan_rsp_frag(struct agent *a, struct cmdu_buff *cmdu)
{
	trace("agent: %s: --->\n", __func__);

	struct tlv_policy d_policy_scan[] = {
		[0] = {
			.type = MAP_TLV_CHANNEL_SCAN_RES,
			.present = TLV_PRESENT_MORE,
			.minlen = 9
		}
	};
	struct tlv *tv_tsp[1][TLV_MAXNUM] = {0};	/* Timestamp TLV */
	struct tlv *tv_scan[256] = {0};	/* Channel Scan Results TLVs */
	int i;
	int num_tlv = 256, num_tlv_copied = 0;
	int ret = 0;
	struct cmdu_buff *frag_cmdu = NULL;

	if (cmdu->datalen < CH_SCAN_RESP_CMDU_MAX_LEN)
		/* No fragmentation required, just send the CMDU as is */
		return agent_send_cmdu(a, cmdu);

	/* If the number of neighbors detected during a channel scan would
	 * mean that the channel scan report message would not fit within
	 * one 1905 CMDU, the Multi-AP Agent shall split the channel scan
	 * report across multiple Channel Scan Result TLVs by splitting
	 * the information related to sets of neighbor BSSs into separate
	 * Channel Scan Result TLVs and setting the NumberofNeighbors field
	 * to the number of neighbors contained in the corresponding TLV.
	 */

	/* Note: assuming neighbors are already evenly split between TLVs */

	ret = map_cmdu_parse_tlvs(cmdu, tv_tsp, 2, a->cfg.map_profile);
	if (ret) {
		dbg("%s: cmdu_parse_tlvs(profile=%d) failed, err = (%d) '%s'\n",
		    __func__, a->cfg.map_profile, ieee1905_error,
		    ieee1905_strerror(ieee1905_error));
		return -1;
	}

	if (!tv_tsp[0][0]) {
		dbg("%s: Missing TIMESTAMP_TLV!\n", __func__);
		return -1;
	}

	ret = cmdu_parse_tlv_single(cmdu, tv_scan, d_policy_scan, &num_tlv);
	if (ret) {
		dbg("%s: map_cmdu_parse_tlvs failed,  err = (%d) '%s'\n",
		    __func__, map_error, map_strerror(map_error));
		return -1;
	}

	if (!tv_scan[0]) {
		dbg("%s: Missing CHANNEL SCAN RESULT_TLV!\n", __func__);
		return -1;
	}

	tv_tsp[0][0]->len = tlv_length(tv_tsp[0][0]); // FIXME
	frag_cmdu = agent_prepare_scan_cmdu(a, tv_tsp[0][0]);
	if (WARN_ON(!frag_cmdu))
		return -1;

	for (i = 0; i < num_tlv; i++) {
		uint16_t tlv_len = tlv_length(tv_scan[i]);

		if (tlv_total_length(tv_scan[i]) > frag_cmdu->end - frag_cmdu->tail) {
			/* No space left for current TLV in CMDU buffer */
			cmdu_put_eom(frag_cmdu);
			agent_send_cmdu(a, frag_cmdu);
			cmdu_free(frag_cmdu);

			num_tlv_copied = 0;

			/* Create next fragment and put next TLVs into it */
			frag_cmdu = agent_prepare_scan_cmdu(a, tv_tsp[0][0]);
		}

		/* Add TLV to CMDU & continue */
		tv_scan[i]->len = tlv_len; // FIXME: assign len smwhr else
		if (cmdu_copy_tlvs(frag_cmdu, &tv_scan[i], 1)) {
			dbg("%s:%d copy TLVs failed, end = %p, tail = %p, len = %d\n",
			    __func__, __LINE__,
			    frag_cmdu->end, frag_cmdu->tail, tv_scan[i]->len);
			return -1;
		}
		num_tlv_copied++;
	}

	if (num_tlv_copied) { /* avoid sending timestamp alone */
		cmdu_put_eom(frag_cmdu);
		agent_send_cmdu(a, frag_cmdu);
	}

	cmdu_free(frag_cmdu);

	return 0;
}

int agent_send_ch_scan_response(struct agent *a, struct wifi_netdev *ndev,
		struct wifi_scan_request_radio *req)
{
	struct cmdu_buff *cmdu_data = NULL;
	int ret = 0;

	dbg("%s: called.\n", __func__);

	/* Generate the response cmdu */
	if (WARN_ON(!ndev))
		return -1;

	if (!req)
		/* Independent channel scan response */
		cmdu_data = agent_gen_independent_ch_scan_response(a, ndev);
	else
		cmdu_data = agent_gen_ch_scan_response_radio(a,
						ndev, req, req->status);

	if (!cmdu_data)
		return -1;

	if (req && req->status == CH_SCAN_STATUS_SCAN_ABORTED) {

		/* Replace old request data with the new one */
		ndev->scan_req = *req;

		/* Will mark as success when results available */
		ndev->scan_req.status = CH_SCAN_STATUS_SCAN_NOT_COMPLETED;

		/* Reset the timer of available time (5min) */
		timer_set(&ndev->available_scan_timer, 300000);
	}

	/* Send the response cmdu */
	ret = agent_send_ch_scan_rsp_frag(a, cmdu_data);

	cmdu_free(cmdu_data);

	return ret;
}

bool agent_ch_scan_succesful(struct agent *a)
{
	if (a->scan_status_code == CH_SCAN_STATUS_SUCCESS)
		return true;

	return false;
}
